/*
 * PureIRCd Internet Relay Chat Daemon, src/s_serv.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 * University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"
#include "dh.h"
#include "zlink.h"
#include "md5crypt.h"
#include "patchlevel.h"

#if defined(AIX) || defined(DYNIXPTX) || defined(SVR3)
#include <time.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"
#if defined( HAVE_STRING_H )
#include <string.h>
#else
/* older unices don't have strchr/strrchr .. help them out */
#include <strings.h>
#undef strchr
#define strchr index
#endif
#include "dich_conf.h"
#include "fdlist.h"
#include "throttle.h"

extern fdlist serv_fdlist;
extern int lifesux;
extern int HTMLOCK;
static char buf[BUFSIZE];
extern int rehashed;
extern int zline_in_progress;

#ifdef HIGHEST_CONNECTION
int max_connection_count = 1, max_client_count = 1;
#endif

/* external variables */
/* external functions */

#ifdef MAXBUFFERS
extern void reset_sock_opts();
#endif

#define AllocCpy(x, y) x = (char *)MyMalloc(strlen(y) + 1); strcpy(x, y)
aGline *expire_gline(aGline *);
aGline *glines = NULL;
aSline *expire_sline(aSline *);
aSline *slines = NULL;

void list_slines(aClient *);
void add_sline(char *, char *, char *, char *, time_t, time_t, char *);
void list_glines(aClient *);
void add_gline(char *, char *, char *, char *, time_t, time_t, char *);

extern char *smalldate(time_t);		/* defined in s_misc.c */
extern void outofmemory(void);		/* defined in list.c */
extern void s_die(void);		/* defined in ircd.c as VOIDSIG */
extern int match(char *, char *);	/* defined in match.c */
extern void report_conf_links(aClient *, aConfList *, int, char);
extern void show_opers(aClient *, char *);
extern void show_servers(aClient *, char *);
extern void count_memory(aClient *, char *);
extern void rehash_ip_hash();		/* defined in s_conf.c */
extern char *find_restartpass();
extern char *find_diepass();

/* Local function prototypes */
static int isnumber(char *);		/* return 0 if not, else return number */
static char *cluster(char *);

int send_rules(aClient *, aClient *, int, char **);
int send_motd(aClient *, aClient *, int, char **);
int send_opermotd(aClient *, aClient *, int, char **);
void read_rules(char *);
void read_motd(char *);
void read_opermotd(char *);
void read_help(char *);
void read_shortmotd(char *);

char glreason[512];
char rules_last_changed_date[MAX_DATE_STRING];		/* enough room for date */
char motd_last_changed_date[MAX_DATE_STRING];		/* enough room for date */
char opermotd_last_changed_date[MAX_DATE_STRING];	/* enough room for date */

static int flush_write(aClient *, char *, int, char *, int, char *);

#ifdef LOCKFILE
/* Shadowfax's lockfile code */
void do_pending_klines(void);

struct pkl
{
    char *comment;	/* Kline Comment */
    char *kline;	/* Actual Kline */
    struct pkl *next;	/* Next Pending Kline */
} *pending_klines = NULL;

time_t pending_kline_time = 0;
#endif /* LOCKFILE */

/*
 * m_version()
 *  parv[0] = sender prefix 
 *  parv[1] = remote server
 */
int m_version(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    extern char serveropts[];

    if (hunt_server(cptr, sptr, ":%s VERSION :%s", 1, parc, parv) == HUNTED_ISME)
    {
	sendto_one(sptr, rpl_str(RPL_VERSION), me.name, parv[0], version, debugmode, me.name, serveropts);
	sendto_umode(MODE_e, "/version requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	    HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
    }
    return 0;
}

/*
 * m_squit()
 * there are two types of squits: those going downstream (to the target server)
 * and those going back upstream (from the target server).
 * previously, it wasn't necessary to distinguish between these two types of 
 * squits because they neatly echoed back all of the QUIT messages during
 * an squit.  This, however, is no longer practical.
 * 
 * To clarify here, DOWNSTREAM signifies an SQUIT heading towards the target
 * server UPSTREAM signifies an SQUIT which has successfully completed,
 * heading out everywhere.
 *
 * acptr is the server being squitted.
 * a DOWNSTREAM squit is where the notice did not come from acptr->from.
 * an UPSTREAM squit is where the notice DID come from acptr->from.
 *
 *  parv[0] = sender prefix 
 *  parv[1] = server name 
 *  parv[2] = comment
 */
int m_squit(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;
    char *server;
    aClient *acptr;
    char *comment = (parc > 2) ? parv[2] : sptr->name;

    if (!IsPrivileged(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc > 1) 
    {
	server = parv[1];

	/* To accomodate host masking, a squit for a masked server
	 * name is expanded if the incoming mask is the same as the
	 * server name for that link to the name of link. */

	while ((*server == '*') && IsServer(cptr))
	{
	    aconf = cptr->serv->nline;

	    if (!aconf)
		break;
	    if (!mycmp(server, my_name_for_link(me.name, aconf)))
		server = cptr->name;
	    break; /* WARNING is normal here */
	    /* NOTREACHED */
	}

	/* The following allows wild cards in SQUIT. Only useful when the command is issued by an oper. */

	for (acptr = client; (acptr = next_client(acptr, server)); 
	     acptr = acptr->next)
	    if (IsServer(acptr) || IsMe(acptr))
		break;
	if (acptr && IsMe(acptr)) 
	{
	    acptr = cptr;
	    server = cptr->sockhost;
	}
    }
    else
    {
	/* This is actually protocol error. But, well, closing the link is very proper answer to that... */

	server = cptr->sockhost;
	acptr = cptr;
    }
    if (!acptr)
    {
	sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], server);
	return 0;
    }
    if (MyClient(sptr) && ((!OPCanGRoute(sptr) && !MyConnect(acptr)) || (!OPCanLRoute(sptr) && MyConnect(acptr)))) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    /* If the server is mine, we don't care about upstream or downstream, just kill it and do the notice. */
    
    if (MyConnect(acptr)) 
    {
	sendto_gnotice("from %s: Received SQUIT %s from %s (%s)", me.name, acptr->name, get_client_name(sptr, HIDEME), comment);
	sendto_serv_butone(NULL, ":%s GNOTICE :Received SQUIT %s from %s (%s)", me.name, server, get_client_name(sptr, HIDEME), comment);
#if defined(USE_SYSLOG) && defined(SYSLOG_SQUIT)
	ircd_log("SQUIT From %s : %s (%s)", parv[0], server, comment);
#endif
	/* I am originating this squit! Not cptr! */
	/* ack, but if cptr is squitting itself.. */

	if (cptr == sptr)
	{
	    exit_client(&me, acptr, sptr, comment);
	    return FLUSH_BUFFER; /* kludge */
	}
	return exit_client(&me, acptr, sptr, comment);
    }

    /* the server is not connected to me. Determine whether this is an upstream or downstream squit */

    if (sptr->from == acptr->from) /* upstream */
    {
	sendto_umode(MODE_d, "Exiting server %s due to upstream squit by %s [%s]", acptr->name, sptr->name, comment);
	return exit_client(cptr, acptr, sptr, comment);
    }

    /* fallthrough: downstream */

    if (!(IsUnconnect(acptr->from))) /* downstream not unconnect capable */
    {
	sendto_umode(MODE_d, "Exiting server %s due to non-unconnect server %s [%s]", acptr->name, acptr->from->name, comment);
	return exit_client(&me, acptr, sptr, comment);
    }
    sendto_umode(MODE_d, "Passing along SQUIT for %s by %s [%s]", acptr->name, sptr->name, comment);
    sendto_one(acptr->from, ":%s SQUIT %s :%s", parv[0], acptr->name, comment);
    return 0;
}

/*
 * ts_servcount() 
 * returns the number of TS servers that are connected to us
 */
int ts_servcount()
{
    int i;
    aClient *acptr;
    int r = 0;

    for (i = 0; i <= highest_fd; i++)
	if ((acptr = local[i]) && IsServer(acptr) && DoesTS(acptr))
	    r++;
    return r;
}

/*
 * m_svinfo()
 *  parv[0] = sender prefix 
 *  parv[1] = TS_CURRENT for the server 
 *  parv[2] = TS_MIN for the server 
 *  parv[3] = server is standalone or connected to non-TS only 
 *  parv[4] = server's idea of UTC time
 */
int m_svinfo(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    time_t deltat, tmptime, theirtime;

    if (!IsServer(sptr) || !MyConnect(sptr))
	return 0;
    if (parc == 2 && mycmp(parv[1], "ZIP") == 0)
    {
	SetZipIn(sptr);
	sptr->serv->zip_in = zip_create_input_session();
	sendto_gnotice("from %s: Input from %s is now compressed", me.name, cptr->name);
	sendto_serv_butone(sptr, ":%s GNOTICE :Input from %s is now compressed", me.name, cptr->name);
	return ZIP_NEXT_BUFFER;
    }
    if (parc < 5 || !DoesTS(sptr))
	return 0;
    if (TS_CURRENT < atoi(parv[2]) || atoi(parv[1]) < TS_MIN) 
    {
	/* a server with the wrong TS version connected; since we're
	 * TS_ONLY we can't fall back to the non-TS protocol so we drop the link  -orabidoo */

	sendto_umode(MODE_s, "Link %s dropped, wrong TS protocol version (%s,%s)", cptr->name, parv[1], parv[2]);
	return exit_client(sptr, sptr, sptr, "Incompatible TS version");
    }

    tmptime = time(NULL);
    theirtime = atol(parv[4]);
    deltat = abs(theirtime - tmptime);

    if (deltat > TS_MAX_DELTA) 
    {
	sendto_gnotice("from %s: Link %s dropped, excessive TS delta (my "
	    "TS=%d, their TS=%d, delta=%d)", me.name, cptr->name, tmptime, theirtime, deltat);
	sendto_serv_butone(sptr, ":%s GNOTICE :Link %s dropped, excessive TS delta (delta=%d)", me.name, cptr->name, deltat);
	return exit_client(sptr, sptr, sptr, "Excessive TS delta");
    }
    if (deltat > TS_WARN_DELTA) 
    {
	sendto_umode(MODE_s, "Link %s notable TS delta (my TS=%d, their TS=%d, delta=%d)", cptr->name, tmptime, theirtime, deltat);
    }
    return 0;
}

/*
 * m_tctrl()
 */
int m_tctrl(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char tempbuff[500];

    if (!IsServer(sptr) || !MyConnect(sptr) || parc < 14)
	return 0;
    if (!strcmp(BASENAME, parv[1]) == 0)
    {
	sendto_one(cptr, "ERROR :Your are running an incompatible IRCd.");
	sendto_gnotice("from %s: Link %s dropped, wrong IRCd version.", me.name, cptr->name);
	return exit_client(cptr, cptr, cptr, "Incompatible IRCd version");
	return 0;
    }
    if (!strcmp(DYN_NETWORK_NETNAME, parv[2]) == 0)
    {
	sendto_one(cptr, "ERROR :Your network name is (%s). All servers needs to have the same. We have (%s).", parv[2], DYN_NETWORK_NETNAME);
	sendto_umode(MODE_s, "Network name mismatch from link %s. Theirs: %s Ours: %s. Dropped link.",
	    cptr->name, parv[2], DYN_NETWORK_NETNAME);
	sendto_gnotice("from %s: Network name mismatch from link %s. Theirs: %s Ours: %s. Dropped link.",
	    me.name, cptr->name, parv[2], DYN_NETWORK_NETNAME);
	ircsprintf(tempbuff, "Your network name is (%s). All servers needs to have the same. We have (%s).", parv[2], DYN_NETWORK_NETNAME);
	return exit_client(cptr, cptr, cptr, tempbuff);
	return 0;
    }
    SetTctrl(cptr);
    return 0;
}

/*
 * m_vs()
 *  parv[0] = sender prefix
 *  parv[1] = target
 *  parv[2] = virtual host
 */
int m_vs(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered(sptr))
	return 0;
    if (!IsServer(cptr))
	return 0;
    if (strlen(parv[2]) < 1)
	return 0;
    if (strlen(parv[2]) > (HOSTLEN - 1))
	return 0;
    if ((acptr = find_person(parv[1], NULL)))
    {
	strncpyzt(acptr->user->virthost, parv[2], HOSTLEN);
	sendto_serv_butone(cptr, ":%s VS %s %s", parv[0], parv[1], parv[2]);
    }
    return 0;
}

/*
 * m_server()
 *  parv[0] = sender prefix 
 *  parv[1] = servername 
 *  parv[2] = serverinfo/hopcount 
 *  parv[3] = serverinfo
 */
int m_server(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    char info[REALLEN+1], *inpath, *host;
    aClient *acptr, *bcptr;
    aConfItem *aconf;
    int hop;
    char nbuf[HOSTLEN * 2 + USERLEN + 5]; /* same size as in s_misc.c */

    info[0] = '\0';
    inpath = get_client_name(cptr, HIDEME);

    if (parc < 2 || *parv[1] == '\0') 
    {
	sendto_one(cptr, "ERROR :No servername");
	return 0;
    }

    hop = 0;
    host = parv[1];

    if (parc > 3 && atoi(parv[2])) 
    {
	hop = atoi(parv[2]);
	strncpyzt(info, parv[3], REALLEN);
    }
    else if (parc > 2) 
    {
	strncpyzt(info, parv[2], REALLEN);

	if ((parc > 3) && ((i = strlen(info)) < (REALLEN - 2))) 
	{
	    (void) strcat(info, " ");
	    (void) strncat(info, parv[3], REALLEN - i - 2);
	    info[REALLEN] = '\0';
	}
    }

    /* July 5, 1997 
     * Rewritten to throw away server cruft from users, 
     * combined the hostname validity test with cleanup of host name,
     * so a cleaned up hostname can be returned as an error if
     * necessary. - Dianora */

    /* yes, the if(strlen) below is really needed!! */

    if (strlen(host) > HOSTLEN)
	host[HOSTLEN] = '\0';
    if (IsPerson(cptr)) 
    {
	/* A local link that has been identified as a USER tries something fishy... ;-) */

	sendto_one(cptr, err_str(ERR_UNKNOWNCOMMAND), me.name, parv[0], "SERVER");
	return 0;
    } 
    else /* hostile servername check */	
    {
	/* Lets check for bogus names and clean them up we don't bother
	 * cleaning up ones from users, becasuse we will never see them
	 * any more - Dianora */

	int bogus_server = 0;
	int found_dot = 0;
	char clean_host[(2 * HOSTLEN) + 1];
	char *s;
	char *d;
	int n;

	s = host;
	d = clean_host;
	n = (2 * HOSTLEN) - 2;

	while (*s && n > 0) 
	{
	    if ((unsigned char) *s < (unsigned char) ' ') /* Is it a control character? */
	    {
		bogus_server = 1;
		*d++ = '^';
		*d++ = (char) ((unsigned char) *s + 0x40); 
		/* turn it into a printable */
		n -= 2;
	    }
	    else if ((unsigned char) *s > (unsigned char) '~')
	    {
		bogus_server = 1;
		*d++ = '.';
		n--;
	    }
	    else 
	    {
		if (*s == '.')
		    found_dot = 1;
		*d++ = *s;
		n--;
	    }
	    s++;
	}
	*d = '\0';

	if ((!found_dot) || bogus_server) 
	{
	    sendto_one(sptr, "ERROR :Bogus server name (%s)", clean_host);
	    return exit_client(cptr, cptr, cptr, "Bogus server name");
	}
    }

    /* check to see this host even has an N line before bothering 
     * anyone about it. Its only a quick sanity test to stop the
     * conference room and win95 ircd dorks. Sure, it will be
     * redundantly checked again in m_server_estab() *sigh* yes there
     * will be wasted CPU as the conf list will be scanned twice. But
     * how often will this happen? - Dianora
     *
     * This should (will be) be recoded to check the IP is valid as well,
     * with a pointer to the valid N line conf kept for later, saving an
     * extra lookup.. *sigh* - Dianora */

    if (!IsServer(cptr)) 
    {
	if (find_conf_name(host, CONF_NOCONNECT_SERVER) == NULL)
	{
	    sendto_umode(MODE_s, "Link %s dropped, no N:Line", get_client_name(cptr, TRUE));
	    return exit_client(cptr, cptr, cptr, "NO N line");
	}
    }
    if ((acptr = find_name(host, NULL))) 
    {
	/* This link is trying feed me a server that I already have 
	 * access through another path -- multiple paths not accepted 
	 * currently, kill this link immediately!! 
	 *
	 * Rather than KILL the link which introduced it, KILL the 
	 * youngest of the two links. -avalon */

	bcptr = (cptr->firsttime > acptr->from->firsttime) ? cptr : acptr->from;
	sendto_one(bcptr, "ERROR :Server %s already exists", host);

	if (bcptr == cptr) 
	{
	    sendto_gnotice("from %s: Link %s cancelled, server %s already exists", me.name, get_client_name(bcptr, HIDEME), host);
	    sendto_serv_butone(bcptr, ":%s GNOTICE :Link %s cancelled, server %s already exists", me.name, get_client_name(bcptr, HIDEME), host);
	    return exit_client(bcptr, bcptr, &me, "Server Exists");
	}
	strcpy(nbuf, get_client_name(bcptr, HIDEME));
	sendto_gnotice("from %s: Link %s cancelled, server %s reintroduced by %s", me.name, nbuf, host, get_client_name(cptr, HIDEME));
	sendto_serv_butone(bcptr, ":%s GNOTICE :Link %s cancelled, server %s reintroduced by %s", me.name, nbuf, host, get_client_name(cptr, HIDEME));
	(void) exit_client(bcptr, bcptr, &me, "Server Exists");
    }

    /* The following if statement would be nice to remove since user
     * nicks never have '.' in them and servers must always have '.' in
     * them. There should never be a server/nick name collision, but it
     * is possible a capricious server admin could deliberately do
     * something strange. -Dianora */

    if ((acptr = find_client(host, NULL)) && acptr != cptr) 
    {
	/* Server trying to use the same name as a person. Would 
	 * cause a fair bit of confusion. Enough to make it hellish for a while and servers to send stuff to the wrong place. */

	sendto_one(cptr, "ERROR :Nickname %s already exists!", host);
	strcpy(nbuf, get_client_name(cptr, HIDEME));
	sendto_gnotice("from %s: Link %s cancelled, servername/nick collision", me.name, nbuf);
	sendto_serv_butone(cptr, ":%s GNOTICE :Link %s cancelled, servername/nick collision", me.name, nbuf);
	return exit_client(cptr, cptr, cptr, "Nick as Server");
    }
    if (IsServer(cptr)) 
    {
	/* Server is informing about a new server behind this link.
	 * Create REMOTE server structure, add it to list and propagate word to my other server links... */

	if (parc == 1 || info[0] == '\0') 
	{
	    sendto_one(cptr, "ERROR :No server info specified for %s", host);
	    return 0;
	}

	/* See if the newly found server is behind a guaranteed leaf
	 * (L-line). If so, close the link. Depreciated. Kinda redundant with Hlines. -epi */

	if (!(aconf = find_conf_host(cptr->confs, host, CONF_HUB)) || (aconf->port && (hop > aconf->port))) 
	{
	    sendto_gnotice("from %s: Non-Hub link %s introduced %s(%s).", me.name, get_client_name(cptr, HIDEME), host,
		aconf ? (aconf->host ? aconf->host : "*") : "!");
	    sendto_serv_butone(cptr,":%s GNOTICE :Non-Hub link %s introduced %s(%s).", me.name, 
		get_client_name(cptr, HIDEME), host, aconf ? (aconf->host ? aconf->host : "*") : "!");
	    sendto_one(cptr, "ERROR :%s has no H:Line for %s.", get_client_name(cptr, HIDEME), host);
	    return exit_client(cptr, cptr, cptr, "Too many servers");
	}

	acptr = make_client(cptr, sptr);
	(void) make_server(acptr);
	acptr->hopcount = hop;
	strncpyzt(acptr->name, host, sizeof(acptr->name));
	strncpyzt(acptr->info, info, REALLEN);
	acptr->serv->up = find_or_add(parv[0]);

	SetServer(acptr);

	/* if this server is behind a U-lined server, make it U-lined as well. - lucas */

	if (IsULine(sptr) || (find_uline(cptr->confs, acptr->name))) 
	{
	    acptr->flags |= FLAGS_ULINE;
	    sendto_umode(MODE_s, "%s introducing U:Lined server %s", cptr->name, acptr->name);
	}

	Count.server++;
	add_client_to_list(acptr);
	(void) add_to_client_hash_table(acptr->name, acptr);

	/* Old sendto_serv_but_one() call removed because we now need to send different names to different servers (domain name matching) */

	for (i = 0; i <= highest_fd; i++) 
	{
	    if (!(bcptr = local[i]) || !IsServer(bcptr) || bcptr == cptr || IsMe(bcptr))
		continue;
	    if (!(aconf = bcptr->serv->nline)) 
	    {
		sendto_gnotice("from %s: Lost N:Line for %s on %s. Closing", me.name, get_client_name(cptr, HIDEME), host);
		sendto_serv_butone(cptr, ":%s GNOTICE :Lost N:Line for %s on %s. Closing", me.name, get_client_name(cptr, HIDEME), host);
		return exit_client(cptr, cptr, cptr, "Lost N:Line");
	    }
	    if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
		continue;
	    sendto_one(bcptr, ":%s SERVER %s %d :%s", parv[0], acptr->name, hop + 1, acptr->info);
	}
	return 0;
    }
    if (!IsUnknown(cptr) && !IsHandshake(cptr))
	return 0;

    /* A local link that is still in undefined state wants to be a
     * SERVER. Check if this is allowed and change status accordingly... */
    /* Reject a direct nonTS server connection if we're TS_ONLY -orabidoo */

    strncpyzt(cptr->name, host, sizeof(cptr->name));
    strncpyzt(cptr->info, info[0] ? info : me.name, REALLEN);
    cptr->hopcount = hop;

    switch (check_server_init(cptr)) 
    {
    case 0:
	return m_server_estab(cptr);
    case 1:
	sendto_umode(MODE_s, "Access check for %s in progress", get_client_name(cptr, HIDEME));
	return 1;
    default:
	ircstp->is_ref++;
	sendto_umode(MODE_s, "Received unauthorized connection from %s.", get_client_host(cptr));
	return exit_client(cptr, cptr, cptr, "No C/N conf lines");
    }
}

static void sendnick_TS(aClient *cptr, aClient *acptr)
{
    static char ubuf[12];

    if (IsPerson(acptr)) 
    {
	send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);

	if (!*ubuf)
	{
	    ubuf[0] = '+';
	    ubuf[1] = '\0';
	}
	if (IsNICKIP(cptr)) 
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu %lu :%s", acptr->name, acptr->hopcount + 1, acptr->tsinfo, ubuf,
		acptr->user->username, acptr->user->host, acptr->user->server, acptr->user->servicestamp, htonl(acptr->ip.s_addr), acptr->info);
	else 
	    sendto_one(cptr, "NICK %s %d %ld %s %s %s %s %lu :%s", acptr->name, acptr->hopcount + 1, acptr->tsinfo, ubuf,
		acptr->user->username, acptr->user->host, acptr->user->server, acptr->user->servicestamp, acptr->info);
	if (HasUmodex(acptr))
	    sendto_one(cptr, ":%s VS %s :%s", me.name, acptr->name, acptr->user->virthost);
    }
}

int do_server_estab(aClient *cptr)
{
    aClient *acptr;
    aConfItem *aconf;
    aChannel *chptr;
    int i;
    /* "refresh" inpath with host */
    char *inpath = cptr->name;

    SetServer(cptr);
    Count.server++;
    Count.myserver++;

    if (IsZipCapable(cptr) && DoZipThis(cptr))
    {
	sendto_one(cptr, "SVINFO ZIP");
	SetZipOut(cptr);
	cptr->serv->zip_out = zip_create_output_session();
    }
#ifdef MAXBUFFERS
    /* let's try to bump up server sock_opts... -Taner */
    reset_sock_opts(cptr->fd, 1);
#endif 
    /* adds to fdlist */
    addto_fdlist(cptr->fd, &serv_fdlist);
#ifndef NO_PRIORITY
    /* this causes the server to be marked as "busy" */
    check_fdlists();
#endif
    cptr->pingval = get_client_ping(cptr);
    cptr->sendqlen = get_sendq(cptr);

    /* error, error, error! if a server is U:'d, and it connects to us, we need to figure that out! So, do it here. - lucas */ 

    if ((find_uline(cptr->confs, cptr->name)))
    {
	Count.myulined++;
	cptr->flags |= FLAGS_ULINE; 
    }
    sendto_gnotice("from %s: Link with %s established (capabilities:%s%s%s%s%s%s%s%s)", me.name, inpath,
	ZipOut(cptr) ? " Compressed" : "", RC4EncLink(cptr) ? " Encrypted" : "", IsULine(cptr) ? " ULined" : "", IsNICKIP(cptr) ? " NickIP" : "",
	    IsSSJoin(cptr) ? " SSJoin" : "", DoesTS(cptr) ? " TS" : " Non-TS", IsTS3(cptr) ? "3" : "", IsTS5(cptr) ? "5" : "");

    /* Notify everyone of the fact that this has just linked: the entire
     * network should get two of these, one explaining the link between
     * me->serv and the other between serv->me */

    sendto_serv_butone(NULL, ":%s GNOTICE :Link with %s established (capabilities: %s)", me.name, inpath, DoesTS(cptr) ? "TS" : "Non-TS");

    (void) add_to_client_hash_table(cptr->name, cptr);

    /* add it to scache */

    (void) find_or_add(cptr->name);

    /* Old sendto_serv_but_one() call removed because we now need to
     * send different names to different servers (domain name
     * matching) Send new server to other servers. */

    for (i = 0; i <= highest_fd; i++) 
    {
	if (!(acptr = local[i]) || !IsServer(acptr) || acptr == cptr || IsMe(acptr))
	    continue;
	if ((aconf = acptr->serv->nline) && !match(my_name_for_link(me.name, aconf), cptr->name))
	    continue;
	sendto_one(acptr, ":%s SERVER %s 2 :%s", me.name, cptr->name, cptr->info);
    }

    /* Pass on my client information to the new server
     * 
     * First, pass only servers (idea is that if the link gets 
     * cancelled beacause the server was already there, there are no
     * NICK's to be cancelled...). Of course, if cancellation occurs,
     * all this info is sent anyway, and I guess the link dies when a
     * read is attempted...? --msa 
     * 
     * Note: Link cancellation to occur at this point means that at
     * least two servers from my fragment are building up connection
     * this other fragment at the same time, it's a race condition,
     * not the normal way of operation... 
     * 
     * ALSO NOTE: using the get_client_name for server names-- see
     * previous *WARNING*!!! (Also, original inpath is
     * destroyed...) */

    aconf = cptr->serv->nline;

    for (acptr = &me; acptr; acptr = acptr->prev) 
    {
	if (acptr->from == cptr)
	    continue;
	if (IsServer(acptr)) 
	{
	    if (match(my_name_for_link(me.name, aconf), acptr->name) == 0)
		continue;
	    sendto_one(cptr, ":%s SERVER %s %d :%s", acptr->serv->up, acptr->name, acptr->hopcount + 1, acptr->info);
	}
    }

    /* send out our SQLINES and SGLINES too */
    for (aconf=conf ; aconf ; aconf=aconf->next) 
    {
	if ((aconf->status & CONF_QUARANTINE) && (aconf->status & CONF_SQLINE)) 
	    sendto_one(cptr, ":%s SQLINE %s :%s", me.name, aconf->name, aconf->passwd);
	if ((aconf->status & CONF_GCOS) && (aconf->status & CONF_SGLINE))
	    sendto_one(cptr, ":%s SGLINE %d :%s:%s", me.name, strlen(aconf->name), aconf->name, aconf->passwd);
    }

    send_blines(cptr);
    send_glines(cptr);
    send_slines(cptr);
    send_szlines(cptr);

    /* Bursts are about to start.. send a BURST */
    if (IsBurst(cptr))
	sendto_one(cptr, "BURST"); 

    /* Send it in the shortened format with the TS, if it's a TS
     * server; walk the list of channels, sending all the nicks that
     * haven't been sent yet for each channel, then send the channel
     * itself -- it's less obvious than sending all nicks first, but
     * on the receiving side memory will be allocated more nicely 
     * saving a few seconds in the handling of a split -orabidoo */

    {
	chanMember *cm;
	static char nickissent = 1;

	nickissent = 3 - nickissent;

	/* flag used for each nick to check if we've sent it yet - must
	 * be different each time and !=0, so we alternate between 1 and
	 * 2 -orabidoo */

	for (chptr = channel; chptr; chptr = chptr->nextch) 
	{
	    for (cm = chptr->members; cm; cm = cm->next) 
	    {
		acptr = cm->cptr;

		if (acptr->nicksent != nickissent) 
		{
		    acptr->nicksent = nickissent;

		    if (acptr->from != cptr)
			sendnick_TS(cptr, acptr);
		}
	    }
	    send_channel_modes(cptr, chptr);
	}
	/* also send out those that are not on any channel */
	for (acptr = &me; acptr; acptr = acptr->prev)
	    if (acptr->nicksent != nickissent) 
	    {
		acptr->nicksent = nickissent;
		if (acptr->from != cptr)
		    sendnick_TS(cptr, acptr);
	    }
    }
    if (ZipOut(cptr))
    {
	unsigned long inb, outb;
	double rat;
	zip_out_get_stats(cptr->serv->zip_out, &inb, &outb, &rat);

	if (inb)
	{
	    sendto_gnotice("from %s: Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
		me.name, cptr->name, inb, outb, rat);
	    sendto_serv_butone(cptr, ":%s GNOTICE :Connect burst to %s: %lu bytes normal, %lu compressed (%3.2f%%)",
		me.name, cptr->name, inb, outb, rat);
	}
    }

    /* stuff a PING at the end of this burst so we can figure out when the other side has finished processing it. */
    cptr->flags |= FLAGS_BURST|FLAGS_PINGSENT;

    if (IsBurst(cptr))
	cptr->flags |= FLAGS_SOBSENT;
    sendto_one(cptr, "PING :%s", me.name);
    return 0;
}

int m_server_estab(aClient *cptr)
{
    aConfItem *aconf, *bconf;
    char *inpath, *host, *s, *encr;
    int split;
    int i;

    inpath = cptr->name; /* "refresh" inpath with host  */
    split = mycmp(cptr->name, cptr->sockhost);
    host = cptr->name;

    if (!(aconf = find_conf(cptr->confs, host, CONF_NOCONNECT_SERVER))) 
    {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :Access denied. No N line for server %s", inpath);
	sendto_umode(MODE_s, "Access denied. No N:Line for server %s", inpath);
	return exit_client(cptr, cptr, cptr, "No N line for server");
    }
    if (!(bconf = find_conf(cptr->confs, host, CONF_CONNECT_SERVER))) 
    {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :Only N (no C) field for server %s", inpath);
	sendto_umode(MODE_s, "Only N (no C) field for server %s", inpath);
	return exit_client(cptr, cptr, cptr, "No C line for server");
    }

    encr = cptr->passwd;

    if (*aconf->passwd && !StrEq(aconf->passwd, encr)) 
    {
	ircstp->is_ref++;
	sendto_one(cptr, "ERROR :No Access (passwd mismatch) %s", inpath);
	sendto_umode(MODE_s, "Access denied (passwd mismatch) %s", inpath);
	return exit_client(cptr, cptr, cptr, "Bad Password");
    }
    memset(cptr->passwd, '\0', sizeof(cptr->passwd));

    if (DYN_HUB_STATE == 0)
    {
	for (i = 0; i <= highest_fd; i++)
	    if (local[i] && IsServer(local[i])) 
	    {
		ircstp->is_ref++;
		sendto_one(cptr, "ERROR :I'm a leaf not a hub");
		return exit_client(cptr, cptr, cptr, "I'm a leaf");
	    }
    }
    /* aconf->port is a CAPAB field, kind-of. kludge. mm, mm. */
    cptr->capabilities |= aconf->port;

    if (IsUnknown(cptr)) 
    {
	if (bconf->passwd[0])
	    sendto_one(cptr, "PASS %s :TS", bconf->passwd);

	/* Pass my info to the new server */

#ifdef HAVE_ENCRYPTION_ON
	if (!WantDKEY(cptr))
	    sendto_one(cptr, "CAPAB TS5 NOQUIT SSJOIN BURST UNCONNECT ZIP NICKIP");
	else
	    sendto_one(cptr, "CAPAB TS5 NOQUIT SSJOIN BURST UNCONNECT DKEY ZIP NICKIP");
#else
	sendto_one(cptr, "CAPAB TS5 NOQUIT SSJOIN BURST UNCONNECT ZIP NICKIP");
#endif
	sendto_one(cptr, "SERVER %s 1 :%s", my_name_for_link(me.name, aconf), (me.info[0]) ? (me.info) : "IRCers United");
    }
    else
    {
	s = (char *) strchr(aconf->host, '@');
	*s = '\0'; /* should never be NULL -- wanna bet? -Dianora */
#ifdef DEBUGMODE
	ircd_log("Check Usernames [%s]vs[%s]", aconf->host, cptr->username);
#endif
	if (match(aconf->host, cptr->username)) 
	{
	    *s = '@';
	    ircstp->is_ref++;
	    sendto_umode(MODE_s, "Username mismatch [%s]v[%s] : %s", aconf->host, cptr->username, get_client_name(cptr, HIDEME));
	    sendto_one(cptr, "ERROR :No Username Match");
	    return exit_client(cptr, cptr, cptr, "Bad User");
	}
	*s = '@';
    }
    sendto_one(cptr, "TCTRL %s %s 0 0 0 0 0 0 0 0 0 0 0 0 0", BASENAME, DYN_NETWORK_NETNAME);

    /* send routing notice, this should never happen anymore */
    if (!DoesTS(cptr)) 
    {
	sendto_gnotice("from %s: Warning: %s linked, non-TS server", me.name, cptr->name);
	sendto_serv_butone(cptr, ":%s GNOTICE :Warning: %s linked, non-TS server", me.name, cptr->name);
    }
    sendto_one(cptr, "SVINFO %d %d 0 :%ld", TS_CURRENT, TS_MIN, (ts_val) timeofday);

    /* sendto one(cptr, "CAPAB ...."); moved to after PASS but before SERVER
     * now in two places.. up above and in s_bsd.c. - lucas
     * This is to make sure we pass on our capabilities before we establish
     * a server connection */

    det_confs_butmask(cptr, CONF_HUB | CONF_NOCONNECT_SERVER | CONF_ULINE);

    /* WARNING 
     * In the following code in place of plain
     * server's name we send what is returned by
     * get_client_name which may add the "sockhost" after the name.
     * It's *very* *important* that there is a SPACE between 
     * the name and sockhost (if present). The receiving server
     * will start the information field from this first blank and
     * thus puts the sockhost into info. ...a bit tricky, but
     * you have been warned, besides code is more neat this way...
     * --msa */

    /* doesnt duplicate cptr->serv if allocated this struct already */

    (void) make_server(cptr);
    cptr->serv->up = me.name;
    cptr->serv->nline = aconf;

    /* now fill out the servers info so nobody knows dink about it. */
    memset((char *)&cptr->ip, '\0', sizeof(struct in_addr));
    strcpy(cptr->hostip, "127.0.0.1");
    strcpy(cptr->sockhost, "localhost");

#ifdef HAVE_ENCRYPTION_ON
    if (!CanDoDKEY(cptr) || !WantDKEY(cptr))
	return do_server_estab(cptr);
    else
    {
	SetNegoServer(cptr); /* VERY IMPORTANT THAT THIS IS HERE */
	sendto_one(cptr, "DKEY START");
    }
#else
    return do_server_estab(cptr);
#endif
    return 0;
}

/* 
 * m_burst()
 *  parv[0] = sender prefix
 *  parv[1] = SendQ if an EOB
 */
int m_burst(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    if (!IsServer(sptr) || sptr != cptr || parc > 2 || !IsBurst(sptr))
	return 0;
    if (parc == 2)
    {
	/* This is an EOB */
	sptr->flags &= ~(FLAGS_EOBRECV);

	if (sptr->flags & (FLAGS_SOBSENT|FLAGS_BURST))
	    return 0;
	/* we've already sent our EOB.. we synched first no need to check IsBurst because we shouldn't receive a BURST if they're not BURST capab */

	if (DYN_HTMLOCKON_BURST == 1)
	    HTMLOCK = NO;
	sendto_gnotice("from %s: synch to %s in %d %s at %s sendq", me.name,
	    *parv, (timeofday-sptr->firsttime), (timeofday-sptr->firsttime)==1?"sec":"secs", parv[1]);
	sendto_serv_butone(NULL, ":%s GNOTICE :synch to %s in %d %s at %s sendq",
	    me.name, sptr->name, (timeofday-sptr->firsttime), (timeofday-sptr->firsttime)==1?"sec":"secs", parv[1]);
    }
    else
    {
	/* SOB.. lock HTM if defined by admin */
	if (DYN_HTMLOCKON_BURST == 1)
	    HTMLOCK = YES;
	sptr->flags |= FLAGS_EOBRECV;
    }
    return 0;
}

/*
 * m_info()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_info(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char **text = infotext;
    static time_t last_used = 0L;

    if (hunt_server(cptr,sptr,":%s INFO :%s",1,parc,parv) == HUNTED_ISME) 
    {
	if (!IsAnOper(sptr))
	{
	    if (!MyConnect(sptr))
		return 0;
	    if ((last_used + MOTD_WAIT) > NOW) 
		return 0;
	    else 
		last_used = NOW;
	}
	sendto_umode(MODE_e, "/info requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	    HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);

	while (*text)
	    sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], *text++);
	sendto_one(sptr, rpl_str(RPL_INFO), me.name, parv[0], "");
	sendto_one(sptr, ":%s %d %s :Birth Date: %s, compile # %s", me.name, RPL_INFO, parv[0], creation, generation);
	sendto_one(sptr, ":%s %d %s :On-line since %s", me.name, RPL_INFO, parv[0], myctime(me.firsttime));
	sendto_one(sptr, rpl_str(RPL_ENDOFINFO), me.name, parv[0]);
    }
    return 0;
}

/*
 * m_links()
 *  parv[0] = sender prefix 
 *  parv[1] = servername mask 
 * or 
 *  parv[0] = sender prefix 
 *  parv[1] = server to query 
 *  parv[2] = servername mask
 */
int m_links(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *mask;
    aClient *acptr;
    char clean_mask[(2 * HOSTLEN) + 1];
    char *s;
    char *d;
    int n;

    if (parc > 2) 
    {
	if (hunt_server(cptr, sptr, ":%s LINKS %s :%s", 1, parc, parv) != HUNTED_ISME)
	    return 0;
	mask = parv[2];
	
    }
    else
	mask = parc < 2 ? NULL : parv[1];

    /* reject non-local requests */
    if (!IsAnOper(sptr) && !MyConnect(sptr))
	return 0;

    /* *sigh* Before the kiddies find this new and exciting way of annoying opers, lets clean up what is sent to all opers -Dianora */

    if (mask) 
    { /* only necessary if there is a mask */
	s = mask;
	d = clean_mask;
	n = (2 * HOSTLEN) - 2;
	while (*s && n > 0) 
	{
	    /* Is it a control character? */
	    if ((unsigned char) *s < (unsigned char) ' ') 
	    {
		*d++ = '^';
		/* turn it into a printable */
		*d++ = (char) ((unsigned char)*s + 0x40); 
		s++;
		n -= 2;
	    }
	    else if ((unsigned char) *s > (unsigned char) '~') 
	    {
		*d++ = '.';
		s++;
		n--;
	    }
	    else 
	    {
		*d++ = *s++;
		n--;
	    }
	}
	*d = '\0';
    }
    if (MyConnect(sptr))
	sendto_umode(MODE_e, "/links %s requested by %s (%s@%s) [%s]", mask ? clean_mask : "all", sptr->name, sptr->user->username,
	    HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);

    for (acptr = client, (void) collapse(mask); acptr; acptr = acptr->next) 
    {
	if (!IsServer(acptr) && !IsMe(acptr))
	    continue;
	if (!BadPtr(mask) && match(mask, acptr->name))
	    continue;
	if (IsULine(acptr) && DYN_HIDE_ULINES && !IsAnOper(sptr))
	    continue;
	sendto_one(sptr, rpl_str(RPL_LINKS), me.name, parv[0], acptr->name, acptr->serv->up,
	    acptr->hopcount, (acptr->info[0] ? acptr->info : "(Unknown Location)"));
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFLINKS), me.name, parv[0], BadPtr(mask) ? "*" : clean_mask);
    return 0;
}

static int report_array[12][3] =
{
    {CONF_CONNECT_SERVER,	RPL_STATSCLINE, 'C'},
    {CONF_NOCONNECT_SERVER,	RPL_STATSNLINE, 'N'},
    {CONF_GCOS,			RPL_STATSGLINE, 'G'},
    {CONF_CLIENT,		RPL_STATSILINE, 'I'},
    {CONF_KILL,			RPL_STATSKLINE, 'K'},
    {CONF_OPERATOR,		RPL_STATSOLINE, 'O'},
    {CONF_HUB,			RPL_STATSHLINE, 'H'},
    {CONF_LOCOP,		RPL_STATSOLINE, 'o'},
    {CONF_QUARANTINE,		RPL_STATSQLINE, 'Q'},
    {CONF_ULINE,		RPL_STATSULINE, 'U'},
    {CONF_VHOST,		RPL_STATSVLINE, 'V'},
    {0,0}
};

static void report_configured_links(aClient *sptr, int mask)
{
    static char null[] = "<NULL>";
    aConfItem *tmp;
    int *p, port;
    char c, *host, *pass, *name;
	
    for (tmp = conf; tmp; tmp = tmp->next)
	if (tmp->status & mask) 
	{
	    for (p = &report_array[0][0]; *p; p += 3)
		if (*p&tmp->status)
		    break;
	    if (!*p)
		continue;
	    c = (char) *(p + 2);
	    host = BadPtr(tmp->host) ? null : tmp->host;
	    pass = BadPtr(tmp->passwd) ? null : tmp->passwd;
	    name = BadPtr(tmp->name) ? null : tmp->name;
	    port = (int) tmp->port;

	    /* block out IP addresses of U: lined servers in 
	     * a C/N request - lucas/xpsycho
	     * modified on request by taz, hide all
	     * ips to other than routing staff or admins
	     * -epi
	     * yet again, make this just hide _all_ ips. how sad.
	     * - lucas */

	    if (tmp->status & CONF_SERVER_MASK)
		host = "*";
	    if (tmp->status == CONF_KILL)
		sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, name, pass);
	    else if (tmp->status & CONF_QUARANTINE)
		sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, pass, name, port, get_conf_class(tmp));
	    else if (tmp->status & CONF_GCOS)
		sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, pass, name, port, get_conf_class(tmp));
	    else if (mask & CONF_OPS)
		sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, name, oflagstr(port), get_conf_class(tmp));
	    else
	    {
		if (!IsAnOper(sptr) && !IsULine(sptr))
		    sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, "*", name, port, get_conf_class(tmp));
		else
		    sendto_one(sptr, rpl_str(p[1]), me.name, sptr->name, c, host, name, port, get_conf_class(tmp));
	    }
	}
	return;
}

/*
 * m_stats() 
 *  parv[0] = sender prefix 
 *  parv[1] = statistics selector (defaults to Message frequency) 
 *  parv[2] = server name (current server defaulted, if omitted) 
 */
/* 
 * m_stats/stats_conf 
 *  Report N/C-configuration lines from this server. This could 
 *  report other configuration lines too, but converting the 
 *  status back to "char" is a bit akward--not worth the code 
 *  it needs...
 * 
 *  Note:   The info is reported in the order the server uses 
 *  it--not reversed as in ircd.conf!
 */
int m_stats(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static char Lformat[] = ":%s %d %s %s %u %u %u %u %u :%u %u %s";
    static char Sformat[] = ":%s %d %s Name SendQ SendM SendBytes RcveM RcveBytes :Open_since Idle TS";
    struct Message *mptr;
    aClient *acptr;
    char stat = parc > 1 ? parv[1][0] : '\0';
    int i;
    int doall = 0, wilds = 0;
    char *name;
    time_t sincetime;
    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s STATS %s :%s", 2, parc, parv) != HUNTED_ISME)
	return 0;
    if (!IsAnOper(sptr) && !IsULine(sptr)) 
    {
	if ((last_used + MOTD_WAIT) > NOW)
	    return 0;
	else
	    last_used = NOW;
    }
    if (parc > 2) 
    {
	name = parv[2];
	if (!mycmp(name, me.name))
	    doall = 2;
	else if (match(name, me.name) == 0)
	    doall = 1;
	if (strchr(name, '*') || strchr(name, '?'))
	    wilds = 1;
    }
    else
	name = me.name;

    if (stat == (char) 0 && !IsULine(sptr))
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "STATS");
	return 0;
    }

    switch (stat) 
    {
    case 'L':
    case 'l':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else

	/* changed behavior totally.  This is what we do now:
	 * #1: if the user is not opered, never return ips for anything
	 * #2: we DON'T deny /stats l for opers.  Ever heard of /sping?
	 *     it's easy to see if you're lagging a server, why restrict
	 *     something used 99% of the time for good when you're not
	 *     doing any harm?
	 * #3: NEVER return all users on a server, UGH, just like
	 *     /trace, this was fiercely obnoxious.  If you don't
	 *     add an argument, you get all SERVER links. */

	    sendto_one(sptr, Sformat, me.name, RPL_STATSLINKINFO, parv[0]);
	if ((parc > 2) && !(doall || wilds))
	{         /* Single client lookup */
	    if (!(acptr = find_person(name, NULL)))
		break;

	    /* sincetime might be greater than timeofday,
	     * store a new value here to avoid sending 
	     * negative since-times. -Rak */

	    sincetime = (acptr->since > timeofday) ? 0 : timeofday - acptr->since;
	    sendto_one(sptr, Lformat, me.name,
		       RPL_STATSLINKINFO, parv[0],
		       (IsAnOper(sptr) ? get_client_name(acptr, TRUE) :
			get_client_name(acptr, HIDEME)),
		       (int) DBufLength(&acptr->sendQ),
		       (int) acptr->sendM, (int) acptr->sendK,
		       (int) acptr->receiveM, (int) acptr->receiveK,
		       timeofday - acptr->firsttime,
		       sincetime, IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS") : "-");
	}
	else
	{
	    for (i = 0; i <= highest_fd; i++) 
	    {
		if (!(acptr = local[i]))
		    continue;
		if (!IsServer(acptr))
		    continue; /* nothing but servers */
		if (IsULine(acptr) && !IsAnOper(sptr))
		    continue;

		sincetime = (acptr->since > timeofday) ? 0 : timeofday - acptr->since;
		sendto_one(sptr, Lformat, me.name,
			   RPL_STATSLINKINFO, parv[0],
			   get_client_name(acptr, HIDEME),
			   (int) DBufLength(&acptr->sendQ),
			   (int) acptr->sendM, (int) acptr->sendK,
			   (int) acptr->receiveM, (int) acptr->receiveK,
			   timeofday - acptr->firsttime,
			   sincetime, IsServer(acptr) ? (DoesTS(acptr) ? "TS" : "NoTS") : "-");
	    }
	}
	break;
    case 'b':
    case 'B':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_blines(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'C':
    case 'c':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_CONNECT_SERVER | CONF_NOCONNECT_SERVER);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'D':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_conf_links(sptr, &ZList1, RPL_STATSZLINE, 'Z');
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'd':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_szlines(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'E':
    case 'e':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	{
	    report_conf_links(sptr, &EList1, RPL_STATSELINE, 'E');
	    report_conf_links(sptr, &EList2, RPL_STATSELINE, 'E');
	    report_conf_links(sptr, &EList3, RPL_STATSELINE, 'E');
	}
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;	
    case 'F':
    case 'f':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	{
	    report_conf_links(sptr, &FList1, RPL_STATSFLINE, 'F');
	    report_conf_links(sptr, &FList2, RPL_STATSFLINE, 'F');
	    report_conf_links(sptr, &FList3, RPL_STATSFLINE, 'F');
	}
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;	
    case 'G':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    list_glines(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'g':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_GCOS);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'H':
    case 'h':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_HUB);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'I':
    case 'i':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_CLIENT);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;	
    case 'k':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_temp_klines(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'K':
	if (parc > 3)
	    report_matching_host_klines(sptr, parv[3]);
	else if (IsAnOper(sptr))
	{
	    report_conf_links(sptr, &KList1, RPL_STATSKLINE, 'K');
	    report_conf_links(sptr, &KList2, RPL_STATSKLINE, 'K');
	    report_conf_links(sptr, &KList3, RPL_STATSKLINE, 'K');
	}
	else if (sptr->user)
	    report_matching_host_klines(sptr, sptr->user->host);
	break;
    case 'M':
    case 'm':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	/* original behaviour was not to report the command, if
	 * the command hadn't been used. I'm going to always
	 * report the command instead -Dianora */
	    for (mptr = msgtab; mptr->cmd; mptr++)
		sendto_one(sptr, rpl_str(RPL_STATSCOMMANDS), me.name, parv[0], mptr->cmd, mptr->count, mptr->bytes);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'N':
    case 'n':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	{
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0], "User Connects Today: ", Count.today);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0], "User Connects past week: ", Count.weekly);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0], "User Connects past month: ", Count.monthly);
	    sendto_one(sptr, rpl_str(RPL_STATSCOUNT), me.name, parv[0], "User Connects past year: ", Count.yearly);
	    sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
		stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	}
	break;
    case 'o':
    case 'O':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_OPS);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'p':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    show_opers(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'Q':
    case 'q':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_QUARANTINE);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'R':
    case 'r':
#ifdef DEBUGMODE
	send_usage(sptr, parv[0]);
#endif
	break;
    case 'S':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	else
	    list_slines(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 's':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	else
	    list_scache(cptr, sptr, parc, parv);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'T':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	else
	    throttle_stats(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 't':
	if (!IsAnOper(sptr)) 
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	else
	    tstats(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'U':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_ULINE);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'u':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	{
	    time_t now;
	    now = timeofday - me.since;
	    sendto_one(sptr, rpl_str(RPL_STATSUPTIME), me.name, parv[0], now / 86400, (now / 3600) % 24, (now / 60) % 60, now % 60);
#ifdef HIGHEST_CONNECTION
	    sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0], max_connection_count, max_client_count, Count.totalrestartcount);
#endif
	    sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
		stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	    break;
	}
    case 'v':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    show_servers(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'V':
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_configured_links(sptr, CONF_VHOST);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'Y':
    case 'y':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    report_classes(sptr);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case 'z':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	else
	    count_memory(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
    case '?':
	if (!HasUmodeo(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name,  parv[0]);
	    return 0;
	}
	else
	    serv_info(sptr, parv[0]);
	sendto_umode(MODE_e, "/stats %c requested by %s (%s@%s) [%s]",
	    stat, sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	break;
     default:
	stat = '*';
	break;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
    return 0;
}
/*
 * m_users()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_users(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (hunt_server(cptr, sptr, ":%s USERS :%s", 1, parc, parv) == HUNTED_ISME) 
    {
	sendto_one(sptr, rpl_str(RPL_LOCALUSERS), me.name, parv[0], Count.local, Count.max_loc);
	sendto_one(sptr, rpl_str(RPL_GLOBALUSERS), me.name, parv[0], Count.total, Count.max_tot);
    }
    return 0;
}

/*
 * m_error() 
 *  parv[0] = sender prefix 
 *  parv[*] = parameters
 */
int m_error(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *para;

    para = (parc > 1 && *parv[1] != '\0') ? parv[1] : "<>";

    Debug((DEBUG_ERROR, "Received ERROR message from %s: %s", sptr->name, para));

    /* Ignore error messages generated by normal user clients 
     * (because ill-behaving user clients would flood opers screen
     * otherwise). Pass ERROR's from other sources to the local
     * operator... */

    if (IsPerson(cptr) || IsUnknown(cptr))
	return 0;
    if (cptr == sptr)
	sendto_umode(MODE_s, "ERROR :from %s -- %s", get_client_name(cptr, HIDEME), para);
    else
	sendto_umode(MODE_s, "ERROR :from %s via %s -- %s", sptr->name, get_client_name(cptr, HIDEME), para);
    return 0;
}

/*
 * m_help()
 *  parv[0] = sender prefix
 */
int m_help(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    aMotd *helpfile_ptr;
    static time_t last_used = 0L;

    /* reject non local requests */
    if ((last_used + MOTD_WAIT) > NOW)
	return 0;
    else
	last_used = NOW;
    if (!IsAnOper(sptr) || (helpfile == (aMotd *) NULL)) 
    {
	for (i = 0; msgtab[i].cmd; i++)
	    sendto_one(sptr, ":%s NOTICE %s :%s", me.name, parv[0], msgtab[i].cmd);
	return 0;
    }

    helpfile_ptr = helpfile;
    while (helpfile_ptr) 
    {
	sendto_one(sptr, ":%s %d %s :%s", me.name, RPL_SETTINGS, parv[0], helpfile_ptr->line);
	helpfile_ptr = helpfile_ptr->next;
    }
    return 0;
}

/*
 * parv[0] = sender parv[1] = host/server mask. 
 * parv[2] = server to query
 * 
 * 199970918 JRL hacked to ignore parv[1] completely and require parc > 3
 * to cause a force
 */
int m_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int send_lusers(aClient *, aClient *, int, char **);

    if (parc > 2) 
    {
	if (hunt_server(cptr, sptr, ":%s LUSERS %s :%s", 2, parc, parv) != HUNTED_ISME)
	    return 0;
    }		
    return send_lusers(cptr,sptr,parc,parv);
}

/*
 * send_lusers
 *  parv[0] = sender
 *  parv[1] = host/server mask.
 *  parv[2] = server to query
 */
int send_lusers(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
#define LUSERS_CACHE_TIME 180
    static long last_time=0;
    static int s_count = 0, c_count = 0, u_count = 0, i_count = 0;
    static int o_count = 0, m_client = 0, m_server = 0, m_ulined = 0;
    int forced;
    aClient *acptr;

    forced = (IsAnOper(sptr) && (parc > 3));

    Count.unknown = 0;
    m_server = Count.myserver;
    m_ulined = Count.myulined;
    m_client = Count.local;
    i_count = Count.invisi;
    u_count = Count.unknown;
    c_count = Count.total - Count.invisi;
    s_count = Count.server;
    o_count = Count.oper;

    if (forced || (timeofday > last_time + LUSERS_CACHE_TIME)) 
    {
	last_time = timeofday;
	/* only recount if more than a second has passed since last request use LUSERS_CACHE_TIME instead... */
	s_count = 0;
	c_count = 0;
	u_count = 0;
	i_count = 0;
	o_count = 0;
	m_client = 0;
	m_server = 0;
	m_ulined = 0;

	for (acptr = client; acptr; acptr = acptr->next) 
	{
	    switch (acptr->status) 
	    {
	    case STAT_SERVER:
		if (MyConnect(acptr))
		{
		    m_server++;

		    if (IsULine(acptr))
			m_ulined++;
		}
	    case STAT_ME:
		s_count++;
		break;
	    case STAT_CLIENT:
		if (IsAnOper(acptr) && !IsULine(acptr))
		    o_count++;
#ifdef SHOW_INVISIBLE_LUSERS
		if (MyConnect(acptr))
		    m_client++;
		if (!HasUmodei(acptr))
		    c_count++;
		else
		    i_count++;
#else
		if (MyConnect(acptr)) 
		{
		    if (HasUmodei(acptr)) 
		    {
			if (IsAnOper(sptr))
			    m_client++;
		    }
		    else
			m_client++;
		}
		if (!HasUmodei(acptr))
		    c_count++;
		else
		    i_count++;
#endif
		break;
	    default:
		u_count++;
		break;
	    }
	}

	/* We only want to reassign the global counts if the recount time
	 * has expired, and NOT when it was forced, since someone may
	 * supply a mask which will only count part of the userbase -Taner */

	if (!forced) 
	{
	    if (m_server != Count.myserver) 
	    {
		sendto_umode(MODE_d, "Local server count off by %d", Count.myserver - m_server);
		Count.myserver = m_server;
	    }
	    if (m_ulined != Count.myulined) 
	    {
		sendto_umode(MODE_d, "Local ulinedserver count off by %d", Count.myulined - m_ulined);
		Count.myulined = m_ulined;
	    }
	    if (s_count != Count.server) 
	    {
		sendto_umode(MODE_d, "Server count off by %d", Count.server - s_count);
		Count.server = s_count;
	    }
	    if (i_count != Count.invisi) 
	    {
		sendto_umode(MODE_d, "Invisible client count off by %d", Count.invisi - i_count);
		Count.invisi = i_count;
	    }
	    if ((c_count + i_count) != Count.total) 
	    {
		sendto_umode(MODE_d, "Total client count off by %d", Count.total - (c_count + i_count));
		Count.total = c_count + i_count;
	    }
	    if (m_client != Count.local) 
	    {
		sendto_umode(MODE_d, "Local client count off by %d", Count.local - m_client);
		Count.local = m_client;
	    }
	    if (o_count != Count.oper) 
	    {
		sendto_umode(MODE_d, "Oper count off by %d", Count.oper - o_count);
		Count.oper = o_count;
	    }
	    Count.unknown = u_count;
	} /* Complain & reset loop */
    } /* Recount loop */
#ifdef SAVE_MAXCLIENT_STATS
    /* save stats */
    if ((timeofday - last_stat_save) > SAVE_MAXCLIENT_STATS_TIME)
    {
	FILE *fp;

	last_stat_save = timeofday;
	fp=fopen(DPATH "/.maxclients", "w");

	if (fp!=NULL)
	{
	    fprintf(fp, "%d %d %li %li %li %ld %ld %ld %ld", Count.max_loc,
		Count.max_tot, Count.weekly, Count.monthly, Count.yearly, Count.start, Count.week, Count.month, Count.year);
	    fclose(fp);
	    sendto_umode(MODE_d, "Saved maxclient statistics");
	}
    }
#endif
#ifndef	SHOW_INVISIBLE_LUSERS
    if (IsAnOper(sptr) && i_count)
#endif
	sendto_one(sptr, rpl_str(RPL_LUSERCLIENT), me.name, parv[0], c_count, i_count, s_count);
#ifndef	SHOW_INVISIBLE_LUSERS
    else
	sendto_one(sptr, ":%s %d %s :There are %d users on %d servers", me.name, RPL_LUSERCLIENT, parv[0], c_count, s_count);
#endif
    if (o_count)
	sendto_one(sptr, rpl_str(RPL_LUSEROP), me.name, parv[0], o_count);
    if (u_count > 0)
	sendto_one(sptr, rpl_str(RPL_LUSERUNKNOWN), me.name, parv[0], u_count);
    
    /* This should be ok */
    if (Count.chan > 0)
	sendto_one(sptr, rpl_str(RPL_LUSERCHANNELS), me.name, parv[0], Count.chan);
    if (DYN_HIDE_ULINES == 0)
	sendto_one(sptr, rpl_str(RPL_LUSERME), me.name, parv[0], m_client, m_server);
    else
	sendto_one(sptr, rpl_str(RPL_LUSERME), me.name, parv[0], m_client, HasUmodeo(sptr) ? m_server : m_server - m_ulined);
    sendto_one(sptr, rpl_str(RPL_LOCALUSERS), me.name, parv[0], Count.local, Count.max_loc);
    sendto_one(sptr, rpl_str(RPL_GLOBALUSERS), me.name, parv[0], Count.total, Count.max_tot);
    sendto_one(sptr, rpl_str(RPL_STATSCONN), me.name, parv[0], max_connection_count, max_client_count, Count.totalrestartcount);

    if (m_client > max_client_count)
	max_client_count = m_client;
    if ((m_client + m_server) > max_connection_count) 
    {
	max_connection_count = m_client + m_server;
	if (max_connection_count % 10 == 0)
	    sendto_umode(MODE_s, "Highest connection count: %d (%d clients) (%d since server was (re)started)",
		max_connection_count, max_client_count, Count.totalrestartcount);
    }
    return 0;
}

/*
 * m_connect()
 *  parv[0] = sender prefix 
 *  parv[1] = servername 
 *  parv[2] = port number 
 *  parv[3] = remote server
 */
int m_connect(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int port, tmpport, retval;
    aConfItem *aconf;
    aClient *acptr;

    if (!IsPrivileged(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return -1;
    }
    if ((MyClient(sptr) && !OPCanGRoute(sptr) && parc > 3) || (MyClient(sptr) && !OPCanLRoute(sptr) && parc <= 3))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (hunt_server(cptr, sptr, ":%s CONNECT %s %s :%s", 3, parc, parv) != HUNTED_ISME)
	return 0;
    if (parc < 2 || *parv[1] == '\0') 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CONNECT");
	return -1;
    }
    if ((acptr = find_server(parv[1], NULL)))
    {
	sendto_one(sptr, ":%s NOTICE %s :Connect: Server %s %s %s.", me.name, parv[0], parv[1], "already exists from", acptr->from->name);
	return 0;
    }

    for (aconf = conf; aconf; aconf = aconf->next)
	if (aconf->status == CONF_CONNECT_SERVER && match(parv[1], aconf->name) == 0)
	    break;

    /* Checked first servernames, then try hostnames. */
    if (!aconf)
	for (aconf = conf; aconf; aconf = aconf->next)
	    if (aconf->status == CONF_CONNECT_SERVER && (match(parv[1], aconf->host) == 0 || match(parv[1], strchr(aconf->host, '@') + 1) == 0))
		break;
    
    if (!aconf) 
    {
	sendto_one(sptr, "NOTICE %s :Connect: No C line found for %s.", parv[0], parv[1]);
	return 0;
    }

    /* Get port number from user, if given. If not specified, use
     * the default form configuration structure. If missing from there, then use the precompiled default. */

    tmpport = port = aconf->port;

    if (parc > 2 && !BadPtr(parv[2])) 
    {
	if ((port = atoi(parv[2])) <= 0) 
	{
	    sendto_one(sptr, "NOTICE %s :Connect: Illegal port number", parv[0]);
	    return 0;
	}
    }
    else if (port <= 0 && (port = PORTNUM) <= 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Connect: missing port number", me.name, parv[0]);
	return 0;
    }

    /* Notify all operators about remote connect requests Let's notify about local connects, too. - lucas */

    sendto_gnotice("from %s: %s CONNECT %s %s from %s", me.name,  IsAnOper(cptr) ? "Local" : "Remote", 
	parv[1], parv[2] ? parv[2] : "", get_client_name(sptr, HIDEME));
    sendto_serv_butone(NULL, ":%s GNOTICE :%s CONNECT %s %s from %s", me.name, IsAnOper(cptr) ? "Local" : "Remote",
	parv[1], parv[2] ? parv[2] : "", get_client_name(sptr, HIDEME));

#if defined(USE_SYSLOG) && defined(SYSLOG_CONNECT)
    ircd_log("CONNECT From %s : %s %s", parv[0], parv[1], parv[2] ? parv[2] : "");
#endif
    
    aconf->port = port;
    switch (retval = connect_server(aconf, sptr, NULL))
    {
    case 0:
	sendto_one(sptr, ":%s NOTICE %s :*** Connecting to %s.", me.name, parv[0], aconf->name);
	break;
    case -1:
	sendto_one(sptr, ":%s NOTICE %s :*** Couldn't connect to %s.", me.name, parv[0], aconf->name);
	break;
    case -2:
	sendto_one(sptr, ":%s NOTICE %s :*** Host %s is unknown.", me.name, parv[0], aconf->name);
	break;
    default:
	sendto_one(sptr, ":%s NOTICE %s :*** Connection to %s failed: %s", me.name, parv[0], aconf->name, strerror(retval));
    }
    aconf->port = tmpport;
    return 0;
}

/*
 * m_wallops() 
 *  parv[0] = sender prefix 
 *  parv[1] = message text
 */
int m_wallops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;
    
    if (BadPtr(message)) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "WALLOPS");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr) && !OPCanWallOps(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return (0);
    }
    sendto_wallops_butone(IsServer(cptr) ? cptr : NULL, sptr, ":%s WALLOPS :%s", parv[0], message);
    return 0;
}

/*
 * m_locops()
 *  parv[0] = sender prefix 
 *  parv[1] = message text
 */
int m_locops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message)) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "LOCOPS");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr) && !OPCanLocOps(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return (0);
    }
    sendto_locops("from %s: %s", parv[0], message);
    return (0);
}

/*
 * m_helpop()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_helpop(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "HELPOP");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodeh(sptr) && !IsULine(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s HELPOP :%s", parv[0], message);
    send_helpop("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_goper()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_goper(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GOPER");
	return 0;
    }
    if (!IsServer(sptr) || !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    sendto_serv_butone(IsServer(cptr) ? cptr : NULL, ":%s GOPER :%s", parv[0], message);
    sendto_ops("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_snotice() 
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_snotice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;
    
    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SNOTICE");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s SNOTICE :%s", parv[0], message);
    sendto_snotice("%s", message);
    return 0;
}

/*
 * m_gnotice() 
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_gnotice(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;
    
    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GNOTICE");
	return 0;
    }
    if (!IsServer(sptr) && MyConnect(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s GNOTICE :%s", parv[0], message);
    sendto_gnotice("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_globops()
 *  parv[0] = sender prefix
 *  parv[1] = message text
 */
int m_globops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLOBOPS");
	return 0;
    }
    if (MyClient(sptr) && !OFlagg(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s GLOBOPS :%s", parv[0], message);
    send_globops("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_chatops()
 *  parv[0] = sender prefix
 *  parv[1] = message text
 */
int m_chatops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CHATOPS");
	return 0;
    }
    if (MyClient(sptr) && (!IsAnOper(sptr))) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s CHATOPS :%s", parv[0], message);
    send_chatops("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_adchat()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_adchat(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ADCHAT");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodeA(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s ADCHAT :%s", parv[0], message);
    send_adchat("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_sachat()
 *  parv[0] = sender prefix
 *  parv[1] = message text
 */
int m_sachat(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message))
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SACHAT");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodea(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s SACHAT :%s", parv[0], message);
    send_sachat("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_tachat()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_tachat(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "TACHAT");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodeT(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s TACHAT :%s", parv[0], message);
    send_tachat("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_nachat()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_nachat(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "NACHAT");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodeN(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s NACHAT :%s", parv[0], message);
    send_nachat("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_cochat()
 *  parv[0] = sender prefix 
 *  parv[1] = message text 
 */
int m_cochat(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "COCHAT");
	return 0;
    }
    if (MyClient(sptr) && !HasUmodeX(sptr) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s COCHAT :%s", parv[0], message);
    send_cochat("from %s: %s", parv[0], message);
    return 0;
}

/*
 * m_abuse()
 *  parv[0] = sender prefix
 *  parv[1] = message text
 */
int m_abuse(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (check_registered(sptr))
	return 0;
    if (BadPtr(message)) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ABUSE");
	return 0;
    }
    if (MyClient(sptr) && !(HasUmoder(sptr) || HasUmodeo(sptr))) 
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- This service is for registered user only.", me.name, parv[0]);	
	return 0;
    }
    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';
    sendto_serv_butone_services(IsServer(cptr) ? cptr : NULL, ":%s ABUSE :%s", parv[0], message);
    send_abuse("from %s: %s", parv[0], message);
    if (MyClient(sptr))
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Your abuse report has been forwarded to the Abuse Administrators.", me.name, parv[0]);
    return 0;
}


/*
 * m_time()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_time(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (hunt_server(cptr, sptr, ":%s TIME :%s", 1, parc, parv) == HUNTED_ISME)
	sendto_one(sptr, rpl_str(RPL_TIME), me.name, parv[0], me.name, date((long) 0));
    return 0;
}

/*
 * m_admin()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_admin(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aConfItem *aconf;
    
    if (hunt_server(cptr, sptr, ":%s ADMIN :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;
    if (IsPerson(sptr))
	sendto_umode(MODE_e, "/admin requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username, sptr->user->host, sptr->user->server);
    if ((aconf = find_admin())) 
    {
	sendto_one(sptr, rpl_str(RPL_ADMINME), me.name, parv[0], me.name);
	sendto_one(sptr, rpl_str(RPL_ADMINLOC1), me.name, parv[0], aconf->host ? aconf->host : "");
	sendto_one(sptr, rpl_str(RPL_ADMINLOC2), me.name, parv[0], aconf->passwd ? aconf->passwd : "");
	sendto_one(sptr, rpl_str(RPL_ADMINEMAIL), me.name, parv[0], aconf->name ? aconf->name : "");
    }
    else
	sendto_one(sptr, err_str(ERR_NOADMININFO), me.name, parv[0], me.name);
    return 0;
}

#ifdef FLUD
extern int flud_num;
extern int flud_time;
extern int flud_block;
#endif
extern int spam_num;
extern int spam_time;
#ifdef ANTI_DRONE_FLOOD
extern int dronetime;
extern int dronecount;
#endif
#ifdef IDLE_CHECK
extern int idle_time;
#endif
extern int server_split_recovery_time;

/*
 * m_set()
 */
int m_set(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *command;

    if (!MyClient(sptr) || !HasUmodeo(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc > 1) 
    {
	command = parv[1];

	if (!strncasecmp(command, "MAX", 3))
	{
	    if (parc > 2) 
	    {
		int new_value = atoi(parv[2]);

		if (new_value > MASTER_MAX) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :You cannot set MAXCLIENTS to > MASTER_MAX (%d)", me.name, parv[0], MASTER_MAX);
		    return 0;
		}
		if (new_value < 32) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :You cannot set MAXCLIENTS to < 32 (%d:%d)", me.name, parv[0], MAXCLIENTS, highest_fd);
		    return 0;
		}
		MAXCLIENTS = new_value;
		sendto_one(sptr, ":%s NOTICE %s :NEW MAXCLIENTS = %d (Current = %d)", me.name, parv[0], MAXCLIENTS, Count.local);
		sendto_umode(MODE_s, "%s (%s@%s) set new MAXCLIENTS to %d (%d current)", parv[0], sptr->user->username,
		    HasUmodex(sptr) ? sptr->user->virthost : sptr->sockhost, MAXCLIENTS, Count.local);
		return 0;
	    }
	    sendto_one(sptr, ":%s NOTICE %s :Current Maxclients = %d (%d)", me.name, parv[0], MAXCLIENTS, Count.local);
	    return 0;
	}
#ifdef FLUD
	else if (!strncasecmp(command, "FLUDNUM", 7)) 
	{
	    if (parc > 2) 
	    {
		int newval = atoi(parv[2]);

		if (newval <= 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :flud NUM must be > 0", me.name, parv[0]);
		    return 0;
		}
		flud_num = newval;
		sendto_umode(MODE_s, "%s has changed flud NUM to %i", parv[0], flud_num);
		sendto_one(sptr, ":%s NOTICE %s :flud NUM is now set to %i", me.name, parv[0], flud_num);
		return 0;
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :flud NUM is currently %i", me.name, parv[0], flud_num);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "FLUDTIME", 8)) 
	{
	    if (parc > 2) 
	    {
		int newval = atoi(parv[2]);

		if (newval <= 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :flud TIME must be > 0", me.name, parv[0]);
		    return 0;
		}
		flud_time = newval;
		sendto_umode(MODE_s, "%s has changed flud TIME to %i", parv[0], flud_time);
		sendto_one(sptr, ":%s NOTICE %s :flud TIME is now set to %i", me.name, parv[0], flud_time);
		return 0;
	    }
	    else 
	    {
		sendto_one(sptr, ":%s NOTICE %s :flud TIME is currently %i", me.name, parv[0], flud_time);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "FLUDBLOCK", 9)) 
	{
	    if (parc > 2)
	    {
		int newval = atoi(parv[2]);

		if (newval < 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :flud BLOCK must be >= 0", me.name, parv[0]);
		    return 0;
		}
		flud_block = newval;

		if (flud_block == 0)
		{
		    sendto_umode(MODE_s, "%s has disabled flud detection/protection", parv[0]);
		    sendto_one(sptr, ":%s NOTICE %s :flud detection disabled", me.name, parv[0]);
		}
		else
		{
		    sendto_umode(MODE_s, "%s has changed flud BLOCK to %i", parv[0], flud_block);
		    sendto_one(sptr, ":%s NOTICE %s :flud BLOCK is now set to %i", me.name, parv[0], flud_block);
		}
		return 0;
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :flud BLOCK is currently %i", me.name, parv[0], flud_block);
		return 0;
	    }
	}
#endif
	else if (!strncasecmp(command, "SPLITDELAY", 10)) 
	{
	    if (parc > 2) 
	    {
		int newval = atoi(parv[2]);

		if (newval < 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :split delay must be >= 0", me.name, parv[0]);
		    return 0;
		}
		if (newval > MAX_SERVER_SPLIT_RECOVERY_TIME)
		    newval = MAX_SERVER_SPLIT_RECOVERY_TIME;
		sendto_umode(MODE_s, "%s has changed spam SERVER SPLIT RECOVERY TIME to %i", parv[0], newval);
		sendto_one(sptr, ":%s NOTICE %s :SERVER SPLIT RECOVERY TIME is now set to %i", me.name, parv[0], newval);
		server_split_recovery_time = (newval * 60);
		return 0;
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :SERVER SPLIT RECOVERY TIME is currenly %i", me.name, parv[0],
		server_split_recovery_time / 60);
	    }
	}
	/* int spam_time = MIN_JOIN_LEAVE_TIME; 
	 * int spam_num = MAX_JOIN_LEAVE_COUNT; */
	else if (!strncasecmp(command, "SPAMNUM", 7)) 
	{
	    if (parc > 2) 
	    {
		int newval = atoi(parv[2]);

		if (newval <= 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :SPAMNUM must be > 0", me.name, parv[0]);
		    return 0;
		}
		if (newval < MIN_SPAM_NUM)
		    spam_num = MIN_SPAM_NUM;
		else
		    spam_num = newval;
		sendto_umode(MODE_s, "%s has changed SPAMNUM to %i", parv[0], spam_num);
		sendto_one(sptr, ":%s NOTICE %s :SPAMNUM is now set to %i", me.name, parv[0], spam_num);
		return 0;
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :SPAMNUM is currently %i", me.name, parv[0], spam_num);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "SPAMTIME", 8)) 
	{
	    if (parc > 2) 
	    {
		int newval = atoi(parv[2]);

		if (newval <= 0) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :SPAMTIME must be > 0", me.name, parv[0]);
		    return 0;
		}
		if (newval < MIN_SPAM_TIME)
		    spam_time = MIN_SPAM_TIME;
		else
		    spam_time = newval;
		sendto_umode(MODE_s, "%s has changed SPAMTIME to %i", parv[0], spam_time);
		sendto_one(sptr, ":%s NOTICE %s :SPAMTIME is now set to %i", me.name, parv[0], spam_time);
		return 0;
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :SPAMTIME is currently %i", me.name, parv[0], spam_time);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "DRONETIME", 9))
	{
	    if (parc > 2)
	    {
		int newval = atoi(parv[2]);

		if (newval < 0)
		{
		    sendto_one(sptr, ":%s NOTICE %s :DRONETIME must be > 0", me.name, parv[0]);
		    return 0;
		}       
                dronetime = newval;

		if (newval == 0)
		    sendto_umode(MODE_s, "%s has disabled the ANTI_DRONE_FLOOD code", parv[0]);
		else
		    sendto_umode(MODE_s, "%s has changed DRONETIME to %i", parv[0], dronetime);
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :DRONETIME is currently %i", me.name, parv[0], dronetime);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "DRONECOUNT", 10))
	{
	    if (parc > 2)
	    {
		int newval = atoi(parv[2]);

		if (newval <= 0)
		{
		    sendto_one(sptr, ":%s NOTICE %s :DRONECOUNT must be > 0", me.name, parv[0]);
		    return 0;
		}
		dronecount = newval;
		sendto_umode(MODE_s, "%s has changed DRONECOUNT to %i", parv[0], dronecount);
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :DRONECOUNT is currently %i", me.name, parv[0], dronecount);
		return 0;
	    }
	}
	else if (!strncasecmp(command, "THROTTLE", 8))
	{
	    char *changed = NULL;
	    char *to = NULL;

	   /* several values available:
	    * ENABLE [on|off] to enable the code
	    * COUNT [n] to set a max count, must be > 1
	    * TIME [n] to set a max time before expiry, must be > 5
	    * RECORDTIME [n] to set a time for the throttle records to expire
	    * HASH [n] to set the size of the hash table, must be bigger than the default */


	   /* only handle individual settings if parc > 3 (they're actually changing stuff) */
	    if (parc > 3)
	    {
		if (!strcasecmp(parv[2], "ENABLE"))
		{
		    changed = "ENABLE";

		    if (tolower(*parv[3]) == 'y' || !strcasecmp(parv[3], "on"))
		    {
			throttle_enable = 1;
			to = "ON";
		    }
		    else if (tolower(*parv[3]) == 'n' || !strcasecmp(parv[3], "off"))
		    {
			throttle_enable = 0;
			to = "OFF";
		    }
		}
		else if (!strcasecmp(parv[2], "COUNT"))
		{
		    int cnt;
		    changed = "COUNT";
		    cnt = atoi(parv[3]);

		    if (cnt > 1)
		    {
			throttle_tcount = cnt;
			to = parv[3];
		    }
		}
		else if (!strcasecmp(parv[2], "TIME"))
		{
		    int cnt;
		    changed = "TIME";
		    cnt = atoi(parv[3]);

		    if (cnt >= 5)
		    {
			throttle_ttime = cnt;
			to = parv[3];
		    }
		}
		else if (!strcasecmp(parv[2], "RECORDTIME"))
		{
		    int cnt;
		    changed = "RECORDTIME";
		    cnt = atoi(parv[3]);

		    if (cnt >= 30)
		    {
			throttle_rtime = cnt;
			to = parv[3];
		    }
		}
		else if (!strcasecmp(parv[2], "HASH"))
		{
		    int cnt;
		    changed = "HASH";
		    cnt = atoi(parv[3]);

		    if (cnt >= THROTTLE_HASHSIZE)
		    {
			throttle_resize(cnt);
			to = parv[3];
		    }
		}
		if (to != NULL)
		{
		    sendto_ops("%s has changed throttle %s to %s", parv[0], changed, to);
		    sendto_one(sptr, ":%s NOTICE %s :set throttle %s to %s", me.name, parv[0], changed, to);
		}
	    }
	    else
	    {
		/* report various things, we cannot easily get the hash size, so leave that alone. */
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE %s", me.name, parv[0], throttle_enable ? "enabled" : "disabled");
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE COUNT=%d", me.name, parv[0], throttle_tcount);
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE TIME=%d sec", me.name, parv[0], throttle_ttime);
		sendto_one(sptr, ":%s NOTICE %s :THROTTLE RECORDTIME=%d sec", me.name, parv[0], throttle_rtime);
	    }
	}
#ifdef IDLE_CHECK
	else if (!strncasecmp(command, "IDLETIME", 8))
	{
	    if (parc > 2)
	    {
		int newval = atoi(parv[2]);
	
		if (newval == 0)
		{
		    sendto_umode(MODE_s, "%s has disabled Idle Check", parv[0]);
		    idle_time = 0;
		}
		else
		{
		    sendto_umode(MODE_s, "%s has changed IDLETIME to %i", parv[0], newval);
		    idle_time = (newval*60);
		}
	    }
	    else
	    {
		sendto_one(sptr, ":%s NOTICE %s :IDLETIME is currently %i", me.name, parv[0], idle_time/60);
		return 0;
	    }
	}
#endif
    }
    else 
    {
	sendto_one(sptr, ":%s NOTICE %s :Options: MAX", me.name, parv[0]);
#ifdef FLUD
	sendto_one(sptr, ":%s NOTICE %s :Options: FLUDNUM, FLUDTIME, FLUDBLOCK", me.name, parv[0]);
#endif
	sendto_one(sptr, ":%s NOTICE %s :Options: SPAMNUM, SPAMTIME", me.name, parv[0]);
	sendto_one(sptr, ":%s NOTICE %s :Options: DRONETIME, DRONECOUNT", me.name, parv[0]);
#ifdef IDLE_CHECK
	sendto_one(sptr, ":%s NOTICE %s :Options: IDLETIME", me.name, parv[0]);
#endif
	sendto_one(sptr, ":%s NOTICE %s :Options: SPLITDELAY", me.name, parv[0]);
	sendto_one(sptr, ":%s NOTICE %s :Options: THROTTLE <ENABLE|COUNT|TIME|BANTIME|HASH> [setting]", me.name, parv[0]);
    }
    return 0;
}

/* m_htm - high traffic mode info */
int m_htm(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#define LOADCFREQ 5
    char *command;
    extern int lifesux, LRV, LCF, noisy_htm; /* in ircd.c */
    extern int currlife;

    if (!MyClient(sptr) || !HasUmodeo(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr,":%s NOTICE %s :HTM is %s(%d), %s. Max rate = %dk/s. Current = %dk/s", me.name, parv[0], lifesux ? "ON" : "OFF",
	lifesux, noisy_htm ? "NOISY" : "QUIET", LRV, currlife);

    if (parc > 1) 
    {
	command = parv[1];

	if (!strcasecmp(command, "TO"))
	{
	    if (parc > 2)
	    {
		int new_value = atoi(parv[2]);

		if (new_value < 10) 
		    sendto_one(sptr, ":%s NOTICE %s :Cannot set LRV < 10!", me.name, parv[0]);
		else
		    LRV = new_value;

		sendto_one(sptr, ":%s NOTICE %s :NEW Max rate = %dk/s. Current = %dk/s", me.name, parv[0], LRV, currlife);
		sendto_umode(MODE_s, "%s!%s@%s set new HTM rate to %dk/s (%dk/s current)", parv[0], sptr->user->username,
		    HasUmodex(sptr) ? sptr->user->virthost : sptr->sockhost, LRV, currlife);
	    }
	    else
		sendto_one(sptr, ":%s NOTICE %s :LRV command needs an integer parameter", me.name, parv[0]);
	}
	else 
	{
	    if (!strcasecmp(command, "ON")) 
	    {
		lifesux = 1;
		sendto_one(sptr, ":%s NOTICE %s :HTM is now ON.", me.name, parv[0]);
		sendto_umode(MODE_s, "Entering high-traffic mode: Forced by %s!%s@%s", parv[0], sptr->user->username,
		    HasUmodex(sptr) ? sptr->user->virthost : sptr->sockhost);
		LCF = 30; /* 30s */
	    }
	    else if (!strcasecmp(command, "OFF")) 
	    {
		if (DYN_HTMLOCKON_BURST == 1)
		{
		    if (HTMLOCK == YES) 
		    {
			if (!HasUmodeA(sptr)) 
			{
			    sendto_one(sptr, ":%s NOTICE %s :Cannot change HTM - Currently LOCKED", me.name, parv[0]);
			    return 0;
			}
			else
			{
			    sendto_one(sptr, ":%s NOTICE %s :Removing HTM LOCK", me.name, parv[0]);
			    sendto_umode(MODE_s, "Resuming standard operation through HTM LOCK: Forced by %s!%s@%s",
				parv[0], sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->sockhost);
			    lifesux = 0;
			    LCF = LOADCFREQ;
			    return 0;
			}
		    }
		}
		lifesux = 0;
		LCF = LOADCFREQ;
		sendto_one(sptr, ":%s NOTICE %s :HTM is now OFF.", me.name, parv[0]);
		sendto_umode(MODE_s, "Resuming standard operation: Forced by %s!%s@%s", parv[0], sptr->user->username,
		    HasUmodex(sptr) ? sptr->user->virthost : sptr->sockhost);
	    }
	    else if (!strcasecmp(command, "QUIET")) 
	    {
		sendto_umode(MODE_s, "HTM is now QUIET");
		noisy_htm = NO;
	    }
	    else if (!strcasecmp(command, "NOISY")) 
	    {
		sendto_umode(MODE_s, "HTM is now NOISY");
		noisy_htm = YES;
	    }
	    else
		sendto_one(sptr, ":%s NOTICE %s :Commands are:HTM [ON] [OFF] [TO int] [QUIET] [NOISY]", me.name, parv[0]);
	}
    }
    return 0;
}

/*
 * cluster() input
 * - pointer to a hostname output 
 * pointer to a static of the hostname masked for use in a kline. side 
 * effects - NONE
 * 
 * reworked a tad -Dianora
 */
static char *cluster(char *hostname)
{
    static char result[HOSTLEN+1];	/* result to return */
    char temphost[HOSTLEN+1];		/* workplace */
    char *ipp;				/* used to find if host is ip # only */
    char *host_mask;			/* used to find host mask portion to '*' */
    /* used to zap last nnn portion of an ip # */
    char *zap_point = (char *) NULL; 
    char *tld;				/* Top Level Domain */
    int is_ip_number;			/* flag if its an IP # */
    int number_of_dots;			/* count # of dots for ip# and domain klines */

    if (!hostname)
	return (char *) NULL; /* EEK! */

    /* If a '@' is found in the hostname, this is bogus and must have
     * been introduced by server that doesn't check for bogus domains
     * (dns spoof) very well. *sigh* just return it... I could also
     * legitimately return (char *)NULL as above.
     * -Dianora */

    if (strchr(hostname, '@')) 
    {
	strncpyzt(result, hostname, HOSTLEN);
	return (result);
    }

    strncpyzt(temphost, hostname, HOSTLEN);

    is_ip_number = YES;		/* assume its an IP# */
    ipp = temphost;
    number_of_dots = 0;

    while (*ipp) 
    {
	if (*ipp == '.') 
	{
	    number_of_dots++;
	    if (number_of_dots == 3)
		zap_point = ipp;
	    ipp++;
	}
	else if (!isdigit(*ipp)) 
	{
	    is_ip_number = NO;
	    break;
	}
	ipp++;
    }
    if (is_ip_number && (number_of_dots == 3)) 
    {
	zap_point++;
	*zap_point++ = '*';	/* turn 111.222.333.444 into ... */
	*zap_point = '\0';	/* 111.222.333.* */
	strncpy(result, temphost, HOSTLEN);
	return (result);
    }
    else 
    {
	tld = strrchr(temphost, '.');
	if (tld) 
	{
	    number_of_dots = 2;

	    if (tld[3])		/* its at least a 3 letter tld */
		number_of_dots = 1;
	    if (tld != temphost)        /* in these days of dns spoofers ... */
		host_mask = tld - 1;	/* Look for host portion to '*' */
	    else
		host_mask = tld;	/* degenerate case hostname is '.com' ect. */

	    while (host_mask != temphost) 
	    {
		if (*host_mask == '.')
		    number_of_dots--;
		if (number_of_dots == 0) 
		{
		    result[0] = '*';
		    strncpy(result + 1, host_mask, HOSTLEN - 1);
		    return (result);
		}
		host_mask--;
	    }
	    result[0] = '*';	/* foo.com => *foo.com */
	    strncpy(result + 1, temphost, HOSTLEN);
	}
	else
	{   /*  no tld found oops. just return it as is */
	    strncpy(result, temphost, HOSTLEN);
	    return (result);
	}
    }
    return (result);
}

/*
 * m_kline()
 */
int m_kline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    struct pkl *k;
#else
    int out;
#endif
    char buffer[1024];
#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    char timebuffer[MAX_DATE_STRING];
    char filenamebuf[1024];
    struct tm *tmptr;
#endif
    char *filename; /* filename to use for kline */
    char *user, *host;
    char *reason;
    char *current_date;
    aClient *acptr;
    char tempuser[USERLEN+2];
    char temphost[HOSTLEN+1];
    aConfItem *aconf;
    int temporary_kline_time = 0;	/* -Dianora */
    time_t temporary_kline_time_seconds = 0;
    int time_specified = 0;
    char *argv;

    if (!MyClient(sptr) || !OPCanKline(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KLINE");
	return 0;
    }

    argv = parv[1];

    if ((temporary_kline_time = isnumber(argv)) >= 0) 
    {
	if (parc < 3) 
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "KLINE");
	    return 0;
	}
	if (temporary_kline_time > (24 * 60))
	    temporary_kline_time = (24 * 60);	/*  Max it at 24 hours */

	temporary_kline_time_seconds = (time_t) temporary_kline_time *(time_t) 60;
	
	/* turn it into minutes */
	argv = parv[2];
	parc--;
	time_specified = 1;
    }
    else
    {
	temporary_kline_time = 0; /* -1 minute klines are bad... :) - lucas */
    }
    if (strchr(argv, ' '))
    {
	sendto_one(sptr, ":%s NOTICE %s :Poorly formatted hostname (contains spaces). Be sure you are using the form: "
	    "/quote KLINE [time] <user@host/nick> :<reason>", me.name, parv[0]);
	return 0;
    }
    if ((host = strchr(argv, '@')) || *argv == '*') 
    {
	/* Explicit user@host mask given */

	if (host) /* Found user@host */
	{
	    user = argv;	/* here is user part */
	    *(host++) = '\0';	/* and now here is host */
	}
	else 
	{
	    user = "*";		/* no @ found, assume its *@somehost */
	    host = argv;
	}

	if (!*host)		/* duh. no host found, assume its '*' host */
	    host = "*";
	strncpyzt(tempuser, user, USERLEN + 2);	/* allow for '*' in front */
	strncpyzt(temphost, host, HOSTLEN);
	user = tempuser;
	host = temphost;
    }
    else 
    {
	/* Try to find user@host mask from nick */

	if (!(acptr = find_chasing(sptr, argv, NULL)))
	    return 0;
	if (!acptr->user)
	    return 0;
	if (IsServer(acptr)) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :Can't KLINE a server, use @'s where appropriate", me.name, parv[0]);
	    return 0;
	}

	/* turn the "user" bit into "*user", blow away '~' if found in original user name (non-idented) */

	tempuser[0] = '*';

	if (*acptr->user->username == '~')
	    strcpy(tempuser + 1, (char *) acptr->user->username + 1);
	else
	    strcpy(tempuser + 1, acptr->user->username);
	user = tempuser;
	host = cluster(acptr->user->host);
    }

    if (time_specified)
	argv = parv[3];
    else
	argv = parv[2];

#ifdef DEFAULT_KLINE_TIME
    if (time_specified == 0)
    {
	temporary_kline_time = DEFAULT_KLINE_TIME;
	temporary_kline_time_seconds = (time_t) temporary_kline_time *(time_t) 60;
    }
#endif
    if (parc > 2) 
    {
	if (strchr(argv, ':')) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character ':' in comment", me.name, parv[0]);
	    return 0;
	}
	if (strchr(argv, '#')) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character '#' in comment", me.name, parv[0]);
	    return 0;
	}
	if (*argv)
	    reason = argv;
	else
	    reason = "No reason";
    }
    else
	reason = "No reason";

    if (!match(user, "akjhfkahfasfjd") && !match(host, "ldksjfl.kss...kdjfd.jfklsjf")) 
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Error -- Can't K-Line *@*", me.name, parv[0]);
	return 0;
    }
#ifdef NON_REDUNDANT_KLINES
    if ((aconf = find_is_klined(host, user))) 
    {
	char *reason;
	reason = aconf->passwd ? aconf->passwd : "<No Reason>";

	sendto_one(sptr, ":%s NOTICE %s :[%s@%s] already K-lined by [%s@%s] - %s", me.name, parv[0], user, host, aconf->name, aconf->host, reason);
	return 0;

    }
#endif

    current_date = smalldate((time_t) 0);

    aconf = make_conf();
    aconf->status = CONF_KILL;
    DupString(aconf->host, host);

    (void) ircsprintf(buffer, "%s (%s)", reason, current_date);
    
    DupString(aconf->passwd, buffer);
    DupString(aconf->name, user);
    aconf->port = 0;

    if (temporary_kline_time) 
    {
	aconf->hold = timeofday + temporary_kline_time_seconds;
	add_temp_kline(aconf);
	rehashed = YES;
	zline_in_progress = NO;
	sendto_umode(MODE_s, "%s added temporary %d min. k-line for [%s@%s] [%s]", parv[0], temporary_kline_time, user, host, reason);
	return 0;
    }

    Class       (aconf) = find_class(0);

    /* when can aconf->host ever be a NULL pointer though? or for that
     * matter, when is aconf->status ever not going to be CONF_KILL at
     * this point?  -Dianora */

    switch (sortable(host))
    {
    case 0:
	l_addto_conf_list(&KList3, aconf, host_field);
	break;
    case 1:
	addto_conf_list(&KList1, aconf, host_field);
	break;
    case -1:
	addto_conf_list(&KList2, aconf, rev_host_field);
	break;
    }

    /* comstud's SEPARATE_QUOTE_KLINES_BY_DATE code */
    /* Note, that if SEPARATE_QUOTE_KLINES_BY_DATE is defined, it
     * doesn't make sense to have LOCKFILE on the kline file */

#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    tmptr = localtime(&NOW);
    strftime(timebuffer, MAX_DATE_STRING, "%y%m%d", tmptr);
    (void) sprintf(filenamebuf, "%s.%s", klinefile, timebuffer);
    filename = filenamebuf;
    sendto_one(sptr, ":%s NOTICE %s :Added K-Line [%s@%s] to %s", me.name, parv[0], user, host, filenamebuf);
#else
    filename = klinefile;
#ifdef KPATH
    sendto_one(sptr, ":%s NOTICE %s :Added K-Line [%s@%s] to server klinefile", me.name, parv[0], user, host);
#else
    sendto_one(sptr, ":%s NOTICE %s :Added K-Line [%s@%s] to server configfile", me.name, parv[0], user, host);
#endif /* KPATH */
#endif /* SEPARATE_QUOTE_KLINES_BY_DATE */

    rehashed = YES;
    zline_in_progress = NO;
    sendto_umode(MODE_s, "%s added K:Line for [%s@%s] [%s]", parv[0], user, host, reason);
    
#if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    /* MDP - Careful, don't cut & paste that ^O */

    if ((k = (struct pkl *) malloc(sizeof(struct pkl))) == NULL) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, user, host, reason);

    if ((k->comment = strdup(buffer)) == NULL) 
    {
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "K:%s:%s (%s):%s\n", host, reason, current_date, user);

    if ((k->kline = strdup(buffer)) == NULL) 
    {
	free(k->comment);
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }

    k->next = pending_klines;
    pending_klines = k;

    do_pending_klines();
    return (0);

#else /*  LOCKFILE - MDP and not SEPARATE_KLINES_BY_DATE */

    if ((out = open(filename, O_RDWR | O_APPEND | O_CREAT)) == -1) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem opening %s ", me.name, parv[0], filename);
	return 0;
    }

# ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    fchmod(out, 0660);
# endif

    (void) ircsprintf(buffer, "#%s!%s@%s K'd: %s@%s:%s\n", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, user, host, reason);

    if (write(out, buffer, strlen(buffer)) <= 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name, parv[0], filename);
	(void) close(out);
	return 0;
    }

    (void) ircsprintf(buffer, "K:%s:%s (%s):%s\n", host, reason, current_date, user);

    if (write(out, buffer, strlen(buffer)) <= 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name, parv[0], filename);
	(void) close(out);
	return 0;
    }

    (void) close(out);
#ifdef USE_SYSLOG
    ircd_log("%s added K:Line for [%s@%s] [%s]", parv[0], user, host, reason);
#endif
    return 0;
#endif /* LOCKFILE */
}

/*
 * isnumber()
 * 
 * inputs               
 * - pointer to ascii string in output             
 * - 0 if not an integer number, else the number side effects  
 * - none return -1 if not an integer. 
 * (if someone types in maxint, oh well..) - lucas
 */
static int isnumber(char *p)
{
    int result = 0;

    while (*p) 
    {
	if (isdigit(*p)) 
	{
	    result *= 10;
	    result += ((*p) & 0xF);
	    p++;
	}
	else
	    return (-1);
    }

    /* in the degenerate case where oper does a /quote kline 0 user@host
     * :reason i.e. they specifically use 0, I am going to return 1
     * instead as a return value of non-zero is used to flag it as a
     * temporary kline */

    /* er, no. we return 0 because 0 means that it's a permanent kline. -lucas
     * oh, and we only do this if DEFAULT_KLINE_TIME is specified. */

#ifndef DEFAULT_KLINE_TIME
    if (result == 0)
	result = 1;
#endif
    return (result);
}

/*
 * m_unkline() 
 *  parv[0] = sender 
 *  parv[1] = address to remove
 */
int m_unkline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int in, out;
    int pairme = NO;
    char buf[256], buff[256];
    char temppath[256];
#ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    char timebuffer[MAX_DATE_STRING];
    char filenamebuf[1024];
    struct tm *tmptr;
#endif
    char *filename; /* filename to use for unkline */
    char *user, *host;
    char *p;
    int nread;
    int error_on_write = NO;
    aConfItem *aconf;
    struct stat oldfilestat;
    mode_t oldumask;

    ircsprintf(temppath, "%s.tmp", klinefile);

    if (check_registered(sptr)) 
	return -1;
    if (!OPCanUnKline(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNKLINE");
	return 0;
    }
    if ((host = strchr(parv[1], '@')) || *parv[1] == '*')
    {
	/* Explicit user@host mask given */

	if (host) /* Found user@host */
	{
	    user = parv[1];	/* here is user part */
	    *(host++) = '\0';	/* and now here is host */
	}
	else 
	{
	    user = "*";		/* no @ found, assume its *@somehost */
	    host = parv[1];
	}
    }
    else
    {
	sendto_one(sptr, ":%s NOTICE %s :Invalid parameters", me.name, parv[0]);
	return 0;
    }
    if ((user[0] == '*') && (user[1] == '\0') && (host[0] == '*') && (host[1] == '\0')) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Cannot UNK-Line everyone", me.name, parv[0]);
	return 0;
    }
# if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    if (lock_kline_file() < 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :%s is locked try again in a few minutes", me.name, parv[0], klinefile);
	return -1;
    }
# endif
# ifdef SEPARATE_QUOTE_KLINES_BY_DATE
    tmptr = localtime(&NOW);
    strftime(timebuffer, MAX_DATE_STRING, "%y%m%d", tmptr);
    (void) sprintf(filenamebuf, "%s.%s", klinefile, timebuffer);
    filename = filenamebuf;
# else
    filename = klinefile;
# endif

    if ((in = open(filename, O_RDONLY)) == -1) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Cannot open %s", me.name, parv[0], filename);
# if defined(LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	return 0;
    }
    if (fstat(in, &oldfilestat) < 0)	/*  Save the old file mode */
	oldfilestat.st_mode = 0644;
    
    oldumask = umask(0);		/* ircd is normally too paranoid */
    
    if ((out = open(temppath, O_WRONLY | O_CREAT, oldfilestat.st_mode)) == -1) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Cannot open %s", me.name, parv[0], temppath);
	(void) close(in);
# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	umask(oldumask);		/* Restore the old umask */
	return 0;
    }
    umask(oldumask);		/* Restore the old umask */
    
    /* #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason K:bar:No
     * reason (1997/08/30 14.56):foo */

    clear_conf_list(&KList1);
    clear_conf_list(&KList2);
    clear_conf_list(&KList3);
    
    while ((nread = dgets(in, buf, sizeof(buf))) > 0)
    {
	buf[nread] = '\0';
	
	if ((buf[1] == ':') && ((buf[0] == 'k') || (buf[0] == 'K')))
	{
	    /* its a K: line */
	    
	    char *found_host;
	    char *found_user;
	    char *found_comment;
	    
	    strcpy(buff, buf);
	    
	    p = strchr(buff, '\n');
	    if (p)
		*p = '\0';
	    p = strchr(buff, '\r');
	    if (p)
		*p = '\0';

	    found_host = buff + 2;	/* point past the K: */
	    
	    p = strchr(found_host, ':');
	    if (p == (char *) NULL) 
	    {
		sendto_one(sptr, ":%s NOTICE %s :K-Line file corrupted", me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :Couldn't find host", me.name, parv[0]);

		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		continue; /* This K line is corrupted ignore */
	    }
	    *p = '\0';
	    p++;
	    
	    found_comment = p;
	    p = strchr(found_comment, ':');
	    if (p == (char *) NULL) 
	    {
		sendto_one(sptr, ":%s NOTICE %s :K-Line file corrupted", me.name, parv[0]);
		sendto_one(sptr, ":%s NOTICE %s :Couldn't find comment", me.name, parv[0]);

		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		continue;	/* This K line is corrupted ignore */
	    }
	    *p = '\0';
	    p++;

	    found_user = p;

	    /* Ok, if its not an exact match on either the user or the
	     * host then, write the K: line out, and I add it back to the
	     * K line tree */

	    if (strcasecmp(host, found_host) || strcasecmp(user, found_user)) 
	    {
		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		aconf = make_conf();
		aconf->status = CONF_KILL;
		DupString(aconf->host, found_host);
		DupString(aconf->name, found_user);
		DupString(aconf->passwd, found_comment);
		aconf->port = 0;
		Class       (aconf) = find_class(0);

		switch (sortable(found_host)) 
		{
		case 0:
		    l_addto_conf_list(&KList3, aconf, host_field);
		    break;
		case 1:
		    addto_conf_list(&KList1, aconf, host_field);
		    break;
		case -1:
		    addto_conf_list(&KList2, aconf, rev_host_field);
		    break;
		}
	    }
	    else
		pairme++;
	}
	else if (buf[0] == '#') 
	{
	    char *userathost;
	    char *found_user;
	    char *found_host;

	    strcpy(buff, buf);

	    /* #Dianora!db@ts2-11.ottawa.net K'd: foo@bar:No reason
	     * K:bar:No reason (1997/08/30 14.56):foo
	     * 
	     * If its a comment coment line, i.e. #ignore this line Then just
	     * ignore the line */

	    p = strchr(buff, ':');

	    if (p == (char *) NULL) 
	    {
		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		continue;
	    }
	    *p = '\0';
	    p++;

	    userathost = p;
	    p = strchr(userathost, ':');

	    if (p == (char *) NULL) 
	    {
		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		continue;
	    }
	    *p = '\0';
	    
	    while (*userathost == ' ')
		userathost++;
	    
	    found_user = userathost;
	    p = strchr(found_user, '@');
	    if (p == (char *) NULL) 
	    {
		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
		continue;
	    }
	    *p = '\0';
	    found_host = p;
	    found_host++;
	    
	    if ((strcasecmp(found_host, host)) || (strcasecmp(found_user, user))) 
	    {
		if (!error_on_write)
		    error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
	    }
	}
	else
	{   /* its the ircd.conf file, and not a Kline or comment */
	    if (!error_on_write)
		error_on_write = flush_write(sptr, parv[0], out, buf, strlen(buf), temppath);
	}
    }
    
    (void) close(in);
    
    /* The result of the rename should be checked too... oh well */
    /*
     * If there was an error on a write above, then its been reported
     * and I am not going to trash the original kline /conf file
     * -Dianora */

    if ((!error_on_write) && (close(out) >= 0))
	(void) rename(temppath, filename);
    else
    {
	sendto_one(sptr, ":%s NOTICE %s :Couldn't write temp kline file, aborted", me.name, parv[0]);
# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
	(void) unlink(LOCKFILE);
# endif
	return -1;
    }
# if defined (LOCKFILE) && !defined(SEPARATE_QUOTE_KLINES_BY_DATE)
    (void) unlink(LOCKFILE);
# endif
    if (pairme == NO) 
    {
	if (!(pairme = remove_temp_kline(host, user, CONF_KILL)))
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- No K:Line for %s@%s", me.name, parv[0], user, host);
	    return 0;
	}
    }
    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- K:Line for [%s@%s] is removed", me.name, parv[0], user, host);
    sendto_umode(MODE_s, "%s has removed the K:Line for: [%s@%s] (%d matches)", parv[0], user, host, pairme);
    return 0;
}

/*
 * flush_write()
 * 
 * inputs               
 * - pointer to client structure of oper requesting unkline 
 * - out is the file descriptor 
 * - buf is the buffer to write 
 * - ntowrite is the expected number of character to be written 
 * - temppath is the temporary file name to be written output
 *   YES for error on write 
 *   NO for success side effects        
 * - if successful, the buf is written to output file if a write failure
 *   happesn, and the file pointed to by temppath, if its non NULL, is
 *   removed.
 * 
 * The idea here is, to be as robust as possible when writing to the kline
 * file.
 * 
 * Yes, I could have dug the opernick out of sptr. I didn't feel like it.
 * so sue me.
 * 
 * -Dianora
 */
static int flush_write(aClient *sptr, char *opernick, int out, char *buf, int ntowrite, char *temppath)
{
    int nwritten;
    int error_on_write = NO;

    nwritten = write(out, buf, ntowrite);

    if (nwritten != ntowrite) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Unable to write to %s", me.name, opernick, temppath);
	error_on_write = YES;
	(void) close(out);

	if (temppath != (char *) NULL)
	    (void) unlink(temppath);
    }
    return (error_on_write);
}

/*
 * re-worked a tad added Rodders dated KLINE code -Dianora
 * 
 * BUGS:        There is a concern here with LOCKFILE handling the LOCKFILE
 * code only knows how to talk to the kline file. Having an extra
 * define allowing D lines to go into ircd.conf or the kline file
 * complicates life. LOCKFILE code should still respect any lock
 * placed.. The fix I believe is that we are going to have to also pass
 * along in the struct pkl struct, which file the entry is to go
 * into... or.. just remove the ZLINES_IN_KPATH option. -Dianora
 */
int m_zline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#if defined(LOCKFILE) && defined(ZLINES_IN_KPATH)
    struct pkl *k;
#else
    int out;
#endif
    char buffer[1024];
    char *host, *reason, *p;
    char *current_date;
    int number_of_dots = 0;
    aClient *acptr;
    aConfItem *aconf;

    if (!MyClient(sptr) || !OFlagz(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2) 
    {
	if (MyClient(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ZLINE");
	return 0;
    }

    host = parv[1];

    p = host;
    while (*p) 
    {
	if (*p == '.') 
	{
	    number_of_dots++;
	    p++;
	}
	else if (*p == '*') 
	{
	    if (number_of_dots != 3) 
	    {
		sendto_one(sptr, ":%s NOTICE %s :Z:Line format error, use xxx.xxx.xxx.xxx or xxx.xxx.xxx.*", me.name, parv[0]);
		return 0;
	    }
	    else
	    {
		*(p + 1) = '\0';
		break;
	    }
	}
	else 
	{
	    if (!isdigit(*p)) 
	    {
		if (!(acptr = find_chasing(sptr, parv[1], NULL)))
		    return 0;
		if (!acptr->user)
		    return 0;
		if (IsServer(acptr)) 
		{
		    sendto_one(sptr, ":%s NOTICE %s :Can't ZLINE a server silly", me.name, parv[0]);
		    return 0;
		}
		if (!MyConnect(acptr)) 
		{
		    sendto_one(sptr, ":%s NOTICE :%s :Can't ZLINE nick on another server", me.name, parv[0]);
		    return 0;
		}

		host = cluster(acptr->hostip);
		break;
	    }
	    p++;
	}
    }
    if (parc > 2)	/* host :reason */
    {
	if (strchr(parv[2], ':')) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character ':' in comment", me.name, parv[0]);
	    return 0;
	}
	if (strchr(parv[2], '#')) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :Invalid character '#' in comment", me.name, parv[0]);
	    return 0;
	}
	
	if (*parv[2])
	    reason = parv[2];
	else
	    reason = "No reason";
    }
    else
	reason = "No reason";

#ifdef NON_REDUNDANT_KLINES
    if ((aconf = find_is_zlined(host))) 
    {
	char *reason = aconf->passwd ? aconf->passwd : "<No Reason>";

	sendto_one(sptr, ":%s NOTICE %s :[%s] already z-lined by [%s] - %s", me.name, parv[0], host, aconf->host, reason);
	return 0;
    }
#endif

    current_date = smalldate((time_t) 0);

    aconf = make_conf();
    aconf->status = CONF_ZLINE;
    DupString(aconf->host, host);

    (void) ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);
    aconf->name = (char *) NULL;
    aconf->port = 0;
    Class       (aconf) = find_class(0);

    /* since a quote zline always ensures a sortable pattern there is no
     * need to check sortable() - Dianora */

    addto_conf_list(&ZList1, aconf, host_field);

    sendto_umode(MODE_s, "%s added Z:Line for [%s] [%s]", parv[0], host, reason);
#ifdef ZLINES_IN_KPATH
    sendto_one(sptr, ":%s NOTICE %s :Added Z:Line [%s] to server klinefile", me.name, parv[0], host);
#else
    sendto_one(sptr, ":%s NOTICE %s :Added Z:Line [%s] to server configfile", me.name, parv[0], host);
#endif
    /* I moved the following 2 lines up here because we still want
     * the server to hunt for 'targetted' clients even if there are
     * problems adding the Z-line the the appropriate file. -ThemBones */
    rehashed = YES;
    zline_in_progress = YES;
#if defined(LOCKFILE) && defined(ZLINES_IN_KPATH)

    if ((k = (struct pkl *) MyMalloc(sizeof(struct pkl))) == NULL) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "#%s!%s@%s D'd: %s:%s (%s)\n", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, host, reason, current_date);

    if ((k->comment = strdup(buffer)) == NULL) 
    {
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }

    (void) ircsprintf(buffer, "D:%s:%s (%s):\n", host, reason, current_date);

    if ((k->kline = strdup(buffer)) == NULL) 
    {
	free(k->comment);
	free(k);
	sendto_one(sptr, ":%s NOTICE %s :Problem allocating memory", me.name, parv[0]);
	return (0);
    }
    k->next = pending_klines;
    pending_klines = k;

    do_pending_klines();
    return (0);

#else /* LOCKFILE - MDP */
    if ((out = open(zlinefile, O_RDWR | O_APPEND | O_CREAT)) == -1) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem opening %s ", me.name, parv[0], zlinefile);
	return 0;
    }

    (void) ircsprintf(buffer, "#%s!%s@%s Z'd: %s:%s (%s)\n", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, host, reason, current_date);

    if (write(out, buffer, strlen(buffer)) <= 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name, parv[0], zlinefile);
	close(out);
	return 0;
    }

    (void) ircsprintf(buffer, "Z:%s:%s (%s):\n", host, reason, current_date);

    if (write(out, buffer, strlen(buffer)) <= 0) 
    {
	sendto_one(sptr, ":%s NOTICE %s :Problem writing to %s", me.name, parv[0], zlinefile);
	close(out);
	return 0;
    }
    (void) close(out);
    return 0;
#endif /* LOCKFILE */
}

/*
 * m_gline()
 *  parv[0] = sender prefix
 *  parv[1] = add/del
 *  parv[2] = user@host
 *  parv[3] = duration (seconds)
 *  parv[4] = reason
 */
int m_gline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *set_on, *set_by, *username, *hostname, *reason, *p, *tmpmask, *x;
    char *current_date;
    char gmt[256];
    time_t secs, expire, set_at;
    aGline *ag;
    int perm, bad_dns, dots;

    if (!HasUmodeo(sptr) && !IsServer(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
	return 0;
    }

    (void) check_expire_gline();

    tmpmask = parv[2];

    if (parc > 2)
    {
	x = strchr(tmpmask, '@');

	if (!x)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Please use a user@mask", me.name, parv[0]);
	    return 0;
	}
    }
    username = strtok(tmpmask, "@");

    if (!username)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Missing username.", me.name, parv[0]);
	return 0;
    }

    hostname = strtok(NULL, "@");

    if (!hostname)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Missing hostname.", me.name, parv[0]);
	return 0;
    }
    if (!strcasecmp(parv[1], "ADD"))
    {
	if (parc < 5)
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
	    return 0;
	}

	set_by = parv[0];
	reason = parv[4];

	if (strlen(reason) > TOPICLEN)
	    reason[TOPICLEN+1] = '\0';
	if (strlen(username) > (USERLEN+1))
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Usernames are limited to %i characters.", me.name, parv[0], (USERLEN+1));
	    return 0;
	}
	if (strlen(hostname) > (HOSTLEN+1))
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames are limited to %i characters.", me.name, parv[0], (HOSTLEN+1));
	    return 0;
	}
	if (!match(username, "akjhfkahfasfjd") && !match(hostname, "ldksjfl.kss...kdjfd.jfklsjf")) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Cant G:LINE such a broad username/hostname mask.", me.name, parv[0]);
	    return 0;
	}

	dots = 0;
	p = hostname;
	bad_dns = NO;

	while (*p)
	{
	    if (!isalnum(*p))
	    {
		if ((*p != '-') && (*p != '.') && (*p != '*'))
		bad_dns = YES;
	    }
	    if (*p == '.')
		dots++;
	    p++;
	}

	if (bad_dns == YES)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- A hostname may contain a-z, 0-9, '-' & '.' & '*' - please use only these.", me.name, parv[0]);
	    return 0;
	}
	if (!dots)
	{
	    if (MyConnect(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- You must have at least one '.' (dot) in the hostname.", me.name, parv[0]);
	    return 0;
	}
	if (find_gline(username, hostname) == 1)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Username/hostname mask is already on the G:Line list", me.name, parv[0]);
	    return 0;
	}

	secs = atol(parv[3]);
	perm = 0;

	if (secs < 0)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Specify a POSITIVE value for a gline expiry time", me.name, parv[0]);
	    return 0;
	}
	else if (secs == 0)
	{
	    perm = 1;
	    expire = secs;
	}
	else
	{
	    perm = 0;
	    expire = secs + timeofday;
	}
#ifdef __OpenBSD__
	strncpy(gmt, asctime(gmtime((clock_t *)&expire)), sizeof(gmt));
#else
	strncpy(gmt, asctime(gmtime((time_t *)&expire)), sizeof(gmt));
#endif
	gmt[strlen(gmt) - 1] = '\0';
	set_at = timeofday;

	if (!IsServer(sptr))
	{
	    if (perm == 1)
	    {
		sendto_umode(MODE_s, "%s added G:Line for %s@%s which will not expire (%s)", set_by, username, hostname, reason);
#ifdef USE_SYSLOG
		ircd_log("%s added G:Line for %s@%s which will not expire (%s)", set_by, username, hostname, reason);
#endif
	    }
	    else
	    {
		sendto_umode(MODE_s, "%s added G:Line for %s@%s which will expire on %s GMT (%s)", set_by, username, hostname, gmt, reason);
#ifdef USE_SYSLOG
		ircd_log("%s added G:Line for %s@%s which will expire on %s GMT (%s)", set_by, username, hostname, gmt, reason);
#endif
	    }
	}
	sendto_serv_butone(cptr, ":%s GLINE ADD %s@%s %l :%s", parv[0], username, hostname, secs, reason);
	current_date = smalldate((time_t) 0);
	set_on = current_date;
	(void) add_gline(set_on, set_by, username, hostname, expire, set_at, reason);
	(void) sweep_gline();
	return 1;
    }
    else if (!strcasecmp(parv[1], "DEL"))
    {
	if (parc < 3)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /gline del <user@host>", me.name, parv[0]);
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "GLINE");
	    return 0;
	}
	if (find_gline(username, hostname) == 0)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Username/hostname mask not found on the G:Line list.", me.name, parv[0]);
	    return 0;
	}

	for (ag = glines; ag; ag = ag->next)
	{
	del_gline(username, hostname);
	sendto_umode(MODE_s, "%s deleted G:Line for %s@%s", parv[0], username, hostname);
	sendto_serv_butone(cptr, ":%s GLINE DEL %s@%s", parv[0], username, hostname);
	return 0;
	}
    }
    else
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /gline <add/del> <user@host> [duration in seconds] [:reason]", me.name, parv[0]);
	return 0;
    }
    return 0;
}

/*
 * find_gline()
 */
int find_gline(username, hostname)
    char *username, *hostname;
{
    aGline *ag;

    (void) check_expire_gline();

    for (ag = glines; ag; ag = ag->next)
    {
	if (!match(ag->username, username) && !match(ag->hostname, hostname))
	    return 1;
    }
    return 0;
}

int sweep_gline()
{
    aClient *acptr;
    int i;

    (void) check_expire_gline();

    for (i = 0; i <= highest_fd; i++)
    {
	if (!(acptr = local[i]) || !IsClient(acptr))
	    continue;
	find_glined_user(acptr);
    }
    return 1;
}

/*
 * find_glined_user()
 */
int find_glined_user(aClient *cptr)
{
    aGline *ag;

    (void) check_expire_gline();

    if (IsServer(cptr) || IsMe(cptr) || IsULine(cptr))
	return 0;

    for (ag = glines; ag; ag = ag->next)
    {
	if ((cptr->user->username) && (((cptr->sockhost) && (!match(ag->hostname, cptr->sockhost) && !match(ag->username, cptr->user->username))) ||
	    ((cptr->hostip) && (!match(ag->hostname, cptr->hostip) && !match(ag->username, cptr->user->username)))))
	{
	    ircsprintf(glreason, "G:Lined: %s (%s)", ag->reason, ag->set_on);
	    return 1;
	}
    }
    return 0;
}

/*
 * add_gline()
 */
void add_gline(set_on, set_by, username, hostname, expire, set_at, reason)
    char *username, *hostname;
    char *set_on, *set_by, *reason;
    time_t expire, set_at;
{
    aGline *ag;

    ag = (aGline *)MyMalloc(sizeof(aGline));

    AllocCpy(ag->set_on, set_on);
    AllocCpy(ag->set_by, set_by);
    AllocCpy(ag->username, username);
    AllocCpy(ag->hostname, hostname);
    AllocCpy(ag->reason, reason);

    ag->expire = expire;
    ag->set_at = set_at;
    ag->next = glines;
    ag->prev = NULL;

    if (glines)
    {
	glines->prev = ag;
    }
    glines = ag;
}
/*
 * del_gline()
 */
int del_gline(username, hostname)
    char *username, *hostname;
{
    aGline *ag;
    
    for (ag = glines; ag; ag = ag->next)
    {
        if (!match(ag->username, username) && !match(ag->hostname, hostname))
        {
            MyFree((char *)ag->set_on);
            MyFree((char *)ag->set_by);
            MyFree((char *)ag->username);
            MyFree((char *)ag->hostname);
            MyFree((char *)ag->reason);

	    if (ag->prev)
		ag->prev->next = ag->next;
	    else
		glines = ag->next;
	    if (ag->next)
		ag->next->prev = ag->prev;

            MyFree((aGline *) ag);
            return 1;
        }
    }
    return 0;
}

/*
 * check_expire_gline()
 */
void check_expire_gline(void)
{
    aGline *ag;
    time_t nowtime = timeofday;

    for (ag = glines; ag; ag = ag->next)
    {
	if ((ag->expire <= nowtime) && (ag->expire != 0))
	    expire_gline(ag);
    }
}

/*
 * expire_gline()
 */
aGline *expire_gline(ag)
    aGline *ag;
{
    sendto_umode(MODE_s, "G:Line for %s@%s (set by %s %l seconds ago) has now expired",
	ag->username, ag->hostname, ag->set_by, (timeofday - ag->set_at), ag->reason);

#ifdef USE_SYSLOG
    ircd_log("G:Line for %s@%s (set by %s %li seconds ago) has now expired",
	ag->username, ag->hostname, ag->set_by, (timeofday - ag->set_at), ag->reason);
#endif
    del_gline(ag->username, ag->hostname);
    return 0;
}

/*
 * send_glines()
 */
void send_glines(aClient *cptr)
{
    aGline *ag;
    time_t expire;

    (void) check_expire_gline();

    for (ag = glines; ag; ag = ag->next)
    {
	expire = ag->expire;

	if (expire < 0)
	    expire = -expire;
	else if (expire > 0)
	    expire = expire - timeofday;
	sendto_one(cptr, ":%s GLINE ADD %s@%s %lu :%s", me.name, ag->username, ag->hostname,
	    (ag->expire == 0) ? ag->expire : ag->expire - timeofday, ag->reason);
    }
}

/*
 * list_glines()
 */
void list_glines(aClient *sptr)
{
    aGline *ag;

    (void) check_expire_gline();

    if (!glines)
    {
	sendto_one(sptr, ":%s %d %s :There are currently no G:Lines in place on %s", me.name, RPL_SETTINGS, sptr->name, me.name);
	return;
    }

    for (ag = glines; ag; ag = ag->next)
    {
	sendto_one(sptr, ":%s %i %s :G %s@%s %s %s (%s)", me.name, RPL_TEXT, sptr->name,
	    ag->username, ag->hostname, ag->reason, ag->set_on, ag->set_by);
    }
}

/*
 * m_sline()
 *  parv[0] = sender prefix
 *  parv[1] = add/del
 *  parv[2] = user@host
 *  parv[3] = duration (seconds)
 *  parv[4] = reason
 */
int m_sline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *set_on, *set_by, *username, *hostname, *reason, *p, *tmpmask, *x;
    char *current_date;
    char gmt[256];
    time_t secs, expire, set_at;
    aSline *as;
    int perm, bad_dns, dots;

    if (!HasUmodeo(sptr) && !IsServer(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SLINE");
	return 0;
    }

    (void) check_expire_sline();

    tmpmask = parv[2];


    if (parc > 2)
    {
/* Check to see if the mask is a user@mask */
	x = strchr(tmpmask, '@');

	if (!x)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Please use a user@mask", me.name, parv[0]);
	    return 0;
	}
    }
    username = strtok(tmpmask, "@");

    if (!username)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Missing username.", me.name, parv[0]);
	return 0;
    }

    hostname = strtok(NULL, "@");

    if (!hostname)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Missing hostname.", me.name, parv[0]);
	return 0;
    }
    if (!strcasecmp(parv[1], "ADD"))
    {
	if (parc < 5)
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SLINE");
	    return 0;
	}

	set_by = parv[0];
	reason = parv[4];

	if (strlen(reason) > TOPICLEN)
	    reason[TOPICLEN + 1] = '\0';
	if (strlen(username) > (USERLEN+1))
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Usernames are limited to %i characters.", me.name, parv[0], (USERLEN+1));
	    return 0;
	}
	if (strlen(hostname) > (HOSTLEN+1))
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames are limited to %i characters.", me.name, parv[0], (HOSTLEN+1));
	    return 0;
	}
	if (!match(username, "akjhfkahfasfjd") && !match(hostname, "ldksjfl.kss...kdjfd.jfklsjf")) 
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Can't S:LINE such a broad username/hostname mask.",
		me.name, parv[0], username, hostname);
	    return 0;
	}

	dots = 0;
	p = hostname;
	bad_dns = NO;

	while (*p)
	{
	    if (!isalnum(*p))
	    {
		if ((*p != '-') && (*p != '.') && (*p != '*'))
		bad_dns = YES;
	    }
	    if (*p == '.')
		dots++;
	    p++;
	}

	if (bad_dns == YES)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- A hostname may contain a-z, 0-9, '-' & '.' & '*' - please use only these.",
		me.name, parv[0]);
	    return 0;
	}
	if (!dots)
	{
	    if (MyConnect(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- You must have at least one '.' (dot) in the hostname.",
		    me.name, parv[0]);
	    return 0;
	}
	if (find_sline(username, hostname) == 1)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Username/hostname mask is already on the S:Line list",
		    me.name, parv[0]);
	    return 0;
	}

	secs = atol(parv[3]);
	perm = 0;

	if (secs < 0)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Specify a POSITIVE value for a S:Line expiry time.",
		    me.name, parv[0]);
	    return 0;
	}
	else if (secs == 0)
	{
	    perm = 1;
	    expire = secs;
	}
	else
	{
	    perm = 0;
	    expire = secs + timeofday;
	}
#ifdef __OpenBSD__
	strncpy(gmt, asctime(gmtime((clock_t *)&expire)), sizeof(gmt));
#else
	strncpy(gmt, asctime(gmtime((time_t *)&expire)), sizeof(gmt));
#endif
	gmt[strlen(gmt) - 1] = '\0';
	set_at = timeofday;

	if (!IsServer(sptr))
	{
	    if (perm == 1)
	    {
		sendto_umode(MODE_s, "%s added S:Line for %s@%s which will not expire (%s)", set_by, username, hostname, reason);
#ifdef USE_SYSLOG
		ircd_log("%s added S:Line for %s@%s which will not expire (%s)", set_by, username, hostname, reason);
#endif
	    }
	    else
	    {
		sendto_umode(MODE_s, "%s added S:Line for %s@%s which will expire on %s GMT (%s)", set_by, username, hostname, gmt, reason);
#ifdef USE_SYSLOG
		ircd_log("%s added S:Line for %s@%s which will expire on %s GMT (%s)", set_by, username, hostname, gmt, reason);
#endif
	    }
	}
	sendto_serv_butone(cptr, ":%s SLINE ADD %s@%s %l :%s", parv[0], username, hostname, secs, reason);
	current_date = smalldate((time_t) 0);
	set_on = current_date;
	(void) add_sline(set_on, set_by, username, hostname, expire, set_at, reason);
	(void) sweep_sline();
	return 1;
    }
    else if (!strcasecmp(parv[1], "DEL"))
    {
	if (parc < 3)
	{
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /sline del <user@host>", me.name, parv[0]);
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SLINE");
	    return 0;
	}
	if (find_sline(username, hostname) == 0)
	{
	    if (!IsServer(sptr) && MyClient(sptr))
		sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Username/hostname mask not found on the S:Line list.", me.name, parv[0]);
	    return 0;
	}

	for (as = slines; as; as = as->next)
	{
	    sendto_umode(MODE_s, "%s deleted S:Line on %s@%s (set by %s %l seconds ago)",
		parv[0], as->username, as->hostname, as->set_by, (timeofday - as->set_at));

#ifdef USE_SYSLOG
	    ircd_log("%s deleted S:Line on %s@%s (set by %s %l seconds ago)",
		parv[0], as->username, as->hostname, as->set_by, (timeofday - as->set_at));
#endif
	    del_sline(username, hostname);
	    sendto_serv_butone(cptr, ":%s SLINE DEL %s@%s", parv[0], username, hostname);
	    return 0;
	}
    }
    else
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Syntax: /sline <add/del> <user@host> [duration in seconds] [:reason]", me.name, parv[0]);
	return 0;
    }
    return 0;
}

/*
 * find_sline()
 */
int find_sline(username, hostname)
    char *username, *hostname;
{
    aSline *as;

    (void) check_expire_sline();

    for (as = slines; as; as = as->next)
    {
	if (!match(as->username, username) && !match(as->hostname, hostname))
	    return 1;
    }
    return 0;
}

int sweep_sline()
{
    aClient *acptr;
    int i;

    (void) check_expire_sline();

    for (i = 0; i <= highest_fd; i++)
    {
	if (!(acptr = local[i]) || !IsClient(acptr))
	    continue;
	find_slined_user(acptr);
    }
    return 1;
}

/*
 * find_slined_user()
 */
int find_slined_user(aClient *cptr)
{
    aSline *as;

    (void) check_expire_sline();

    if (IsServer(cptr) || IsMe(cptr) || IsULine(cptr))
	return 0;

    for (as = slines; as; as = as->next)
    {
	if ((cptr->user->username) && (((cptr->sockhost) && (!match(as->hostname, cptr->sockhost) && !match(as->username, cptr->user->username))) ||
	    ((cptr->hostip) && (!match(as->hostname, cptr->hostip) && !match(as->username, cptr->user->username)))))
	    return 1;
    }
    return 0;
}

/*
 * add_sline()
 */
void add_sline(set_on, set_by, username, hostname, expire, set_at, reason)
    char *username, *hostname;
    char *set_on, *set_by, *reason;
    time_t expire, set_at;
{
    aSline *as;

    as = (aSline *)MyMalloc(sizeof(aSline));

    AllocCpy(as->set_on, set_on);
    AllocCpy(as->set_by, set_by);
    AllocCpy(as->username, username);
    AllocCpy(as->hostname, hostname);
    AllocCpy(as->reason, reason);

    as->expire = expire;
    as->set_at = set_at;
    as->next = slines;
    as->prev = NULL;

    if (slines)
	slines->prev = as;

    slines = as;
}

/*
 * del_sline()
 */
int del_sline(username, hostname)
    char *username, *hostname;
{
    aSline *as;

    for (as = slines; as; as = as->next)
    {
	if (!match(as->username, username) && !match(as->hostname, hostname))
	{
	    MyFree((char *)as->set_on);
	    MyFree((char *)as->set_by);
	    MyFree((char *)as->username);
	    MyFree((char *)as->hostname);
	    MyFree((char *)as->reason);

	    if (as->prev)
		as->prev->next = as->next;
	    else
		slines = as->next;
	    if (as->next)
		as->next->prev = as->prev;

	    MyFree((aSline *) as);
	    return 1;
	}
    }
    return 0;
}

/*
 * check_expire_sline()
 */
void check_expire_sline(void)
{
    aSline *as;
    time_t nowtime = timeofday;

    for (as = slines; as; as = as->next)
    {
	if ((as->expire <= nowtime) && (as->expire != 0))
	    expire_sline(as);
    }
}

/*
 * expire_sline()
 */
aSline *expire_sline(as)
    aSline *as;
{
    sendto_umode(MODE_s, "S:Line for %s@%s (set by %s reason: %s %l seconds ago) has now expired",
	as->username, as->hostname, as->set_by, as->reason, (timeofday - as->set_at));
#ifdef USE_SYSLOG
    ircd_log("S:Line for %s@%s (set by %s reason: %s %li seconds ago) has now expired",
	as->username, as->hostname, as->set_by, as->reason, (timeofday - as->set_at));
#endif
    del_sline(as->username, as->hostname);
    return 0;
}

/*
 * send_slines()
 */
void send_slines(aClient *cptr)
{
    aSline *as;
    time_t expire;

    (void) check_expire_sline();

    for (as = slines; as; as = as->next)
    {
	expire = as->expire;

	if (expire < 0)
	    expire = -expire;
	else if (expire > 0)
	    expire = expire - timeofday;
	sendto_one(cptr, ":%s SLINE ADD %s@%s %lu :%s", me.name, as->username, as->hostname,
	    (as->expire == 0) ? as->expire : as->expire - timeofday, as->reason);
    }
}

/*
 * list_slines()
 */
void list_slines(aClient *sptr)
{
    aSline *as;

    (void) check_expire_sline();

    if (!slines)
    {
	sendto_one(sptr, ":%s %d %s :There are currently no S:Lines in place on %s", me.name, RPL_SETTINGS, sptr->name, me.name);
	return;
    }

    for (as = slines; as; as = as->next)
    {
	sendto_one(sptr, ":%s %i %s :S %s@%s %s %s (%s)", me.name, RPL_TEXT, sptr->name, as->username, as->hostname, as->reason, as->set_on, 
	    as->set_by);
    }
}

/*
 * m_rehash()
 */
int m_rehash(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int x;

    if (MyClient(sptr) && !OFlagr(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (!MyClient(sptr) && !(HasUmodeT(sptr) || HasUmodeN(sptr)) && !IsULine(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    x = 0;

    if (parc > 1) 
    {
	if (mycmp(parv[1], "-dns") == 0)
	{
	    flush_cache();	/* flush the dns cache */
	    res_init();		/* re-read /etc/resolv.conf */

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "DNS");
	    sendto_umode(MODE_s, "%s is rehashing DNS while whistling innocently", parv[0]);
	    ircd_log("%s is rehashing DNS while whistling innocently", parv[0]);
	    return 0;
	}
	if (mycmp(parv[1], "-szlines") == 0)
	{
	    remove_szline("*", 1);	/* flush the dns cache */

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "SLINES");
	    sendto_umode(MODE_s, "%s is removing SZLINES while whistling innocently", parv[0]);
	    ircd_log("%s is removing SZLINES while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-tklines") == 0)
	{
	    flush_temp_klines();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "temp klines");
	    sendto_umode(MODE_s, "%s is clearing temp klines while whistling innocently", parv[0]);
	    ircd_log("%s is clearing temp klines while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-gar") == 0)
	{
	    block_garbage_collect();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "garbage collecting");
	    sendto_umode(MODE_s, "%s is garbage collecting while whistling innocently", parv[0]);
	    ircd_log("%s is garbage collecting while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-motd") == 0)
	{
	    read_motd(MOTD);

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], MOTD);
	    sendto_umode(MODE_s, "Rehashing MOTD file (conf/ircd.motd) on request of \2%s\2", parv[0]);
	    ircd_log("Rehashing MOTD file (conf/ircd.motd) on request of %s.", parv[0]);

	    if (DYN_SHORTMOTD_STATE == 1)
	    {
		read_shortmotd(SHORTMOTD);

		if (MyClient(sptr))
		    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], SHORTMOTD);
		sendto_umode(MODE_s, "Rehashing SHORTMOTD file (%s) on request of \2%s\2", SHORTMOTD, parv[0]);
		ircd_log("Rehashing SHORTMOTD file (%s) on request of %s.", SHORTMOTD, parv[0]);
	    }
	    return (0);
	}
	else if (mycmp(parv[1], "-ip") == 0)
	{
	    rehash_ip_hash();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "ip hash");
	    sendto_umode(MODE_s, "Rehashing iphash on request of \2%s\2", parv[0]);
	    ircd_log("Rehashing iphash on request of %s.", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-akills") == 0)
	{
	    do_rehash_akills();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "akills");
	    sendto_umode(MODE_s, "Rehashing akills on request of \2%s\2", parv[0]);
	    ircd_log("Rehashing akills on request of %s.", parv[0]);
	    return 0;
	}
	else if(mycmp(parv[1], "-throttles") == 0)
	{
	    throttle_rehash();
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "throttles");
	    sendto_umode(MODE_s, "%s is rehashing throttles", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-chr") == 0)
	{
	    cr_rehash();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str (RPL_REHASHING), me.name, parv[0], IRCD_RESTRICT);
	    sendto_umode(MODE_s, "Rehashing channel restrict configuration file (%s) on request of \2%s\2", IRCD_RESTRICT, parv[0]);
	    ircd_log("Rehashing channel restrict configuration file (%s) on request of %s.", IRCD_RESTRICT, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-dcc") == 0)
	{
	    dcc_rehash();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], IRCD_DCCDENY);
	    sendto_umode(MODE_s, "Rehashing dcc restrict configuration file (%s) on request of \2%s\2", IRCD_DCCDENY, parv[0]);
	    ircd_log("Rehashing dcc restrict configuration file (%s) on request of %s.", IRCD_DCCDENY, parv[0]);
	    return 0;
        }
	else if (mycmp(parv[1], "-dyn") == 0)
	{
	    load_conf(DCONF, 1);
	    build_protoctl();

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], DCONF);
	    sendto_umode(MODE_s, "Rehashing dynamic configuration file (%s) on request of \2%s\2", DCONF, parv[0]);
	    ircd_log("Rehashing dynamic configuration file (%s) on request of %s.", DCONF, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-help") == 0)
        {
	    read_help(HELPFILE);

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], HELPFILE);
            sendto_umode(MODE_s, "Rehashing help file (%s) on request of \2%s\2", HELPFILE, parv[0]);
	    ircd_log("Rehashing help file (%s) on request of %s", HELPFILE, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-omotd") == 0)
	{
	    read_opermotd(OMOTD);

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], OMOTD);
	    sendto_umode(MODE_s, "Rehashing OPERMOTD file (%s) on request of \2%s\2", OMOTD, parv[0]);
	    ircd_log("Rehashing OPERMOTD file (%s) on request of %s", OMOTD, parv[0]);
	    return (0);
	}
	else if (mycmp(parv[1], "-rules") == 0)
	{
	    read_rules(RULES);

	    if (MyClient(sptr))
		sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], RULES);
	    sendto_umode(MODE_s, "Rehashing RULES file (%s) on request of \2%s\2", RULES, parv[0]);
	    ircd_log("Rehashing RULES file (%s) on request of %s", RULES, parv[0]);
	    return (0);
	}
	if (MyClient(sptr) && !(HasUmodeN(sptr) || HasUmodeT(sptr)))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
/* This little number allows us to send a -param to another server, but only if it exists */
	if (parv[2] != NULL)
	{
	    if ((x = hunt_server(cptr, sptr, ":%s REHASH %s %s", 1, parc, parv)) != HUNTED_ISME)
		return 0;
	}
	if (parv[2] == NULL)
	{
	    if ((x = hunt_server(cptr, sptr, ":%s REHASH %s", 1, parc, parv)) != HUNTED_ISME)
		return 0;
	}
    }
    if (cptr != sptr)
    {
	if (parv[2] == NULL)
	{
	    sendto_serv_butone(&me, ":%s GLOBOPS :Remote static server configuration file (ircd.conf) rehash ordered by \2%s\2.", me.name, sptr->name);
	    sendto_umode(MODE_s, "Remote static server configuration file (%s) rehash ordered by \2%s\2.", configfile, parv[0]);
	    ircd_log("Remote static server configuration file (%s) rehash ordered by %s.", configfile, parv[0]);
	    return rehash(cptr, sptr, (parc > 1) ? ((*parv[1] == 'q') ? 2 : 0) : 0);
	}
	else
	{
	    if (mycmp(parv[2], "-dns") == 0)
	    {
		flush_cache();		/* flush the dns cache */
		res_init();		/* re-read /etc/resolv.conf */
		sendto_umode(MODE_s, "Remote DNS rehash ordered by \2%s\2.", parv[0]);
		ircd_log("Remote DNS rehash ordered by %s.", parv[0]);
		return 0;
	    }
	    if (mycmp(parv[2], "-szlines") == 0)
	    {
		remove_szline("*", 1);		/* flush the dns cache */
		sendto_umode(MODE_s, "%s is remotely removing SZLINES while whistling innocently", parv[0]);
		ircd_log("%s is remotely removing SZLINES while whistling innocently", parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-tklines") == 0)
	    {
		flush_temp_klines();
		sendto_umode(MODE_s, "%s is clearing temp klines while whistling innocently", parv[0]);
		ircd_log("%s is clearing temp klines while whistling innocently", parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-gar") == 0)
	    {
		block_garbage_collect();
		sendto_umode(MODE_s, "%s is remotely garbage collecting while whistling innocently", parv[0]);
		ircd_log("%s is remotely garbage collecting while whistling innocently", parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "motd") == 0)
	    {
		read_motd(MOTD);
		sendto_umode(MODE_s, "Remotely rehashing MOTD file (%s) on request of \2%s\2", MOTD, parv[0]);
		ircd_log("Remotely rehashing MOTD file (%s) on request of %s.", MOTD, parv[0]);

		if (DYN_SHORTMOTD_STATE == 1)
		{
		    read_shortmotd(SHORTMOTD);
		    sendto_umode(MODE_s, "Remotely rehashing SHORTMOTD file (%s) on request of \2%s\2", SHORTMOTD, parv[0]);
		    ircd_log("Remotely rehashing SHORTMOTD file (%s) on request of %s.", SHORTMOTD, parv[0]);
		}
		return (0);
	    }
	    else if (mycmp(parv[2], "-ip") == 0)
	    {
		rehash_ip_hash();
		sendto_ops("Remotely rehashing iphash on request of \2%s\2", parv[0]);
		ircd_log("Remotely rehashing iphash on request of %s.", parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-akills") == 0)
	    {
		do_rehash_akills();
		sendto_ops("Rehashing akills on request of \2%s\2", parv[0]);
		ircd_log("Rehashing akills on request of %s.", parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-chr") == 0)
	    {
		cr_rehash();
		sendto_ops("Remotely rehashing channel restrict configuration file (%s) on request of \2%s\2", IRCD_RESTRICT, parv[0]);
		ircd_log("Remotely rehashing channel restrict configuration file (%s) on request of %s.", IRCD_RESTRICT, parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-dcc") == 0)
	    {
		dcc_rehash();
		sendto_ops("Remotely rehashing dcc restrict configuration file (%s) on request of \2%s\2", IRCD_DCCDENY, parv[0]);
		ircd_log("Remotely rehashing dcc restrict configuration file (%s) on request of %s.", IRCD_DCCDENY, parv[0]);
		return 0;
            }
	    else if (mycmp(parv[2], "-dyn") == 0)
	    {
		load_conf(DCONF, 1);
		build_protoctl();
		sendto_ops("Remotely rehashing dynamic configuration file (%s) on request of \2%s\2", DCONF, parv[0]);
		ircd_log("Remotely rehashing dynamic configuration file (%s) on request of %s.", DCONF, parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-help") == 0)
            {
		read_help(HELPFILE);
		sendto_ops("Remotely rehashing help file (%s) on request of \2%s\2", HELPFILE, parv[0]);
		ircd_log("Remotely rehashing help file (%s) on request of %s", HELPFILE, parv[0]);
		return 0;
	    }
	    else if (mycmp(parv[2], "-omotd") == 0)
	    {
		read_opermotd(OMOTD);
		sendto_ops("Remotely rehashing OPERMOTD file (%s) on request of \2%s\2", OMOTD, parv[0]);
		ircd_log("Remotely rehashing OPERMOTD file (%s) on request of %s", OMOTD, parv[0]);
		return (0);
	    }
	    else if (mycmp(parv[2], "-rules") == 0)
	    {
		read_rules(RULES);
		sendto_ops("Remotely rehashing RULES file (%s) on request of \2%s\2", RULES, parv[0]);
#ifdef USE_SYSLOG
		ircd_log("Remotely rehashing RULES file (%s) on request of %s", RULES, parv[0]);
#endif
		return (0);
	    }
	}
    }
    else 
    {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], configfile);
	sendto_umode(MODE_s, "Rehashing static server configuration file (%s) on request of \2%s\2", configfile, parv[0]);
#ifdef USE_SYSLOG
	ircd_log("Rehashing static server configuration file (%s) on request of %s", configfile, parv[0]);
#endif
	return rehash(cptr, sptr, (parc > 1) ? ((*parv[1] == 'q') ? 2 : 0) : 0);
    }
    return 0;			/* shouldn't ever get here */
}

/*
 * m_grehash()
 */
int m_grehash(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!MyClient(sptr) && !(HasUmodeT(sptr) || HasUmodeN(sptr)) && !IsULine(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc > 1) 
    {
	if (mycmp(parv[1], "-dns") == 0)
	{
	    flush_cache();	/* flush the dns cache */
	    res_init();		/* re-read /etc/resolv.conf */
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "DNS");
	    sendto_serv_butone(cptr, ":%s REHASH -dns", parv[0]);
	    sendto_umode(MODE_s, "Global DNS rehash ordered by \2%s\2.", parv[0]);
	    ircd_log("Global DNS rehash ordered by %s.", parv[0]);
	    return 0;
	}
	if (mycmp(parv[1], "-szlines") == 0)
	{
	    remove_szline("*", 1); /* flush the dns cache */
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "SLINES");
	    sendto_serv_butone(cptr, ":%s REHASH -szlines", parv[0]);
	    sendto_umode(MODE_s, "%s is removing SZLINES while whistling innocently", parv[0]);
	    ircd_log("%s is removing SZLINES while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-tklines") == 0)
	{
	    flush_temp_klines();
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "temp klines");
	    sendto_serv_butone(cptr, ":%s REHASH -tklines", parv[0]);
	    sendto_umode(MODE_s, "%s is clearing temp klines while whistling innocently", parv[0]);
	    ircd_log("%s is clearing temp klines while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-gar") == 0)
	{
	    block_garbage_collect();
	    sendto_serv_butone(cptr, ":%s REHASH -gar", parv[0]);
	    sendto_umode(MODE_s, "%s is garbage collecting while whistling innocently", parv[0]);
	    ircd_log("%s is garbage collecting while whistling innocently", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-motd") == 0)
	{
	    read_motd(MOTD);
	    sendto_serv_butone(cptr, ":%s REHASH -motd", parv[0]);
	    sendto_umode(MODE_s, "Global MOTD file (%s) rehash ordered by \2%s\2.", MOTD, parv[0]);
	    ircd_log("Global MOTD file (%s) rehash ordered by %s.", MOTD, parv[0]);

	    if (DYN_SHORTMOTD_STATE == 1)
	    {
		read_shortmotd(SHORTMOTD);
		sendto_umode(MODE_s, "Global SHORTMOTD file (%s) rehash ordered by \2%s\2.", SHORTMOTD, parv[0]);
		ircd_log("Global SHORTMOTD file (%s) rehash ordered by %s.", SHORTMOTD, parv[0]);
	    }
	    return (0);
	}
	else if (mycmp(parv[1], "-ip") == 0)
	{
	    rehash_ip_hash();
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "ip hash");
	    sendto_serv_butone(cptr, ":%s REHASH -ip", parv[0]);
	    sendto_umode(MODE_s, "Global iphash rehash ordered by \2%s\2.", parv[0]);
	    ircd_log("Global iphash rehash ordered by %s.", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-akills") == 0)
	{
	    do_rehash_akills();
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], "akills");
	    sendto_serv_butone(cptr, ":%s REHASH -akills", parv[0]);
	    sendto_umode(MODE_s, "Global akills rehash ordered by \2%s\2.", parv[0]);
	    ircd_log("Global akills rehash ordered by %s.", parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-chr") == 0)
	{
	    cr_rehash();
	    sendto_serv_butone(cptr, ":%s REHASH -chr", parv[0]);
	    sendto_umode(MODE_s, "Global channel restrict configuration file (%s) rehash ordered by \2%s\2.", IRCD_RESTRICT, parv[0]);
	    ircd_log("Global channel restrict configuration file (%s) rehash ordered by %s.", IRCD_RESTRICT, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-dcc") == 0)
	{
	    dcc_rehash();
	    sendto_one (sptr, rpl_str(RPL_REHASHING), me.name, parv[0], IRCD_DCCDENY);
	    sendto_serv_butone(cptr, ":%s REHASH -dcc", parv[0]);
	    sendto_umode(MODE_s, "Global dcc restrict configuration file (%s) rehash ordered by \2%s\2.", IRCD_DCCDENY, parv[0]);
	    ircd_log("Global dcc restrict configuration file (%s) rehash ordered by %s.", IRCD_DCCDENY, parv[0]);
	    return 0;
        }
	else if (mycmp(parv[1], "-dyn") == 0)
	{
	    load_conf(DCONF, 1);
	    build_protoctl();
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], DCONF);
	    sendto_serv_butone(cptr, ":%s REHASH -dyn", parv[0]);
	    sendto_ops("Global dynamic configuration file (%s) rehash ordered by \2%s\2.", DCONF, parv[0]);
	    ircd_log("Global dynamic configuration file (%s) rehash ordered by %s.", DCONF, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-help") == 0)
	{
	    read_help(HELPFILE);
	    sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], HELPFILE);
            sendto_ops("Global HELP file (%s) rehash ordered by \2%s\2.", HELPFILE, parv[0]);
	    sendto_serv_butone(cptr, ":%s REHASH -help", parv[0]);
	    ircd_log("Global HELP file (%s) rehash ordered by %s.", HELPFILE, parv[0]);
	    return 0;
	}
	else if (mycmp(parv[1], "-omotd") == 0)
	{
	    read_opermotd(OMOTD);
	    sendto_serv_butone(cptr, ":%s REHASH -omotd", parv[0]);
	    sendto_ops("Global OPERMOTD file (%s) rehash ordered by \2%s\2.", OMOTD, parv[0]);
	    ircd_log("Global OPERMOTD file (%s) rehash ordered by %s.", OMOTD, parv[0]);
	    return (0);
	}
	else if (mycmp(parv[1], "-rules") == 0)
	{
	    read_rules(RULES);
	    sendto_umode(MODE_s, "Global RULES file (%s) rehash ordered by \2%s\2.", RULES, parv[0]);
	    sendto_serv_butone(cptr, ":%s REHASH -rules", parv[0]);
	    ircd_log("Global RULES file (%s) rehash ordered by %s.", RULES, parv[0]);
	    return (0);
	}
    }
    else 
    {
	sendto_one(sptr, rpl_str(RPL_REHASHING), me.name, parv[0], configfile);
	sendto_umode(MODE_s, "Global static server configuration file (%s) rehash ordered by \2%s\2.", configfile, parv[0]);
#ifdef USE_SYSLOG
	ircd_log("Global static server configuration file (%s) rehash ordered by %s.", configfile, parv[0]);
#endif
	sendto_serv_butone(&me, ":%s REHASH", parv[0]);
	return rehash(cptr, sptr, (parc > 1) ? ((*parv[1] == 'q') ? 2 : 0) : 0);
    }
    return 0;	/* shouldn't ever get here */
}

/*
 * m_restart()
 */
int m_restart(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *pass = NULL, *encr;
#ifdef CRYPT_XLINE_PASSWORD
# ifdef CRYPTXPASS_SED
    extern char *crypt();
# endif
#endif

    if (MyClient(sptr) && !OFlagR(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if ((pass = (char *) find_restartpass())) 
    {
	if (parc < 2) 
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "RESTART");
	    return 0;
	}
#ifdef CRYPT_XLINE_PASSWORD
# ifdef CRYPTXPASS_MD5
    if (parv[1] && pass)
	encr = libshadow_md5_crypt(parv[1], pass);
    else
	encr = "";
# else
    if (parv[1] && pass)
	encr = crypt(parv[1], pass);
    else
	encr = "";
# endif /* CRYPTXPASS_MD5 */
#else
    encr = parv[1];
#endif /* CRYPT_OLINE_PASSWORD */

	if (strcmp(pass, encr)) 
	{
	    sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	    return 0;
	}
    }
    sendto_umode(MODE_s, "IRCd is restarting by request of %s", parv[0]);
#ifdef USE_SYSLOG
    ircd_log("IRCd RESTART by %s", get_client_name(sptr, FALSE));
#endif
    sprintf(buf, "Server RESTART by %s", get_client_name(sptr, HIDEME));
    restart(buf);
    return 0;
}

/*
 * m_trace()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_trace(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;
    aClient *acptr=NULL;
    aClass *cltmp;
    char *tname;
    int doall, link_s[MAXCONNECTIONS], link_u[MAXCONNECTIONS];
    int cnt = 0, wilds = 0, dow = 0;
	
    tname = (parc > 1) ? parv[1] : me.name;

    if (!IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (DYN_HIDE_ULINES == 1)
    {
	if ((acptr = next_client_double(client, tname)))
	{
	    if (!(IsAnOper(sptr)) && IsULine(acptr))
	    {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return 0;
	    }
	    acptr = NULL; /* shrug, we borrowed it, reset it just in case */
	}
    }
    if (parc > 2)
	if (hunt_server(cptr, sptr, ":%s TRACE %s :%s", 2, parc, parv))
	    return 0;

    switch (hunt_server(cptr, sptr, ":%s TRACE :%s", 1, parc, parv)) {
    case HUNTED_PASS:	/*  note: gets here only if parv[1] exists */
    {
	aClient *ac2ptr = next_client_double(client, tname);
	if (ac2ptr)
	    sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0], version, debugmode, tname, ac2ptr->from->name);
	else
	    sendto_one(sptr, rpl_str(RPL_TRACELINK), me.name, parv[0], version, debugmode, tname, "ac2ptr_is_NULL!!");
	return 0;
    }
    case HUNTED_ISME:
	break;
    default:
	return 0;
    }
    if (parv[1] && !strchr(parv[1],'.') && (strchr(parv[1], '*') || strchr(parv[1], '?'))) 
    {
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name, parv[0], parv[1]);
	return 0;        
    }
    sendto_umode(MODE_e, "/trace requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);

    doall = (parv[1] && (parc > 1)) ? !match(tname, me.name) : TRUE;
    wilds = !parv[1] || strchr(tname, '*') || strchr(tname, '?');
    dow = wilds || doall;
    if (!IsAnOper(sptr) || !dow) /* non-oper traces must be full nicks */
	/* lets also do this for opers tracing nicks */
    {
	char *name;
	int class;
	acptr = hash_find_client(tname,(aClient *)NULL);

	if (!acptr || !IsPerson(acptr)) 
	{
	    /* this should only be reached if the matching target is this server */
	    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name, parv[0], tname);
	    return 0;
			  
	}
	name = get_client_name(acptr,FALSE);
	class = get_client_class(acptr);

	if (IsAnOper(acptr)) 
	{
	    sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR), me.name, parv[0], class, name, timeofday - acptr->lasttime);
	}
	else
	{
	    sendto_one(sptr,rpl_str(RPL_TRACEUSER), me.name, parv[0], class, name, timeofday - acptr->lasttime);
	}
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE),me.name, parv[0], tname);
	return 0;	 
    }
    if (dow && lifesux && !HasUmodeo(sptr)) 
    {
	sendto_one(sptr, rpl_str(RPL_LOAD2HI), me.name, parv[0]);
	return 0;
    }
	
    memset((char *) link_s, '\0', sizeof(link_s));
    memset((char *) link_u, '\0', sizeof(link_u));
    /* Count up all the servers and clients in a downlink. */

    if (doall)
	for (acptr = client; acptr; acptr = acptr->next) 
	{
	    if (IsPerson(acptr))
		link_u[acptr->from->fd]++;
	    else if (IsServer(acptr))
		link_s[acptr->from->fd]++;
	}
	
    /* report all direct connections */
	
    for (i = 0; i <= highest_fd; i++) 
    {
	char *name;
	int class;
		
	if (!(acptr = local[i]))	/* Local Connection? */
	    continue;
	if (!HasUmodeo(sptr) && !IsAnOper(acptr) && (acptr != sptr))
	    continue;
	if (!doall && wilds && match(tname, acptr->name))
	    continue;
	if (!dow && mycmp(tname, acptr->name))
	    continue;
	if (IsAnOper(sptr))
	    name = get_client_name(acptr, FALSE);
	else
	    name = get_client_name(acptr, HIDEME);
	class = get_client_class(acptr);
		
	switch (acptr->status) 
	{
	case STAT_CONNECTING:
	    sendto_one(sptr, rpl_str(RPL_TRACECONNECTING), me.name, parv[0], class, name);
	    cnt++;
	    break;
	case STAT_HANDSHAKE:
	    sendto_one(sptr, rpl_str(RPL_TRACEHANDSHAKE), me.name, parv[0], class, name);
	    cnt++;
	    break;
	case STAT_ME:
	    break;
	case STAT_UNKNOWN:
	    sendto_one(sptr, rpl_str(RPL_TRACEUNKNOWN), me.name, parv[0], class, name, acptr->firsttime ? timeofday - acptr->firsttime : -1);
	    cnt++;
	    break;
	case STAT_CLIENT:
	    if (HasUmodeo(sptr) || (IsAnOper(acptr) && !HasUmodei(acptr))) 
	    {
		if (IsAnOper(acptr))
		    sendto_one(sptr, rpl_str(RPL_TRACEOPERATOR), me.name, parv[0], class, name, timeofday - acptr->lasttime);
		else
		    sendto_one(sptr,rpl_str(RPL_TRACEUSER), me.name, parv[0], class, name, timeofday - acptr->lasttime);
	    }
	    break;
	case STAT_SERVER:
	    sendto_one(sptr, rpl_str(RPL_TRACESERVER),
		       me.name, parv[0], class, link_s[i], link_u[i], name, 
		       *(acptr->serv->bynick) ? acptr->serv->bynick : "*", 
		       *(acptr->serv->byuser) ? acptr->serv->byuser : "*", 
		       *(acptr->serv->byhost) ? acptr->serv->byhost : me.name);
	    cnt++;
	    break;
	case STAT_LOG:
	    sendto_one(sptr, rpl_str(RPL_TRACELOG), me.name, parv[0], LOGFILE, acptr->port);
	    cnt++;
	    break;
	default:
	    sendto_one(sptr, rpl_str(RPL_TRACENEWTYPE), me.name, parv[0], name);
	    cnt++;
	    break;
	}
    }

    /* Add these lines to summarize the above which can get rather long and messy when done remotely - Avalon */

    if (!IsAnOper(sptr) || !cnt) 
    {
	if (cnt) 
	{
	    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	    return 0;
	}
	/* let the user have some idea that its at the end of the trace */
	sendto_one(sptr, rpl_str(RPL_TRACESERVER), me.name, parv[0], 0, link_s[me.fd],
	    link_u[me.fd], me.name, "*", "*", me.name, timeofday - acptr->lasttime);
	sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
	return 0;
    }
    for (cltmp = FirstClass(); doall && cltmp; cltmp = NextClass(cltmp))
	if (Links(cltmp) > 0)
	    sendto_one(sptr, rpl_str(RPL_TRACECLASS), me.name, parv[0], Class (cltmp), Links(cltmp));
    sendto_one(sptr, rpl_str(RPL_ENDOFTRACE), me.name, parv[0], tname);
    return 0;
}

/*
 * m_rules()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_rules(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static time_t last_used = 0L;

    if (hunt_server (cptr, sptr, ":%s RULES :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;
    if ((last_used + MOTD_WAIT) > NOW)
	return 0;
    else
	last_used = NOW;
    sendto_umode(MODE_e, "/rules requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
    send_rules(cptr, sptr, parc, parv);
    return 0;
}

/*
 * send_rules()
 *  parv[0] = sender prefix
 *  parv[1] = servername
 */
int send_rules(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aRules *temp;
    struct tm *tm;

    tm = rules_tm;

    if (rules == (aRules *) NULL)
    {
	sendto_one(sptr, err_str(ERR_NORULES), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr, rpl_str(RPL_RULESSTART), me.name, parv[0], me.name);

    if (tm)
	sendto_one(sptr, ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_RULES, parv[0], tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year,
	    tm->tm_hour, tm->tm_min);
    temp = rules;

    while (temp)
    {
	sendto_one(sptr, rpl_str(RPL_RULES), me.name, parv[0], temp->line);
	temp = temp->next;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFRULES), me.name, parv[0]);
    return 0;
}

/*
 * read_rules()
 */
void read_rules(char *filename)
{
    aRules *temp, *last;
    struct stat sb;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    while (rules)
    {
	temp = rules->next;
	MyFree (rules);
	rules = temp;
    }

    fd = open (filename, O_RDONLY);

    if (fd == -1)
	return;
    fstat (fd, &sb);
    rules_tm = localtime (&sb.st_mtime);
    last = (aRules *) NULL;

    while (dgets (fd, buffer, MOTDLINELEN - 1) > 0)
    {
	if ((tmp = (char *) strchr (buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr (buffer, '\r')))
	    *tmp = '\0';

	temp = (aRules *) MyMalloc (sizeof (aRules));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aRules *) NULL;

	if (!rules)
	    rules = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close (fd);

    if (rules_tm)
	(void) sprintf(rules_last_changed_date, "%d/%d/%d %d:%02d", rules_tm->tm_mday, rules_tm->tm_mon + 1, 1900 + rules_tm->tm_year,
	    rules_tm->tm_hour, rules_tm->tm_min);
}

/*
 * m_opermotd()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_opermotd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s OPERMOTD :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;
    if (!IsAnOper(sptr))
    {
	sendto_one (sptr, err_str (ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if ((last_used + MOTD_WAIT) > NOW)
	return 0;
    else
	last_used = NOW;
    sendto_umode(MODE_e, "/opermotd requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host,
	sptr->user->server);
    send_opermotd(cptr, sptr, parc, parv);
    return 0;
}

/*
 * send_opermotd
 *  parv[0] = sender prefix
 *  parv[1] = servername
 */
int send_opermotd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aOMotd *temp;
    struct tm  *tm;
    
    tm = omotd_tm;
    if (omotd == (aOMotd *) NULL) 
    {
	sendto_one(sptr, err_str(ERR_NOOPERMOTD), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr, rpl_str(RPL_OPERMOTDSTART), me.name, parv[0], me.name);

    if (tm)
	sendto_one(sptr, ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_MOTD, parv[0], tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year,
	    tm->tm_hour, tm->tm_min);

    temp = omotd;
    while (temp)
    {
	sendto_one(sptr, rpl_str(RPL_OPERMOTD),  me.name, parv[0], temp->line);
	temp = temp->next;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFOPERMOTD), me.name, parv[0]);
    return 0;
}

/*
 * read_opermotd()
 */
void read_opermotd(char *filename)
{
    aOMotd *temp, *last;
    struct stat sb;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /* Clear out the old MOTD */

    while (omotd) 
    {
	temp = omotd->next;
	MyFree(omotd);
	omotd = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;
    fstat(fd, &sb);
    omotd_tm = localtime(&sb.st_mtime);
    last = (aOMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
    {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aOMotd *) MyMalloc(sizeof(aOMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aOMotd *) NULL;
	if (!omotd)
	    omotd = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);

    if (omotd_tm)
	(void) sprintf(opermotd_last_changed_date, "%d/%d/%d %d:%02d", omotd_tm->tm_mday, omotd_tm->tm_mon + 1,
		1900 + omotd_tm->tm_year, omotd_tm->tm_hour, omotd_tm->tm_min);
}

/*
 * m_motd()
 *  parv[0] = sender prefix 
 *  parv[1] = servername
 */
int m_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    static time_t last_used = 0L;

    if (hunt_server(cptr, sptr, ":%s MOTD :%s", 1, parc, parv) != HUNTED_ISME)
	return 0;
    if ((last_used + MOTD_WAIT) > NOW)
	return 0;
    else
	last_used = NOW;
    sendto_umode(MODE_e, "/motd requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
    send_motd(cptr, sptr, parc, parv);
    return 0;
}

/*
** send_motd
**  parv[0] = sender prefix
**  parv[1] = servername
**
** This function split off so a server notice could be generated on a
** user requested motd, but not on each connecting client.
** -Dianora
*/
int send_motd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aMotd *temp;
    struct tm  *tm;
    
    tm = motd_tm;
    if (motd == (aMotd *) NULL) 
    {
	sendto_one(sptr, err_str(ERR_NOMOTD), me.name, parv[0]);
	return 0;
    }
    sendto_one(sptr, rpl_str(RPL_MOTDSTART), me.name, parv[0], me.name);

    if (tm)
	sendto_one(sptr, ":%s %d %s :- %d/%d/%d %d:%02d", me.name, RPL_MOTD, parv[0], tm->tm_mday, tm->tm_mon + 1, 1900 + tm->tm_year,
	    tm->tm_hour, tm->tm_min);

    temp = motd;
    while (temp) 
    {
	sendto_one(sptr, rpl_str(RPL_MOTD),  me.name, parv[0], temp->line);
	temp = temp->next;
    }
    sendto_one(sptr, rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
    return 0;
}

/*
 * read_motd() - From CoMSTuD, added Aug 29, 1996
 */
void read_motd(char *filename)
{
    aMotd *temp, *last;
    struct stat sb;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /* Clear out the old MOTD */

    while (motd) 
    {
	temp = motd->next;
	MyFree(motd);
	motd = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;
    fstat(fd, &sb);
    motd_tm = localtime(&sb.st_mtime);
    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
    {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!motd)
	    motd = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);

    if (motd_tm)
	(void) sprintf(motd_last_changed_date, "%d/%d/%d %d:%02d", motd_tm->tm_mday, motd_tm->tm_mon + 1, 1900 + motd_tm->tm_year, motd_tm->tm_hour, motd_tm->tm_min);
}

void read_shortmotd(char *filename)
{
    aMotd *temp, *last;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /* Clear out the old MOTD */

    while (shortmotd)
    {
	temp = shortmotd->next;
	MyFree(shortmotd);
	shortmotd = temp;
    }
    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;
    
    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
    {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!shortmotd)
	    shortmotd = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);
}

/*
 * read_help() - modified from from CoMSTuD's read_motd added Aug 29,
 * 1996 modifed  Aug 31 1997 - Dianora
 * 
 * Use the same idea for the oper helpfile
 */
void read_help(char *filename)
{
    aMotd *temp, *last;
    char buffer[MOTDLINELEN], *tmp;
    int fd;

    /* Clear out the old HELPFILE */

    while (helpfile) 
    {
	temp = helpfile->next;
	MyFree(helpfile);
	helpfile = temp;
    }

    fd = open(filename, O_RDONLY);
    if (fd == -1)
	return;

    last = (aMotd *) NULL;

    while (dgets(fd, buffer, MOTDLINELEN - 1) > 0) 
    {
	if ((tmp = (char *) strchr(buffer, '\n')))
	    *tmp = '\0';
	if ((tmp = (char *) strchr(buffer, '\r')))
	    *tmp = '\0';
	temp = (aMotd *) MyMalloc(sizeof(aMotd));

	strncpyzt(temp->line, buffer, MOTDLINELEN);
	temp->next = (aMotd *) NULL;
	if (!helpfile)
	    helpfile = temp;
	else
	    last->next = temp;
	last = temp;
    }
    close(fd);
}

/*
 * m_close()
 */
int m_close(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    int i;
    int closed = 0;

    if (!MyOper(sptr)) 
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    for (i = highest_fd; i; i--) 
    {
	if (!(acptr = local[i]))
	    continue;
	if (!IsUnknown(acptr) && !IsConnecting(acptr) && !IsHandshake(acptr))
	    continue;
	sendto_one(sptr, rpl_str(RPL_CLOSING), me.name, parv[0], get_client_name(acptr, TRUE), acptr->status);
	(void) exit_client(acptr, acptr, acptr, "Oper Closing");
	closed++;
    }
    sendto_one(sptr, rpl_str(RPL_CLOSEEND), me.name, parv[0], closed);
    sendto_umode(MODE_s, "%s (%s@%s) closed %d unknown connections", sptr->name,
	sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, closed);
    return 0;
}

/*
 * m_die()
 */
int m_die(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    int i;
    char *pass = NULL, *encr;
#ifdef CRYPT_XLINE_PASSWORD
# ifdef CRYPTXPASS_SED
    extern char *crypt();
# endif
#endif

    if (!OFlagD(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if ((pass = (char *) find_diepass())) 
    {
	if (parc < 2) 
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "DIE");
	    return 0;
	}
#ifdef CRYPT_XLINE_PASSWORD
# ifdef CRYPTXPASS_MD5
    if (parv[1] && pass)
	encr = libshadow_md5_crypt(parv[1], pass);
    else
	encr = "";
# else
    if (parv[1] && pass)
	encr = crypt(parv[1], pass);
    else
	encr = "";
# endif /* CRYPTXPASS_MD5 */
#else
    encr = parv[1];
#endif /* CRYPT_XLINE_PASSWORD */
	if (strcmp(pass, encr)) 
	{
	    sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
	    return 0;
	}
    }

    for (i = 0; i <= highest_fd; i++) 
    {
	if (!(acptr = local[i]))
	    continue;
	if (IsClient(acptr))
	    sendto_one(acptr, ":%s NOTICE %s :IRCd terminating by request of %s", me.name, acptr->name, sptr->name);
	else if (IsServer(acptr))
	    sendto_one(acptr, ":%s ERROR :Terminated by %s", me.name, get_client_name(sptr, TRUE));
    }
    (void) s_die();
    return 0;
}

/*
 * m_capab() 
 *  Communicate what I can do to another server 
 *  This has to be able to be sent and understood while
 *  the client is UNREGISTERED. Therefore, we
 *  absolutely positively must not check to see if
 *  this is a server or a client. It's probably an unknown!
 */
int m_capab(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    int i;

    /* If it's not local, or it has already set capabilities, silently ignore it. */

    if (cptr != sptr || cptr->capabilities)
	return 0;

    for (i = 1; i < parc; i++) 
    {
	if (strcmp(parv[i], "TS3") == 0)
	    SetTS3(cptr);
	else if (strcmp(parv[i], "TS5") == 0)
	    SetTS5(cptr);
	else if (strcmp(parv[i], "NOQUIT") == 0)
	    SetNoQuit(cptr);
	else if (strcmp(parv[i], "SSJOIN") == 0)
	    SetSSJoin(cptr);
	else if (strcmp(parv[i], "BURST") == 0)
	    SetBurst(cptr);
	else if (strcmp(parv[i], "UNCONNECT") == 0)
	    SetUnconnect(cptr);
	else if (strcmp(parv[i], "DKEY") == 0)
	    SetDKEY(cptr);
	else if (strcmp(parv[i], "ZIP") == 0)
	    SetZipCapable(cptr);
	else if (strcmp(parv[i], "NICKIP") == 0)
	    SetNICKIP(cptr);
    }
    return 0;
}

/* Shadowfax's LOCKFILE code */
#ifdef LOCKFILE
int lock_kline_file()
{
    int fd;
    /* Create Lockfile */

    if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) 
    {
	sendto_umode(MODE_s, "%s is locked, K:Lines pending", klinefile);
	pending_kline_time = time(NULL);
	return (-1);
    }
    (void) close(fd);
    return (1);
}

void do_pending_klines()
{
    int fd;
    char s[20];
    struct pkl *k, *ok;

    if (!pending_klines)
	return;

    /* Create Lockfile */
    if ((fd = open(LOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0666)) < 0) 
    {
	sendto_umode(MODE_s, "%s is locked, K:Lines pending", klinefile);
	pending_kline_time = time(NULL);
	return;
    }
    (void) ircsprintf(s, "%d\n", getpid());
    (void) write(fd, s, strlen(s));
    close(fd);

    /* Open klinefile */
    if ((fd = open(klinefile, O_WRONLY | O_APPEND)) == -1) 
    {
	sendto_umode(MODE_s, "Pending K:Lines cannot be written, cannot open %s", klinefile);
	unlink(LOCKFILE);
	return;
    }

    /* Add the Pending Klines */

    k = pending_klines;
    while (k) 
    {
	write(fd, k->comment, strlen(k->comment));
	write(fd, k->kline, strlen(k->kline));
	free(k->comment);
	free(k->kline);
	ok = k;
	k = k->next;
	free(ok);
    }
    pending_klines = NULL;
    pending_kline_time = 0;

    close(fd);

    /* Delete the Lockfile */
    unlink(LOCKFILE);
}
#endif

/*
 * m_svskill()
 *  parv[0] = servername
 *  parv[1] = client
 *  parv[2] = nick stamp
 *  parv[3] = kill message
 */ 
int m_svskill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aClient *acptr;
    char *comment;
    char reason[TOPICLEN + 1];
    ts_val ts = 0;

    if (parc > 3) 
    {
	comment = parv[3] ? parv[3] : parv[0];
	ts = atol(parv[2]);
    }
    else
	comment = (parc > 2 && parv[2]) ? parv[2] : parv[0];
      
    if (!IsULine(sptr))
	return -1;
    if ((acptr = find_client(parv[1], NULL)) && (!ts || ts == acptr->tsinfo))
    {
	if (MyClient(acptr))
	{
	    strcpy(reason, "SVSKilled: ");
	    strncpy(reason + 11, comment, TOPICLEN - 11);
	    reason[TOPICLEN] = '\0';
	    return exit_client(acptr, acptr, sptr, reason);
	}
	if (acptr->from == cptr)
	{
	    sendto_umode(MODE_d, "Received wrong-direction SVSKILL for %s (behind %s) from %s", acptr->name, cptr->name, get_client_name(sptr, HIDEME));
	    return 0;
	}
	else 
	    sendto_serv_butone(cptr, ":%s SVSKILL %s :%s", parv[0], parv[1], comment);
    }
    return 0;
}

/*
 * m_akill()
 *  parv[1]=host 
 *  parv[2]=user
 *  parv[3]=length
 *  parv[4]=akiller
 *  parv[5]=time set
 *  parv[6]=reason
 */      
int m_akill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aConfItem *aconf, *ac2;
    char *user, *host, *reason, *akiller, buffer[1024], *current_date;   
    time_t length=0, timeset=0;

    if (!IsServer(sptr) || (parc < 6))
	return 0;
    if (!IsULine(sptr)) 
    {
	sendto_serv_butone(&me,":%s GLOBOPS :Non U:Lined server %s trying to AKILL!", me.name, sptr->name);
	send_globops("from %s: Non U:Lined server %s trying to AKILL!", me.name, sptr->name);
	return 0;
    }

    host=parv[1];
    user=parv[2];
    akiller=parv[4];
    length=atoi(parv[3]);
    timeset=atoi(parv[5]);
    reason=(parv[6] ? parv[6] : "<no reason>");

    current_date=smalldate((time_t)timeset);
    /* cut reason down a little, eh? */
    /* 250 chars max */
    if (strlen(reason)>250)
	reason[251]=0;

    /* if this is already klined, don't akill it now */
    ac2=find_is_klined(host, user);

    if (ac2!=NULL) 
    {
	/* pass along the akill anyways */
	sendto_serv_butone(cptr, ":%s AKILL %s %s %d %s %d :%s", sptr->name, host, user, length, akiller, timeset, reason);
	return 0;
    }
    /* fill out aconf */
    aconf=make_conf();
    aconf->status=CONF_AKILL;
    DupString(aconf->host, host);
    DupString(aconf->name, user);
    
    ircsprintf(buffer, "%s (%s)", reason, current_date);

    DupString(aconf->passwd, buffer);
	        
    /* whether or not we have a length, this is still a temporary akill */
    /* if the length is over a day, or nonexistant, we call this a 'forever'
     * akill and set ->hold to 0xFFFFFFFF to indicate such
     * this is a hack to prevent
     * forever akills from being removed unless by an explicit /rehash */

    if (length>86400 || !length)
	aconf->hold=0xFFFFFFFF;  
    else
	aconf->hold=timeset+length;

    add_temp_kline(aconf);
    rehashed=YES;
    zline_in_progress=NO;

    /* now finally send it off to any other servers! */
    sendto_serv_butone(cptr, ":%s AKILL %s %s %d %s %d :%s", sptr->name, host, user, length, akiller, timeset, reason);
    return 0;
}

/*
 * m_rakill()
 */  
int m_rakill(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    if (!IsServer(sptr))
	return 0;
    /* just quickly find the akill and be rid of it! */
    if (parc<3) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "RAKILL");
	return 0;
    }
    if (!IsULine(sptr)) 
    {
	sendto_serv_butone(&me, ":%s GLOBOPS :Non U:Lined server %s trying to RAKILL!",  me.name, sptr->name);
	send_globops("from %s: Non U:Lined server %s trying to RAKILL!", me.name, sptr->name);
	return 0;
    }
    remove_temp_kline(parv[1], parv[2], CONF_AKILL);
    sendto_serv_butone(cptr, ":%s RAKILL %s %s", sptr->name, parv[1], parv[2]);
    return 0;
}

	
/*
 * RPL_NOWON   - Online at the moment (Succesfully added to WATCH-list)
 * RPL_NOWOFF  - Offline at the moement (Succesfully added to WATCH-list)
 * RPL_WATCHOFF   - Succesfully removed from WATCH-list.
 * ERR_TOOMANYWATCH - Take a guess :>  Too many WATCH entries.
 */
static void show_watch(aClient *cptr, char *name, int rpl1, int rpl2) 
{
    aClient *acptr;

    if ((acptr = find_person(name, NULL)))
	sendto_one(cptr, rpl_str(rpl1), me.name, cptr->name, acptr->name, acptr->user->username, HasUmodex(acptr) ? acptr->user->virthost
	    : acptr->user->host, acptr->lasttime);
    else
	sendto_one(cptr, rpl_str(rpl2), me.name, cptr->name, name, "*", "*", 0);
}
	
/*
 * m_watch()
 */
int m_watch(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aClient *acptr;
    char *s, *p, *user;
    char def[2] = "l";

    if (parc < 2) 
    {
	/* Default to 'l' - list who's currently online */
	parc = 2;
	parv[1] = def;
    }
    
    for (p = NULL, s = strtoken(&p, parv[1], ", "); s;
	 s = strtoken(&p, NULL, ", ")) 
    {
	if ((user = (char *)strchr(s, '!')))
	    *user++ = '\0'; /* Not used */
		
	/* Prefix of "+", they want to add a name to their WATCH list. */

	if (*s == '+') 
	{
	    if (*(s+1)) 
	    {
		if (sptr->watches >= MAXWATCH) 
		{
		    sendto_one(sptr, err_str(ERR_TOOMANYWATCH), me.name, cptr->name, s+1);					
		    continue;
		}				
		add_to_watch_hash_table(s+1, sptr);
	    }
	    show_watch(sptr, s+1, RPL_NOWON, RPL_NOWOFF);
	    continue;
	}
	
	/* Prefix of "-", coward wants to remove somebody from their WATCH list.  So do it. :-) */

	if (*s == '-') 
	{
	    del_from_watch_hash_table(s+1, sptr);
	    show_watch(sptr, s+1, RPL_WATCHOFF, RPL_WATCHOFF);
	    continue;
	}

	/*Fancy "C" or "c", they want to nuke their WATCH list and start over, so be it. */

	if (*s == 'C' || *s == 'c') 
	{
	    hash_del_watch_list(sptr);
	    continue;
	}
		
	/* Now comes the fun stuff, "S" or "s" returns a status report of
	 * their WATCH list.  I imagine this could be CPU intensive if its
	 * done alot, perhaps an auto-lag on this? */

	if (*s == 'S' || *s == 's') 
	{
	    Link *lp;
	    aWatch *anptr;
	    int count = 0;
							
	    /* Send a list of how many users they have on their WATCH list and how many WATCH lists they are on. */

	    anptr = hash_get_watch(sptr->name);

	    if (anptr)
		for (lp = anptr->watch, count = 1; (lp = lp->next); count++);
	    sendto_one(sptr, rpl_str(RPL_WATCHSTAT), me.name, parv[0], sptr->watches, count);
			
	    /* Send a list of everybody in their WATCH list. Be careful not to buffer overflow. */

	    if ((lp = sptr->watch) == NULL)
	    {
		sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
		continue;
	    }
	    *buf = '\0';
	    strcpy(buf, lp->value.wptr->nick);
	    count = strlen(parv[0])+strlen(me.name)+10+strlen(buf);
	    while ((lp = lp->next)) 
	    {
		if (count+strlen(lp->value.wptr->nick)+1 > BUFSIZE - 2) 
		{
		    sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name, parv[0], buf);
		    *buf = '\0';
		    count = strlen(parv[0])+strlen(me.name)+10;
		}
		strcat(buf, " ");
		strcat(buf, lp->value.wptr->nick);
		count += (strlen(lp->value.wptr->nick)+1);
	    }
	    sendto_one(sptr, rpl_str(RPL_WATCHLIST), me.name, parv[0], buf);
	    sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
	    continue;
	}

	/* Well that was fun, NOT.  Now they want a list of everybody in
	 * their WATCH list AND if they are online or offline? Sheesh,
	 * greedy arn't we? */

	if (*s == 'L' || *s == 'l') 
	{
	    Link *lp = sptr->watch;
			
	    while (lp) 
	    {
		if ((acptr = find_person(lp->value.wptr->nick, NULL)))
		    sendto_one(sptr, rpl_str(RPL_NOWON), me.name, parv[0], acptr->name, acptr->user->username, HasUmodex(acptr) ? acptr->user->virthost
			: acptr->user->host, acptr->tsinfo);

		/* But actually, only show them offline if its a capital 'L' (full list wanted). */

		else if (isupper(*s))
		    sendto_one(sptr, rpl_str(RPL_NOWOFF), me.name, parv[0], lp->value.wptr->nick, "*", "*", lp->value.wptr->lasttime);
		lp = lp->next;
	    }
	    sendto_one(sptr, rpl_str(RPL_ENDOFWATCHLIST), me.name, parv[0], *s);
	    continue;
	}
	/* Hmm.. unknown prefix character.. Ignore it. :-) */
    }
    return 0;
}

/*
 * m_svswatch()
 */
int m_svswatch(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *s, *p, *user;

    /*if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;*/
    if (!(acptr = find_person(parv[1], NULL)))
	return 0;
    for (p = NULL, s = strtoken(&p, parv[2], ", "); s;
	 s = strtoken(&p, NULL, ", "))
    {
	if ((user = (char *)strchr(s, '!')))
	    *user++ = '\0'; /* Not used */

	/* Prefix of "+", they want to add a name to their WATCH list. */

	if (*s == '+')
	{
	    if (*(s+1))
	    {
		if (acptr->watches >= MAXWATCH)
		{
		    sendto_one(sptr, err_str(ERR_TOOMANYWATCH), me.name, cptr->name, s+1);					
		    continue;
		}
		add_to_watch_hash_table(s+1, acptr);
	    }
	    show_watch(acptr, s+1, RPL_NOWON, RPL_NOWOFF);
	    continue;
	}

	/* Prefix of "-", coward wants to remove somebody from their WATCH list. So do it. :-) */

	if (*s == '-')
	{
	    del_from_watch_hash_table(s+1, acptr);
	    show_watch(acptr, s+1, RPL_WATCHOFF, RPL_WATCHOFF);
	    continue;
	}
	if (*s == '*') 
	{
	    hash_del_watch_list(acptr);
	    continue;
	}
	sendto_ops("%s SVSWATCH %s :%s", parv[0], parv[1], parv[2]);
    }
    return 0;
}

int m_sqline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aConfItem *aconf;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    if (parc < 2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SQLINE");
	return 0;
    }

    /* get rid of redundancies */
    parv[1] = collapse(parv[1]);

    /* if we have any Q:lines (SQ or Q) that match this Q:line, just return (no need to waste cpu */

    if (!(aconf=find_conf_name(parv[1], *parv[1]=='#' ? CONF_QUARANTINED_CHAN : CONF_QUARANTINED_NICK)))
    {
	/* okay, it doesn't suck, build a new conf for it */
	aconf=make_conf();

        /* Q:line and SQline, woo */
        if (*parv[1]=='#')
	    aconf->status=(CONF_QUARANTINED_CHAN|CONF_SQLINE); 
        else
	    aconf->status=(CONF_QUARANTINED_NICK|CONF_SQLINE);
	DupString(aconf->name, parv[1]);
	DupString(aconf->passwd, (parv[2]!=NULL ? parv[2] : "Reserved"));
	aconf->next=conf;
	conf=aconf;
    }
    sendto_serv_butone(cptr, ":%s SQLINE %s :%s", sptr->name, parv[1], aconf->passwd);
    return 0;
}
	
int m_unsqline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    int matchit = 0;
    char *mask;
    aConfItem *aconf, *ac2, *ac3;
    
    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    if (parc < 2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNSQLINE");
	return 0;
    }
    if (parc == 3) 
    {
	matchit=atoi(parv[1]);
	mask=parv[2];
    }
    else
	mask=parv[1];

    /* find the sqline and remove it */
    /* this was way too complicated and ugly. Fixed. -lucas */
    /* Changed this to use match.  Seems to make more sense? */
    /* Therefore unsqline * will clear the sqline list. */
    
    ac2 = ac3 = aconf = conf;
    while(aconf)
    {
	if ((aconf->status & (CONF_QUARANTINE)) && (aconf->status & (CONF_SQLINE)) && 
	    aconf->name && (matchit ? !match(mask, aconf->name) : !mycmp(mask, aconf->name))) 
	{
	    if (conf==aconf) 
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf=ac2;
	} 
	else
	{
	    ac3=aconf;
	    aconf=aconf->next;
	}
    }
    if (parc == 3) 
	sendto_serv_butone(cptr, ":%s UNSQLINE %d :%s", sptr->name, matchit, mask);
    else
	sendto_serv_butone(cptr, ":%s UNSQLINE :%s", sptr->name, mask);
    return 0;
}

int m_sgline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aConfItem *aconf;
    int len;
    char *mask, *reason;
    
    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    if (parc<3)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SGLINE");
	return 0;
    }

    len=atoi(parv[1]);
    mask=parv[2];

    if ((strlen(mask) > len) && (mask[len])==':') 
    {
	mask[len] = '\0';
	reason = mask+len+1;
    } 
    else 
    { /* Bogus */
	return 0;
    }
    
    /* if we have any G:lines (SG or G) that match this G:line, just return (no need to waste cpu */

    if (!(aconf=find_conf_name(mask, CONF_GCOS)))
    {
	/* okay, it doesn't suck, build a new conf for it */
	aconf=make_conf();
	aconf->status=(CONF_GCOS|CONF_SGLINE); /* G:line and SGline, woo */
	DupString(aconf->name, mask);
	DupString(aconf->passwd, reason ? reason : "Reason Unspecified");
	aconf->next=conf;
	conf=aconf;
    }
    sendto_serv_butone(cptr, ":%s SGLINE %d :%s:%s", sptr->name, len, aconf->name,aconf->passwd);
    return 0;
}
	
int m_unsgline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    int matchit=0;
    char *mask;
    aConfItem *aconf, *ac2, *ac3;
   
    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    if (parc<2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNSGLINE");
	return 0;
    }
    if (parc==3) 
    {
	matchit=atoi(parv[1]);
	mask=parv[2];
    }
    else
	mask=parv[1];

    /* find the sgline and remove it */
    /* this was way too complicated and ugly. Fixed. -lucas */
    /* Changed this to use match.  Seems to make more sense? */
    /* Therefore unsgline * will clear the sgline list. */
    ac2 = ac3 = aconf = conf;
    while(aconf)
    {
	if ((aconf->status & (CONF_GCOS)) && (aconf->status & (CONF_SGLINE)) && 
	    aconf->name && (matchit ? !match(mask, aconf->name) : !mycmp(mask, aconf->name))) 
	{
	    if (conf==aconf) 
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf=ac2;
	}
	else
	{
	    ac3=aconf;
	    aconf=aconf->next;
	}
    }   
    if (parc==3)
	sendto_serv_butone(cptr, ":%s UNSGLINE %d :%s", sptr->name, matchit, mask);
    else
	sendto_serv_butone(cptr, ":%s UNSGLINE :%s",sptr->name,mask);
    return 0;
}

int m_szline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aConfItem *aconf;

    if (!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    if (parc<2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SZLINE");
	return 0;
    }
    if (!(aconf=find_is_zlined(parv[1])))
    {
	/* okay, it doesn't suck, build a new conf for it */
	aconf=make_conf();
	aconf->status=(CONF_ZLINE|CONF_SZLINE); /* Z:line and SZline, woo */
	DupString(aconf->host, parv[1]);
	DupString(aconf->passwd, (parv[2]!=NULL ? parv[2] : "No Reason"));
	aconf->name=(char *)NULL;
	add_szline(aconf);
    }
    sendto_serv_butone(cptr, ":%s SZLINE %s :%s", sptr->name, parv[1], aconf->passwd);
    return 0;
}
	
int m_unszline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    int matchit = 0;
    char *mask;

    if (!IsServer(sptr) || !IsULine(sptr))
	return 0;
    if (parc<2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNSZLINE");
	return 0;
    }
    if (parc==3)
    {
	matchit=atoi(parv[1]);
	mask=parv[2];
    }
    else
	mask=parv[1];
    
    remove_szline(mask, matchit ? 1 : 0);
    if (parc==3)
	sendto_serv_butone(cptr, ":%s UNSZLINE %d %s", sptr->name, matchit, mask);
    else
	sendto_serv_butone(cptr, ":%s UNSZLINE %s", sptr->name, mask);
    return 0;
}

#define DKEY_GOTIN  0x01
#define DKEY_GOTOUT 0x02
#define DKEY_DONE(x) (((x) & (DKEY_GOTIN|DKEY_GOTOUT)) == (DKEY_GOTIN|DKEY_GOTOUT))

int m_dkey(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (!(IsNegoServer(sptr) && parc > 1))
    {
	if (IsPerson(sptr))
	    return 0;
	return exit_client(sptr, sptr, sptr, "Not negotiating now");
    }
#ifdef HAVE_ENCRYPTION_ON
    if (mycmp(parv[1], "START") == 0)
    {
	char keybuf[1024];

	if (parc != 2)
	    return exit_client(sptr, sptr, sptr, "DKEY START failure");

	if (sptr->serv->sessioninfo_in != NULL && sptr->serv->sessioninfo_out != NULL)
	    return exit_client(sptr, sptr, sptr, "DKEY START duplicate?!");

	sptr->serv->sessioninfo_in = dh_start_session();
	sptr->serv->sessioninfo_out = dh_start_session();

	sendto_umode(MODE_s, "Initiating diffie-hellman key exchange with %s", sptr->name);

	dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_in);
	sendto_one(sptr, "DKEY PUB I %s", keybuf);

	dh_get_s_public(keybuf, 1024, sptr->serv->sessioninfo_out);
	sendto_one(sptr, "DKEY PUB O %s", keybuf);
	return 0;
    }
    if (mycmp(parv[1], "PUB") == 0)
    {
	char keybuf[1024];

	if (parc != 4 || !sptr->serv->sessioninfo_in || !sptr->serv->sessioninfo_out)
	    return exit_client(sptr, sptr, sptr, "DKEY PUB failure");

	if (mycmp(parv[2], "O") == 0) /* their out is my in! */
	{
	    if (!dh_generate_shared(sptr->serv->sessioninfo_in, parv[3]))
		return exit_client(sptr, sptr, sptr, "DKEY PUB O invalid");
	    sptr->serv->dkey_flags |= DKEY_GOTOUT;
	}
	else if (mycmp(parv[2], "I") == 0) /* their out is my in! */
	{
	    if (!dh_generate_shared(sptr->serv->sessioninfo_out, parv[3]))
		return exit_client(sptr, sptr, sptr, "DKEY PUB I invalid");
	    sptr->serv->dkey_flags |= DKEY_GOTIN;
	}
	else
	    return exit_client(sptr, sptr, sptr, "DKEY PUB bad option");

	if (DKEY_DONE(sptr->serv->dkey_flags))
	{
	    int hexlen;
	    unsigned char hexbuf[1024];

	    sendto_one(sptr, "DKEY DONE");
	    SetRC4OUT(sptr);
	    
	    if (dh_get_s_shared(keybuf, 1024, sptr->serv->sessioninfo_in) == NULL || dh_hexstr_to_raw(keybuf, hexbuf, &hexlen) == 0)
		return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
	    sptr->serv->rc4_in = rc4_initstate(hexbuf, hexlen);

	    if (dh_get_s_shared(keybuf, 1024, sptr->serv->sessioninfo_out) == NULL || dh_hexstr_to_raw(keybuf, hexbuf, &hexlen) == 0)
		return exit_client(sptr, sptr, sptr, "Could not setup encrypted session");
	    sptr->serv->rc4_out = rc4_initstate(hexbuf, hexlen);
	    dh_end_session(sptr->serv->sessioninfo_in);
	    dh_end_session(sptr->serv->sessioninfo_out);
	    sptr->serv->sessioninfo_in = sptr->serv->sessioninfo_out = NULL;
	    return 0;
	}
	return 0;
    }
    if (mycmp(parv[1], "DONE") == 0)
    {
	if (!((sptr->serv->sessioninfo_in == NULL && sptr->serv->sessioninfo_out == NULL) && (sptr->serv->rc4_in != NULL && sptr->serv->rc4_out != NULL)))
	    return exit_client(sptr, sptr, sptr, "DKEY DONE when not done!");
	SetRC4IN(sptr);
	sendto_umode(MODE_s, "Diffie-Hellman exchange with %s complete, connection encrypted.", sptr->name);
	sendto_one(sptr, "DKEY EXIT");
	return RC4_NEXT_BUFFER;
    }
    if (mycmp(parv[1], "EXIT") == 0)
    {
	if (!(IsRC4IN(sptr) && IsRC4OUT(sptr)))
	    return exit_client(sptr, sptr, sptr, "DKEY EXIT when not in proper stage");
	ClearNegoServer(sptr);
	return do_server_estab(sptr);
    }
#endif
    return 0;
}

/*
 * m_sethost()
 *  parv[0] = sender
 *  parv[1] = newhost
 */
int m_sethost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *p;
    int bad_dns, dots;

    if (check_registered(sptr))
	return 0;
    if (!MyConnect(sptr))
	return 0;
    if (!IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc != 2)
	return 0;
    if (parv[1] == NULL)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETHOST");
	return 0;
    }
    if (strlen(parv[1]) < 4)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames must contain a minimum of 4 characters.", me.name, parv[0]);
	return 0;
    }
    if (strlen(parv[1]) > (HOSTLEN - 1))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames are limited to %i characters.", me.name, parv[0], (HOSTLEN - 1));
	return 0;
    }

    dots = 0;
    p = parv[1];
    bad_dns = NO;

    while (*p)
    {
	if (!isalnum(*p))
	{
	    if ((*p != '-') && (*p != '.'))
	    bad_dns = YES;
	}
	if (*p == '.')
	    dots++;
	p++;
    }

    if (bad_dns == YES)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- A hostname may contain a-z, 0-9, '-' & '.' - please use only these.", me.name, parv[0]);
	return 0;
    }
    if (!dots)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- You must have at least one '.' (dot) in the hostname.", me.name, parv[0]);
	return 0;
    }
    if (!HasUmodex(sptr))
    {
	SetUmodex(sptr);
	sendto_one(sptr, ":%s MODE %s :+x", sptr->name, sptr->name);
    }
    strncpyzt(sptr->user->virthost, parv[1], HOSTLEN);
    sendto_serv_butone(cptr, ":%s VS %s :%s", me.name, parv[0], parv[1]);
    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Your virtual hostname is now [%s]", me.name, parv[0], parv[1]);
    return 0;
}

/*
 * m_setident() (Based on m_setident() by Stskeeps)
 *  parv[0] = sender
 *  parv[1] = ident
 */
int m_setident(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *s;
    int legalident = 1;

    if (check_registered(sptr))
	return 0;
    if (MyConnect(sptr) && !IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }

    if (parc != 2)
	return 0;
    if (parv[1] == NULL)
    {
	if (MyConnect(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETIDENT");
	return 0;
    }
    if (strlen(parv[1]) < 1)
    {
	if (MyConnect(sptr))
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETIDENT");
	return 0;
    }
    if (strlen(parv[1]) > (USERLEN - 1))
    {
	if (MyConnect(sptr))
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Virtual idents are limited to %i characters.",
		me.name, sptr->name, parv[1], (USERLEN-1));
	return 0;
    }

    for (s = parv[1]; *s; s++)
    {
	if (!isallowed(*s))
	{
	    legalident = 0;
	}
	if (*s == '~')
	    legalident = 1;
    }

    if (legalident == 0)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- An Ident may contain a-z, 0-9, '-', '~' & '.' - please use only these",
	    me.name, parv[0]);
	return 0;
    }

    strcpy(sptr->user->username, parv[1]);
    sendto_serv_butone(cptr, ":%s SETIDENT %s", parv[0], parv[1]);

    if (MyConnect(sptr))
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Your virtual ident is now [%s]", me.name, parv[0], parv[1]);
    return 0;
}

/* 
 * m_setname() (Based on m_setname() by Stskeeps)
 *  parv[0] = sender prefix
 *  parv[1] = gecos
 */
int m_setname(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (check_registered(sptr))
	return 0;
    if (!IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc != 2)
	return 0;
    if (strlen(parv[1]) > (REALLEN - 2))
    {
	if (MyConnect(sptr))
	    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Realnames are limited to %i characters.", me.name, sptr->name, (REALLEN-2));
	return 0;
    }
    if (strlen(parv[1]) < 1)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETNAME");
	return 0;
    }

    strcpy(sptr->info, parv[1]);
    sendto_serv_butone(cptr, ":%s SETNAME %s", parv[0], parv[1]);

    if (MyConnect(sptr))
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Your realname is now [%s]", me.name, parv[0], parv[1]);
    return 0;
}

/*
 * m_chgident() (Based on m_chgident() by Stskeeps)
 *  parv[0] = sender
 *  parv[1] = nickname
 *  parv[2] = identname
 */
int m_chgident(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *s;
    int legalident = 1;

    if (check_registered(sptr))
	return 0;
    if (MyClient(sptr))
    {
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
    }
    if (parc != 3)
	return 0;
    if (strlen(parv[2]) < 1)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CHGIDENT");
	return 0;
    }
    if (strlen(parv[2]) > (USERLEN - 1))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Idents are limited to %i characters.", me.name, sptr->name, (USERLEN-1));
	return 0;
    }

    for (s = parv[2]; *s; s++)
    {
	if (!isallowed(*s))
	{
	    legalident = 0;
	}
    }

    if (legalident == 0)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- An Ident may contain a-z, 0-9, '-', '~' & '.' - please use only these",
	    me.name, parv[0]);
	return 0;
    }
    if ((acptr = find_person(parv[1], NULL)))
    {
	if (!IsULine(sptr))
	    /* someone forgot to check their work here, so I did. -OvErRiTe */
	    sendto_umode(MODE_s, "%s changed the virtual ident of %s (%s@%s) to be %s", sptr->name, acptr->name, acptr->user->username,
		acptr->user->host, parv[2]);
	strcpy(acptr->user->username, parv[2]);
	sendto_serv_butone(cptr, ":%s CHGIDENT %s :%s", sptr->name, acptr->name, parv[2]);
	return 0;
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, parv[1]);
	return 0;
    }
    return 0;
}

/* 
 * m_chghost() (Based on m_chghost() by Stskeeps)
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 *  parv[2] = hostname
 */
int m_chghost(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    char *p;
    int bad_dns, dots;

    if (check_registered(sptr))
	return 0;
    if (!MyClient(sptr))
	return 0;
    if (!IsAnOper(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc != 3)
	return 0;
    if (strlen(parv[2]) < 4)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames must contain a minimum of 4 characters.", me.name, sptr->name);
	return 0;
    }
    if (strlen(parv[2]) > (HOSTLEN - 1))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Hostnames are limited to %i characters.", me.name, sptr->name, (HOSTLEN-1));
	return 0;
    }

    dots = 0;
    p = parv[2];
    bad_dns = NO;

    while (*p)
    {
	if (!isalnum(*p))
	{
	    if ((*p != '-') && (*p != '.'))
	    bad_dns = YES;
	}
	if (*p == '.')
	    dots++;
	p++;
    }

    if (bad_dns == YES)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- A hostname may contain a-z, 0-9, '-' & '.' & ':'- please use only these.",
	    me.name, parv[0]);
	return 0;
    }
    if (!dots)
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- You must have at least one '.' (dot) in the hostname.", me.name, parv[0]);
	return 0;
    }
    if ((acptr = find_person(parv[1], NULL)))
    {
	if (!IsULine(sptr))
	{
	    /* someone forgot to double check their work here so I did. -OvErRiTe */
	    sendto_umode(MODE_s, "%s changed the virtual hostname of %s (%s@%s) to be %s", sptr->name, acptr->name,
		acptr->user->username, acptr->user->host, parv[2]);
	    sendto_serv_butone(cptr, ":%s SNOTICE :%s changed the virtual hostname of %s (%s@%s) to be %s", me.name, 
		sptr->name, acptr->name, acptr->user->username, acptr->user->host, parv[2]);
	}
	if (!HasUmodex(acptr))
	    SetUmodex(acptr);
	strncpyzt(acptr->user->virthost, parv[2], HOSTLEN);
	sendto_serv_butone(cptr, ":%s VS %s :%s", me.name, acptr->name, parv[2]);
	return 0;
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, parv[1]);
	return 0;
    }
    return 0;
}

/* 
 * m_chgname() (Based on m_chgname() by Stskeeps)
 *  parv[0] = sender prefix
 *  parv[1] = nickname
 *  parv[2] = realname
 */
int m_chgname(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;

    if (check_registered(sptr))
	return 0;
    if (MyClient(sptr))
    {
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
    }
    if (parc != 3)
	return 0;
    if (strlen(parv[2]) < 1)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "CHGNAME");
	return 0;
    }
    if (strlen(parv[2]) > (REALLEN - 1))
    {
	sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Realnames are limited to %i characters.", me.name, sptr->name, (REALLEN-1));
	return 0;
    }
    if ((acptr = find_person(parv[1], NULL)))
    {
	if (!IsULine(sptr))
	    sendto_umode(MODE_s, "%s changed the GECOS of %s (%s@%s) to be %s",
		sptr->name, acptr->name, acptr->user->username, HasUmodex(acptr) ? acptr->user->virthost : acptr->user->host, parv[2]);
	strcpy(acptr->info, parv[2]);
	sendto_serv_butone(cptr, ":%s CHGNAME %s %s", sptr->name, acptr->name, parv[2]);
	return 0;
    }
    else
    {
	sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, parv[1]);
	return 0;
    }
    return 0;
}

/*
 * m_addline() (Based on m_addline() by Stskeeps)
 */
int m_addline(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    FILE *conf;
    char *text;
    text = parc > 1 ? parv[1] : NULL;

    if (!MyConnect(sptr))
	return 0;
    if (!HasUmodeA(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ADDLINE");
	return 0;
    }
    conf = fopen(configfile, "a");

    if (conf == NULL)
	return 0;
    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Wrote (%s) to ircd.conf", me.name, parv[0], text);
    fprintf(conf, "# Added by %s!%s@%s\n", parv[0], sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host);
    fprintf(conf, "%s\n", text);
    fclose(conf);
    return 1;
}

/*
 * m_addmotd() (Based on m_addmotd() by Stskeeps)
 */
int m_addmotd(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    FILE *conf;
    char *text;
    text = parc > 1 ? parv[1] : NULL;

    if (!MyConnect(sptr))
	return 0;
    if (!HasUmodeA(sptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	return 0;
    }
    if (parc < 2)
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "ADDMOTD");
	return 0;
    }
    conf = fopen(MOTD, "a");

    if (conf == NULL)
 	return 0;
    sendto_one(sptr, ":%s NOTICE %s :*** Notice -- Wrote (%s) to file: ircd.motd", me.name, parv[0], text);
    fprintf(conf, "%s\n", text);
    fclose(conf);
    return 1;
}


/*
 * dump_map (used by m_map)
 */
void dump_map(aClient * cptr, aClient * server, char *mask, int prompt_length, int length)
{
    static char prompt[64];
    char *p = &prompt[prompt_length];
    int cnt = 0, local = 0;
    aClient *acptr;

    *p = '\0';

    if (prompt_length > 60)
	sendto_one(cptr, rpl_str(RPL_MAPMORE), me.name, cptr->name, prompt, server->name);
    else
    {
	for (acptr = client; acptr; acptr = acptr->next)
	{
	    if (IsPerson(acptr))
	    {
		++cnt;
		if (!strcmp(acptr->user->server, server->name))
		    ++local;
	    }
	}
	sendto_one(cptr, rpl_str(RPL_MAP), me.name, cptr->name, prompt, length, server->name, local, (local * 100) / cnt);
	cnt = 0;
    }
    if (prompt_length > 0)
    {
	p[-1] = ' ';
	if (p[-2] == '`')
		p[-2] = ' ';
    }
    if (prompt_length > 60)
	return;
    strcpy (p, "|-");

    for (acptr = client; acptr; acptr = acptr->next)
    {
	if (!IsServer(acptr) || strcmp(acptr->serv->up, server->name))
	    continue;
	if (match(mask, acptr->name))
	    acptr->flags &= ~FLAGS_MAP;
	else
	{
	    acptr->flags |= FLAGS_MAP;
	    cnt++;
	}
    }
    for (acptr = client; acptr; acptr = acptr->next)
    {
	if (IsULine(acptr) && DYN_HIDE_ULINES && !IsAnOper(cptr))
	    continue;
	if (!(acptr->flags & FLAGS_MAP) || !IsServer(acptr) || strcmp(acptr->serv->up, server->name))
	    continue;
	if (--cnt == 0)
	    *p = '`';
	dump_map(cptr, acptr, mask, prompt_length + 2, length - 2);
     }
     if (prompt_length > 0)
	p[-1] = '-';
}

/*
 * m_map()
 *  parv[0] = sender prefix
 *  parv[1] = server mask
 */
int m_map(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient *acptr;
    int longest = strlen (me.name);
    static time_t last_used = 0L;

    if (check_registered (sptr))
	return 0;
    if (!IsAnOper(sptr)) 
    {
	if ((last_used + MOTD_WAIT) > NOW)
	    return 0;
	else
	    last_used = NOW;

    }
    if ((DYN_ALLOW_MAP == 0) && !IsAnOper(cptr))
    {
	sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	sendto_umode(MODE_s, "/map requested by %s (%s@%s) [%s] - (non oper(request ignored))",
	    sptr->name, sptr->user->username, HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);
	return 0;
    }
    sendto_umode(MODE_s, "/map requested by %s (%s@%s) [%s]", sptr->name, sptr->user->username,
	HasUmodex(sptr) ? sptr->user->virthost : sptr->user->host, sptr->user->server);

    if (parc < 2)
	parv[1] = "*";

    for (acptr = client; acptr; acptr = acptr->next)
	if (IsServer(acptr) && (strlen(acptr->name) + acptr->hopcount * 2) > longest)
	    longest = strlen(acptr->name) + acptr->hopcount * 2;

    if (longest > 60)
	longest = 60;
    longest += 2;
    dump_map(sptr, &me, parv[1], 0, longest);
    sendto_one(sptr, rpl_str(RPL_MAPEND), me.name, parv[0]);
    return 0;
}

/*
 * m_lag()
 *  parv[0] = prefix
 *  parv[1] = server to query
 */
int m_lag(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    if (MyClient(sptr))
	if (!IsAnOper(sptr))
	{
	    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
	    return 0;
	}
	if (parc < 2)
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "LAG");
	    return 0;
	}
	if (*parv[1] == '\0')
	{
	    sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "LAG");
	    return 0;
	}
	if (hunt_server(cptr, sptr, ":%s LAG :%s", 1, parc, parv) == HUNTED_NOSUCH)
	{
	    return 0;
	}
	sendto_one(sptr, ":%s NOTICE %s :*** Lag reply -- %s %s %l", me.name, sptr->name, me.name, parv[1], timeofday);
	return 0;
}
